/*****************************************************************************/
/**
 *  @file   main.cpp
 *  @brief  Example program for kvs::Tubeline class.
 *  @author Naohisa Sakamoto
 */
/*----------------------------------------------------------------------------
 *
 *  Copyright (c) Visualization Laboratory, Kyoto University.
 *  All rights reserved.
 *  See http://www.viz.media.kyoto-u.ac.jp/kvs/copyright/ for details.
 *
 *  $Id: main.cpp 641 2010-10-17 02:15:17Z naohisa.sakamoto $
 */
/*****************************************************************************/
#include <kvs/Message>
#include <kvs/StructuredVolumeObject>
#include <kvs/StructuredVolumeImporter>
#include <kvs/LineObject>
#include <kvs/LineRenderer>
#include <kvs/Streamline>
#include <kvs/Tubeline>
#include <kvs/TornadoVolumeData>
#include <kvs/PolygonObject>
#include <kvs/glut/Application>
#include <kvs/glut/Screen>


/*===========================================================================*/
/**
 *  @brief  Main function.
 *  @param  argc [i] argument counter
 *  @param  argv [i] argument values
 */
/*===========================================================================*/
int main( int argc, char** argv )
{
    kvs::glut::Application app( argc, argv );

    /* Read volume data from the specified data file. If the data file is not
     * specified, scalar hydrogen volume data is created by using
     * kvs::HydrogenVolumeData class.
     */
    kvs::StructuredVolumeObject* volume = NULL;
    if ( argc > 1 ) volume = new kvs::StructuredVolumeImporter( std::string( argv[1] ) );
    else            volume = new kvs::TornadoVolumeData( kvs::Vector3ui( 16, 16, 16 ) );

    if ( !volume )
    {
        kvsMessageError( "Cannot create a structured volume object." );
        return( false );
    }

    std::vector<kvs::Real32> v;
    kvs::Vector3i min_coord( 7, 7,  0 );
    kvs::Vector3i max_coord( 10, 10, 15 );
    for ( int k = min_coord.z(); k < max_coord.z(); k++ )
    {
        for ( int j = min_coord.y(); j < max_coord.y(); j++ )
        {
            for ( int i = min_coord.x(); i < max_coord.x(); i++ )
            {
                v.push_back( static_cast<kvs::Real32>(i) );
                v.push_back( static_cast<kvs::Real32>(j) );
                v.push_back( static_cast<kvs::Real32>(k) );
            }
        }
    }
    kvs::PointObject* point = new kvs::PointObject( kvs::ValueArray<kvs::Real32>( v ) );
    if ( !point )
    {
        kvsMessageError( "Cannot creat a point object.");
        delete volume;
        return( false );
    }

    const kvs::TransferFunction transfer_function( 256 );
    kvs::LineObject* line = new kvs::Streamline( volume, point, transfer_function );
    if ( !line )
    {
        kvsMessageError( "Cannot creat a streamline object.");
        delete point;
        return( false );
    }

    line->setSize( 0.2 );

    delete point;
    delete volume;

    const size_t ndivisions = 20;
    kvs::PolygonObject* object = new kvs::Tubeline( line, ndivisions );
    if ( !object )
    {
        kvsMessageError( "Cannot creat a tubeline object.");
        delete line;
        return( false );
    }

    delete line;

    kvs::glut::Screen screen( &app );
    screen.registerObject( object );
    screen.setGeometry( 0, 0, 512, 512 );
    screen.setTitle( "kvs::Tubeline" );
    screen.show();

    return( app.run() );
}
