#ifdef ENABLE_MPI

#include "DistributedMetropolisSampling.h"
#include <algorithm>
#include <kvs/CellByCellParticleGenerator>
#include <kvs/ColorMap>
#include <kvs/RGBColor>
#include <kvs/CellBase>
#include <kvs/TetrahedralCell>
#include <kvs/QuadraticTetrahedralCell>
#include <kvs/HexahedralCell>
#include <kvs/QuadraticHexahedralCell>
#include <kvs/PyramidalCell>
#include <kvs/UnstructuredVolumeObject>
#include <mpi.h>
#include "mpi_controller.h"
/*** debug ***/
#include "common.h"
#include <cuda.h>
extern int time_series;
extern int ts_file_no;
extern int integral_index;
extern int random_index;
extern float TIMER[10][100];
/*************/

extern "C" {
size_t calc_particles_gpu_metropolis(
        size_t cell_type,
        size_t nnodes,
        size_t ncells,
        const unsigned int* connections,
        const float* coords,
        const float* values,
        const float* density_map,
        size_t density_resolution,
        float min_value,
        float max_value);

void generate_particles_gpu_metropolis(
        const unsigned char* color_map,
        size_t total_particles,
		kvs::ValueArray<kvs::Real32> &m_coords,
		kvs::ValueArray<kvs::UInt8>  &m_colors,
		kvs::ValueArray<kvs::Real32> &m_normals);
void pageLockedMemoryAlloc(
			   float** m_coords,
			   unsigned char** m_colors,
			   float** m_normals,
			   size_t n);
void pageLockedMemoryFree(
			  float** m_coords,
			  unsigned char** m_colors,
			  float** m_normals);
void pageableMemoryAlloc(
			   float** m_coords,
			   unsigned char** m_colors,
			   float** m_normals,
			   size_t n);
void pageableMemoryFree(
			  float** m_coords,
			  unsigned char** m_colors,
			  float** m_normals);
}

namespace Generator = kvs::CellByCellParticleGenerator;

namespace kvs {
namespace test {

void DistributedMetropolisSampling::generate_particles_core_gpu(
	const kvs::UnstructuredVolumeObject *volume,
	const float *density_map,
	const unsigned char *color_map,
	size_t resolution, float min_value, float max_value,
	kvs::ValueArray<kvs::Real32> *vertex_coords,
	kvs::ValueArray<kvs::UInt8>  *vertex_colors,
	kvs::ValueArray<kvs::Real32> *vertex_normals)
{
	int mpi_rank = 0, mpi_size = 0;
	MPI_Comm_rank(MPI_COMM_WORLD, &mpi_rank);
	MPI_Comm_size(MPI_COMM_WORLD, &mpi_size);

	std::cout << "node " << mpi_rank << " start." << std::endl;

	size_t cellType = (volume != NULL ? volume->cellType() : 0);
	size_t nnodes   = (volume != NULL ? volume->nnodes() : 0);
	size_t ncells   = (volume != NULL ? volume->ncells() : 0);

 	/*** debug ***/
	double start = get_time();
 	MPI_Bcast(&time_series, 1, MPI_INTEGER, 0, MPI_COMM_WORLD);
  	std::cout << mpi_rank << ":time_series = " << time_series << std::endl;
	if(time_series==0){
	MPI_Bcast(&cellType, 1, MPI_UNSIGNED, 0, MPI_COMM_WORLD);
	MPI_Bcast(&nnodes, 1, MPI_UNSIGNED, 0, MPI_COMM_WORLD);
	MPI_Bcast(&ncells, 1, MPI_UNSIGNED, 0, MPI_COMM_WORLD);
	}
	MPI_Bcast(&resolution, 1, MPI_UNSIGNED, 0, MPI_COMM_WORLD);
	MPI_Bcast(&min_value, 1, MPI_FLOAT, 0, MPI_COMM_WORLD);
	MPI_Bcast(&max_value, 1, MPI_FLOAT, 0, MPI_COMM_WORLD);
	/*************/

	kvs::Real32 *coords = NULL, *values = NULL;
	float *dist_density_map = NULL;
	unsigned char *dist_color_map = NULL;
	/*** debug ***/
	if(volume != NULL){
		coords = const_cast<kvs::Real32 *>(volume->coords().pointer());
		values = reinterpret_cast<kvs::Real32 *>(const_cast<void *>(volume->values().pointer()));
	}else{
		coords = new float[nnodes * 3];
		values = new float[nnodes];
	}
	if(density_map != NULL){
		dist_density_map = const_cast<float *>(density_map);
	}else{
		dist_density_map = new float[resolution];
	}
	if(color_map != NULL){
		dist_color_map = const_cast<unsigned char *>(color_map);
	}else{
		dist_color_map = new unsigned char[resolution * 3];
	}
	/*************/

	/*** debug ***/
	double generate_particles_start;
	size_t particles_num;
	MPI_Bcast(dist_density_map, resolution, MPI_FLOAT, 0, MPI_COMM_WORLD);
	MPI_Bcast(dist_color_map, resolution * 3, MPI_UNSIGNED_CHAR, 0, MPI_COMM_WORLD);

	if(time_series==0){
	MPI_Bcast(coords, nnodes * 3, MPI_FLOAT, 0, MPI_COMM_WORLD);
	MPI_Bcast(values, nnodes, MPI_FLOAT, 0, MPI_COMM_WORLD);

	int cells_per_node = (static_cast<int>(ncells) + mpi_size - 1) / mpi_size;

	std::vector<int> sendcnts(mpi_size), senddispls(mpi_size);
	for(int i = 0; i < mpi_size; ++i){
		sendcnts[i] =
			(std::min<int>(cells_per_node * (i + 1), ncells) - cells_per_node * i) * cellType;
		if(i != 0){ senddispls[i] = senddispls[i - 1] + sendcnts[i - 1]; }
	}
	std::vector<kvs::UInt32> connections(sendcnts[mpi_rank]);
	MPI_Scatterv(
		(volume != NULL ? const_cast<kvs::UInt32 *>(volume->connections().pointer()) : NULL),
		&sendcnts[0], &senddispls[0], MPI_UNSIGNED, &connections[0], connections.size(),
		MPI_UNSIGNED, 0, MPI_COMM_WORLD);

	generate_particles_start = get_time();
// 	printf("\n%d:GPU:data transfer: %lf ms\n", mpi_rank,generate_particles_start-start);
	/*** timer ***/
	TIMER[mpi_rank][1] += generate_particles_start-start;
	/*************/
	particles_num = calc_particles_gpu_metropolis(
		cellType, nnodes, sendcnts[mpi_rank] / cellType, &connections[0],
		&coords[0], &values[0], dist_density_map, resolution, min_value, max_value);

	}else{ // time_series==1
	generate_particles_start = get_time();
// 	printf("\n%d:GPU:data transfer: %lf ms\n", mpi_rank,generate_particles_start-start);
	/*** timer ***/
	TIMER[mpi_rank][1] += generate_particles_start-start;
	/*************/
	particles_num = calc_particles_gpu_metropolis(
	        cellType, nnodes, ncells, volume->connections().pointer(),
		&coords[0], &values[0], dist_density_map, resolution, min_value, max_value);
	}
	/*************/

 	kvs::ValueArray<kvs::Real32> m_coords;
  	kvs::ValueArray<kvs::UInt8> m_colors;
  	kvs::ValueArray<kvs::Real32> m_normals;
	generate_particles_gpu_metropolis(
					  dist_color_map, particles_num, m_coords, m_colors, m_normals);

	/*** debug ***/
	double generate_particles_end = get_time();
// 	printf("\n%d:GPU:generate_particles: %lf ms\n", mpi_rank,generate_particles_end - generate_particles_start);
	/*** timer ***/
	TIMER[mpi_rank][2] += generate_particles_end - generate_particles_start;
	/*************/
	/*************/

	if(time_series==0){
	unsigned int result_particles_num = m_coords.size();
	std::vector<unsigned int> particles_nums(mpi_size);
	MPI_Allgather(
		&result_particles_num, 1, MPI_UNSIGNED, &particles_nums[0],
		1, MPI_UNSIGNED, MPI_COMM_WORLD);

	unsigned int total_particles = 0;
	if(volume != NULL){
		for(int i = 0; i < mpi_size; ++i){ total_particles += particles_nums[i]; }
		vertex_coords->allocate(total_particles);
		vertex_colors->allocate(total_particles);
		vertex_normals->allocate(total_particles);
	}

	std::vector<int> recvcnts(mpi_size), recvdispls(mpi_size);
	for(int i = 0; i < mpi_size; ++i){
		recvcnts[i] = particles_nums[i];
		if(i != 0){ recvdispls[i] = recvdispls[i - 1] + recvcnts[i - 1]; }
	}

	/*** debug ***/
	MPI_Gatherv(
		m_coords.pointer(), m_coords.size(), MPI_FLOAT,
		(volume != NULL ? vertex_coords->pointer() : NULL),
		&recvcnts[0], &recvdispls[0], MPI_FLOAT, 0, MPI_COMM_WORLD);
	MPI_Gatherv(
		m_colors.pointer(), m_colors.size(), MPI_UNSIGNED_CHAR,
		(volume != NULL ? vertex_colors->pointer() : NULL),
		&recvcnts[0], &recvdispls[0], MPI_UNSIGNED_CHAR, 0, MPI_COMM_WORLD);
	MPI_Gatherv(
		m_normals.pointer(), m_normals.size(), MPI_FLOAT,
		(volume != NULL ? vertex_normals->pointer() : NULL),
		&recvcnts[0], &recvdispls[0], MPI_FLOAT, 0, MPI_COMM_WORLD);
	/*************/
	}

	if(volume == NULL){
		delete [] coords;
		delete [] values;
	}
	if(density_map == NULL){
		delete [] dist_density_map;
	}
	if(color_map == NULL){
		delete [] dist_color_map;
	}
	double end = get_time();
// 	printf("\n%d:GPU:data merge: %lf ms\n", mpi_rank,end - generate_particles_end);
	/*** timer ***/
	TIMER[mpi_rank][3] += end - generate_particles_end;
	/*************/
}

void DistributedMetropolisSampling::generate_particles_master_gpu(
	const kvs::UnstructuredVolumeObject *volume,
	const float *density_map,
	const unsigned char *color_map,
	size_t resolution, float min_value, float max_value,
	kvs::ValueArray<kvs::Real32> &vertex_coords,
	kvs::ValueArray<kvs::UInt8>  &vertex_colors,
	kvs::ValueArray<kvs::Real32> &vertex_normals)
{
	generate_particles_core_gpu(
		volume, density_map, color_map, resolution, min_value, max_value,
		&vertex_coords, &vertex_colors, &vertex_normals);
}

/*** debug ***/
void DistributedMetropolisSampling::generate_particles_slave_ts_gpu(
	const kvs::UnstructuredVolumeObject *volume
)
{
	generate_particles_core_gpu(volume, NULL, NULL, 0, 0.0f, 0.0f, NULL, NULL, NULL);
}
/*************/


void DistributedMetropolisSampling::generate_particles_slave_gpu()
{
	generate_particles_core_gpu(NULL, NULL, NULL, 0, 0.0f, 0.0f, NULL, NULL, NULL);
}

void DistributedMetropolisSampling::generate_particles_core_cpu(
	const kvs::UnstructuredVolumeObject *volume,
	const float *raw_density_map,
	const unsigned char *raw_color_map,
	size_t resolution, float min_value, float max_value,
	kvs::ValueArray<kvs::Real32> *vertex_coords,
	kvs::ValueArray<kvs::UInt8>  *vertex_colors,
	kvs::ValueArray<kvs::Real32> *vertex_normals)
{
	int mpi_rank = 0, mpi_size = 0;
	MPI_Comm_rank(MPI_COMM_WORLD, &mpi_rank);
	MPI_Comm_size(MPI_COMM_WORLD, &mpi_size);

	std::cout << "node " << mpi_rank << " start." << std::endl;

	size_t cellType = (volume != NULL ? volume->cellType() : 0);
	size_t veclen   = (volume != NULL ? volume->veclen() : 0);
	size_t nnodes   = (volume != NULL ? volume->nnodes() : 0);
	size_t ncells   = (volume != NULL ? volume->ncells() : 0);

	MPI_Bcast(&cellType, 1, MPI_UNSIGNED, 0, MPI_COMM_WORLD);
	MPI_Bcast(&veclen, 1, MPI_UNSIGNED, 0, MPI_COMM_WORLD);
	MPI_Bcast(&nnodes, 1, MPI_UNSIGNED, 0, MPI_COMM_WORLD);
	MPI_Bcast(&ncells, 1, MPI_UNSIGNED, 0, MPI_COMM_WORLD);
	MPI_Bcast(&resolution, 1, MPI_UNSIGNED, 0, MPI_COMM_WORLD);
	MPI_Bcast(&min_value, 1, MPI_FLOAT, 0, MPI_COMM_WORLD);
	MPI_Bcast(&max_value, 1, MPI_FLOAT, 0, MPI_COMM_WORLD);

	int cells_per_node = (static_cast<int>(ncells) + mpi_size - 1) / mpi_size;
	std::vector<int> sendcnts(mpi_size), senddispls(mpi_size);
	for(int i = 0; i < mpi_size; ++i){
		sendcnts[i] =
			(std::min<int>(cells_per_node *(i + 1), ncells) - cells_per_node * i) * cellType;
		if(i != 0){ senddispls[i] = senddispls[i - 1] + sendcnts[i - 1]; }
	}
	size_t local_ncells = sendcnts[mpi_rank] / cellType;

	kvs::ValueArray<kvs::Real32> coords;
	kvs::ValueArray<kvs::UInt32> connections(sendcnts[mpi_rank]);
	kvs::AnyValueArray values;
	float *density_map = NULL;
	kvs::ValueArray<kvs::UInt8> dist_color_map;
	if(volume != NULL){
		coords = volume->coords();
		values = volume->values();
		density_map = const_cast<float *>(raw_density_map);
		dist_color_map = kvs::ValueArray<kvs::UInt8>(raw_color_map, resolution * 3);
	}else{
		coords.allocate(nnodes * 3);
		values.allocate<kvs::Real32>(nnodes);
		density_map = new float[resolution];
		dist_color_map.allocate(resolution * 3);
	}

	MPI_Bcast(coords.pointer(), nnodes * 3, MPI_FLOAT, 0, MPI_COMM_WORLD);
	MPI_Bcast(values.pointer(), nnodes, MPI_FLOAT, 0, MPI_COMM_WORLD);
	MPI_Bcast(density_map, resolution, MPI_FLOAT, 0, MPI_COMM_WORLD);
	MPI_Bcast(dist_color_map.pointer(), resolution * 3, MPI_UNSIGNED_CHAR, 0, MPI_COMM_WORLD);

	MPI_Scatterv(
		(volume != NULL ? const_cast<kvs::UInt32 *>(volume->connections().pointer()) : NULL),
		&sendcnts[0], &senddispls[0], MPI_UNSIGNED, connections.pointer(),
		sendcnts[mpi_rank], MPI_UNSIGNED, 0, MPI_COMM_WORLD);

	kvs::UnstructuredVolumeObject *object = new kvs::UnstructuredVolumeObject(
		static_cast<kvs::VolumeObjectBase::CellType>(cellType),
		nnodes, local_ncells, veclen,
		coords, connections, values);

	kvs::CellBase<kvs::Real32> *cell = NULL;
	switch(cellType){
	case kvs::VolumeObjectBase::Tetrahedra:
		cell = new kvs::TetrahedralCell<kvs::Real32>(object);
		break;
	case kvs::VolumeObjectBase::QuadraticTetrahedra:
		cell = new kvs::QuadraticTetrahedralCell<kvs::Real32>(object);
		break;
	case kvs::VolumeObjectBase::Hexahedra:
		cell = new kvs::HexahedralCell<kvs::Real32>(object);
		break;
	case kvs::VolumeObjectBase::QuadraticHexahedra:
		cell = new kvs::QuadraticHexahedralCell<kvs::Real32>(object);
		break;
	case kvs::VolumeObjectBase::Pyramid:
		cell = new kvs::PyramidalCell<kvs::Real32>(object);
		break;
	}

	kvs::ColorMap color_map(dist_color_map, min_value, max_value);
	const size_t max_range = resolution - 1;
	const float normalize_factor = max_range / ( max_value - min_value );

	std::vector<kvs::Real32> o_coords;
	std::vector<kvs::UInt8>  o_colors;
	std::vector<kvs::Real32> o_normals;

	for(size_t index = 0; index < local_ncells; ++index){
		// Bind the cell which is indicated by 'index'.
		cell->bindCell( index );

		// Calculate a density.
		const float  average_scalar = cell->averagedScalar();
		const float average_degree_float = (average_scalar - min_value) * normalize_factor;
		size_t average_degree = 0;
		if (average_degree_float < 0) {
			average_degree = 0; // round to 0.
		} else {
			average_degree = static_cast<size_t>(average_degree_float);
		}
		const float  average_density = density_map[ average_degree ];

		// Calculate a number of particles in this cell.
		const float volume_of_cell = cell->volume();
		const float p = average_density * volume_of_cell;
		size_t nparticles_in_cell = static_cast<size_t>( p );

		if ( p - nparticles_in_cell > Generator::GetRandomNumber() ) { ++nparticles_in_cell; }
		if( nparticles_in_cell == 0 ) continue;

		// Calculate itnitial value
		/* NOTE: The gradient vector of the cell is reversed for shading on the rendering process.
		 */
		kvs::Vector3f point = cell->randomSampling();
		float         scalar = cell->scalar();
		size_t degree;
		float degree_float = (scalar - min_value) * normalize_factor;
		if (degree_float < 0) {
			degree = 0;
		} else {
			degree = static_cast<size_t>(degree_float);
		}
		degree = kvs::Math::Clamp<size_t>( degree, 0, max_range );
		float         density = density_map[ degree ];
		kvs::Vector3f g = -cell->gradient();

		kvs::Vector3f point_trial;
		float         scalar_trial;
		size_t        degree_trial;
		float         density_trial;
		kvs::Vector3f g_trial;

//		while( kvs::Math::IsZero( density ) )
		const size_t max_loop = nparticles_in_cell * 10;
		for ( size_t i = 0; i < max_loop; i++ )
		{
			point = cell->randomSampling();
			degree_float = (cell->scalar() - min_value) * normalize_factor;
			if (degree_float < 0 ) {
				degree = 0;
			} else {
				degree = static_cast<size_t>(degree_float);
			}
			degree = kvs::Math::Clamp<size_t>( degree, 0, max_range );
			density = density_map[ degree ];
			g = -cell->gradient();
			if ( !kvs::Math::IsZero( density ) ) break;
		}

		//Generate N particles
		size_t nduplications = 0; // number of duplications
		size_t counter = 0;
		while( counter < nparticles_in_cell )
		{
			//set trial position and density
			point_trial = cell->randomSampling();
			scalar_trial = cell->scalar();
			const float degree_trial_float = (scalar_trial - min_value) * normalize_factor;
			if (degree_trial_float < 0) {
				degree_trial = 0;
			} else {
				degree_trial = static_cast<size_t>(degree_trial_float);
			}
			degree_trial = kvs::Math::Clamp<size_t>( degree_trial, 0, max_range );
			density_trial = density_map[ degree_trial ];
			g_trial = -cell->gradient();

			//calculate ratio
			double ratio = density_trial / density;

			if( ratio >= 1.0 ) // accept trial point
			{
				// calculate color
				const kvs::RGBColor color( color_map.at( scalar_trial ) );

				// calculate normal
				const kvs::Vector3f normal( g_trial );

				o_coords.push_back( point_trial.x() );
				o_coords.push_back( point_trial.y() );
				o_coords.push_back( point_trial.z() );

				o_colors.push_back( color.r() );
				o_colors.push_back( color.g() );
				o_colors.push_back( color.b() );

				o_normals.push_back( normal.x() );
				o_normals.push_back( normal.y() );
				o_normals.push_back( normal.z() );

				// update point
				point = point_trial;
				degree = degree_trial;
				density = density_trial;
				g = g_trial;

				counter++;
			}
			else
			{
				if( ratio >= Generator::GetRandomNumber() ) // accept point trial
				{
					// calculate color
					const kvs::RGBColor color( color_map.at( scalar_trial ) );

					// calculate normal
					const kvs::Vector3f normal( g_trial );

					o_coords.push_back( point_trial.x() );
					o_coords.push_back( point_trial.y() );
					o_coords.push_back( point_trial.z() );

					o_colors.push_back( color.r() );
					o_colors.push_back( color.g() );
					o_colors.push_back( color.b() );

					o_normals.push_back( normal.x() );
					o_normals.push_back( normal.y() );
					o_normals.push_back( normal.z() );

					// update point
					point = point_trial;
					degree = degree_trial;
					density = density_trial;
					g = g_trial;

					counter++;
				}
				else // accept current point
				{
#ifdef DUPLICATION
					// calculate color
					const kvs::RGBColor color( color_map.scalar( degree ) );

					//calculate normal
					const kvs::Vector3f normal( g );

					o_coords.push_back( point_trial.x() );
					o_coords.push_back( point_trial.y() );
					o_coords.push_back( point_trial.z() );

					o_colors.push_back( color.r() );
					o_colors.push_back( color.g() );
					o_colors.push_back( color.b() );

					o_normals.push_back( normal.x() );
					o_normals.push_back( normal.y() );
					o_normals.push_back( normal.z() );

					counter++;
#else
					nduplications++;
					if ( nduplications > max_loop ) break;
					else continue;
#endif
				}
			}
		}
	}

	size_t particles_num = o_coords.size() / 3;
	std::vector<unsigned int> particles_nums(mpi_size);
	MPI_Allgather(
		&particles_num, 1, MPI_UNSIGNED, &particles_nums[0],
		1, MPI_UNSIGNED, MPI_COMM_WORLD);

	std::vector<int> recvcnts(mpi_size), recvdispls(mpi_size);
	for(int i = 0; i < mpi_size; ++i){
		recvcnts[i] = particles_nums[i] * 3;
		if(i != 0){ recvdispls[i] = recvdispls[i - 1] + recvcnts[i - 1]; }
	}

	if(volume != NULL){
		unsigned int total_particles = 0;
		for(int i = 0; i < mpi_size; ++i){ total_particles += particles_nums[i]; }
		vertex_coords->allocate(total_particles * 3);
		vertex_colors->allocate(total_particles * 3);
		vertex_normals->allocate(total_particles * 3);
	}

	MPI_Gatherv(
		&o_coords[0], o_coords.size(), MPI_FLOAT,
		(volume != NULL ? vertex_coords->pointer() : NULL),
		&recvcnts[0], &recvdispls[0], MPI_FLOAT, 0, MPI_COMM_WORLD);
	MPI_Gatherv(
		&o_colors[0], o_colors.size(), MPI_UNSIGNED_CHAR,
		(volume != NULL ? vertex_colors->pointer() : NULL),
		&recvcnts[0], &recvdispls[0], MPI_UNSIGNED_CHAR, 0, MPI_COMM_WORLD);
	MPI_Gatherv(
		&o_normals[0], o_normals.size(), MPI_FLOAT,
		(volume != NULL ? vertex_normals->pointer() : NULL),
		&recvcnts[0], &recvdispls[0], MPI_FLOAT, 0, MPI_COMM_WORLD);

	delete cell;
	delete object;
	if(volume == NULL){
		delete [] density_map;
	}
}

void DistributedMetropolisSampling::generate_particles_master_cpu(
	const kvs::UnstructuredVolumeObject *volume,
	const float *density_map,
	const unsigned char *color_map,
	size_t resolution, float min_value, float max_value,
	kvs::ValueArray<kvs::Real32> &vertex_coords,
	kvs::ValueArray<kvs::UInt8>  &vertex_colors,
	kvs::ValueArray<kvs::Real32> &vertex_normals)
{
	generate_particles_core_cpu(
		volume, density_map, color_map, resolution, min_value, max_value,
		&vertex_coords, &vertex_colors, &vertex_normals);
}

void DistributedMetropolisSampling::generate_particles_slave_cpu()
{
	generate_particles_core_cpu(NULL, NULL, NULL, 0, 0.0f, 0.0f, NULL, NULL, NULL);
}

} // end of namespace test
} // end of namespace kvs

#endif // ENABLE_MPI

