#ifdef ENABLE_MPI

#include "DistributedRejectionSampling.h"
#include <algorithm>
#include <kvs/CellByCellParticleGenerator>
#include <kvs/ColorMap>
#include <kvs/RGBColor>
#include <kvs/CellBase>
#include <kvs/TetrahedralCell>
#include <kvs/QuadraticTetrahedralCell>
#include <kvs/HexahedralCell>
#include <kvs/QuadraticHexahedralCell>
#include <kvs/PyramidalCell>
#include <kvs/UnstructuredVolumeObject>
#include <mpi.h>
#include "mpi_controller.h"
/*** debug ***/
#include "common.h"
#include <cuda.h>
extern int time_series;
extern int ts_file_no;
extern int integral_index;
extern int random_index;
extern int gpuMultiStream;
extern float TIMER[10][100];
/*************/

extern "C" {
size_t calc_particles_gpu_rejection(
        size_t cell_type,
        size_t nnodes,
        size_t ncells,
        const unsigned int* connections,
        const float* coords,
        const float* values,
        const float* density_map,
        size_t density_resolution,
        float min_value,
        float max_value);

void generate_particles_gpu_rejection(
        const unsigned char* color_map,
        size_t total_particles,
	float* m_coords,
	unsigned char* m_colors,
	float* m_normals);
void pageLockedMemoryAlloc(
			   float** m_coords,
			   unsigned char** m_colors,
			   float** m_normals,
			   size_t n);
void pageLockedMemoryFree(
			  float** m_coords,
			  unsigned char** m_colors,
			  float** m_normals);
void pageableMemoryAlloc(
			   float** m_coords,
			   unsigned char** m_colors,
			   float** m_normals,
			   size_t n);
void pageableMemoryFree(
			  float** m_coords,
			  unsigned char** m_colors,
			  float** m_normals);
}

namespace Generator = kvs::CellByCellParticleGenerator;

namespace kvs {
namespace test {

void DistributedRejectionSampling::generate_particles_core_gpu(
	const kvs::UnstructuredVolumeObject *volume,
	const float *density_map,
	const unsigned char *color_map,
	size_t resolution, float min_value, float max_value,
	kvs::ValueArray<kvs::Real32> *vertex_coords,
	kvs::ValueArray<kvs::UInt8>  *vertex_colors,
	kvs::ValueArray<kvs::Real32> *vertex_normals)
{
	int mpi_rank = 0, mpi_size = 0;
	MPI_Comm_rank(MPI_COMM_WORLD, &mpi_rank);
	MPI_Comm_size(MPI_COMM_WORLD, &mpi_size);

	std::cout << "node " << mpi_rank << " start." << std::endl;

	size_t cellType = (volume != NULL ? volume->cellType() : 0);
	size_t nnodes   = (volume != NULL ? volume->nnodes() : 0);
	size_t ncells   = (volume != NULL ? volume->ncells() : 0);

 	/*** debug ***/
	double start = get_time();
 	MPI_Bcast(&time_series, 1, MPI_INTEGER, 0, MPI_COMM_WORLD);
  	std::cout << mpi_rank << ":time_series = " << time_series << std::endl;
	if(time_series==0){
	MPI_Bcast(&cellType, 1, MPI_UNSIGNED, 0, MPI_COMM_WORLD);
	MPI_Bcast(&nnodes, 1, MPI_UNSIGNED, 0, MPI_COMM_WORLD);
	MPI_Bcast(&ncells, 1, MPI_UNSIGNED, 0, MPI_COMM_WORLD);
	}
	MPI_Bcast(&resolution, 1, MPI_UNSIGNED, 0, MPI_COMM_WORLD);
	MPI_Bcast(&min_value, 1, MPI_FLOAT, 0, MPI_COMM_WORLD);
	MPI_Bcast(&max_value, 1, MPI_FLOAT, 0, MPI_COMM_WORLD);
 	/*************/

	kvs::Real32 *coords = NULL, *values = NULL;
	float *dist_density_map = NULL;
	unsigned char *dist_color_map = NULL;
	if(volume != NULL){
		coords = const_cast<kvs::Real32 *>(volume->coords().pointer());
		values = reinterpret_cast<kvs::Real32 *>(const_cast<void *>(volume->values().pointer()));
	}else{
		coords = new float[nnodes * 3];
		values = new float[nnodes];
	}
	if(density_map != NULL){
		dist_density_map = const_cast<float *>(density_map);
	}else{
		dist_density_map = new float[resolution];
	}
	if(color_map != NULL){
		dist_color_map = const_cast<unsigned char *>(color_map);
	}else{
		dist_color_map = new unsigned char[resolution * 3];
	}

	double generate_particles_start;
	size_t particles_num;

	MPI_Bcast(dist_density_map, resolution, MPI_FLOAT, 0, MPI_COMM_WORLD);
	MPI_Bcast(dist_color_map, resolution * 3, MPI_UNSIGNED_CHAR, 0, MPI_COMM_WORLD);

	if(time_series==0){
	MPI_Bcast(coords, nnodes * 3, MPI_FLOAT, 0, MPI_COMM_WORLD);
	MPI_Bcast(values, nnodes, MPI_FLOAT, 0, MPI_COMM_WORLD);

	int cells_per_node = (static_cast<int>(ncells) + mpi_size - 1) / mpi_size;
	std::vector<int> sendcnts(mpi_size), senddispls(mpi_size);
	for(int i = 0; i < mpi_size; ++i){
		sendcnts[i] =
			(std::min<int>(cells_per_node * (i + 1), ncells) - cells_per_node * i) * cellType;
		if(i != 0){ senddispls[i] = senddispls[i - 1] + sendcnts[i - 1]; }
	}
	std::vector<kvs::UInt32> connections(sendcnts[mpi_rank]);

	MPI_Scatterv(
		(volume != NULL ? const_cast<kvs::UInt32 *>(volume->connections().pointer()) : NULL),
		&sendcnts[0], &senddispls[0], MPI_UNSIGNED, &connections[0], connections.size(),
		MPI_UNSIGNED, 0, MPI_COMM_WORLD);

	generate_particles_start = get_time();
// 	printf("\n%d:GPU:data transfer: %lf ms\n", mpi_rank,generate_particles_start-start);
	/*** timer ***/
	TIMER[mpi_rank][1] += generate_particles_start-start;
	/*************/
	particles_num = calc_particles_gpu_rejection(
		cellType, nnodes, sendcnts[mpi_rank] / cellType, &connections[0],
		&coords[0], &values[0], dist_density_map, resolution, min_value, max_value);

	}else{ // time_series==1
	generate_particles_start = get_time();
// 	printf("\n%d:GPU:data transfer: %lf ms\n", mpi_rank,generate_particles_start-start);
	/*** timer ***/
	TIMER[mpi_rank][1] += generate_particles_start-start;
	/*************/
	particles_num = calc_particles_gpu_rejection(
	        cellType, nnodes, ncells, volume->connections().pointer(),
		&coords[0], &values[0], dist_density_map, resolution, min_value, max_value);
	}

   	size_t         o_size  = particles_num*3;
     	float*         o_coords;
     	unsigned char* o_colors;
     	float*         o_normals;
     	if(gpuMultiStream==0){
    	  pageableMemoryAlloc(&o_coords,&o_colors,&o_normals,o_size);
     	}else{
    	  pageLockedMemoryAlloc(&o_coords,&o_colors,&o_normals,o_size);
     	}
// 	kvs::ValueArray<kvs::Real32> o_coords(o_size);
// 	kvs::ValueArray<kvs::UInt8> o_colors(o_size);
// 	kvs::ValueArray<kvs::Real32> o_normals(o_size);
  	generate_particles_gpu_rejection
 	  (
 	   dist_color_map, particles_num,
  	   &o_coords[0], &o_colors[0], &o_normals[0]
 	   );

	double generate_particles_end = get_time();
// 	printf("\n%d:GPU:generate_particles: %lf ms\n", mpi_rank,generate_particles_end - generate_particles_start);
	/*** timer ***/
	TIMER[mpi_rank][2] += generate_particles_end - generate_particles_start;
	/*************/

	if(time_series==0){
  	std::vector<unsigned int> particles_nums(mpi_size);
  	MPI_Allgather(
  		&particles_num, 1, MPI_UNSIGNED, &particles_nums[0],
  		1, MPI_UNSIGNED, MPI_COMM_WORLD);

  	if(volume != NULL){
	        unsigned int total_particles = 0;
  		for(int i = 0; i < mpi_size; ++i){ total_particles += particles_nums[i]; }
     		vertex_coords->allocate(total_particles * 3);
    		vertex_colors->allocate(total_particles * 3);
    		vertex_normals->allocate(total_particles * 3);
  	}

  	std::vector<int> recvcnts(mpi_size), recvdispls(mpi_size);
  	for(int i = 0; i < mpi_size; ++i){
  		recvcnts[i] = particles_nums[i] * 3;
  		if(i != 0){ recvdispls[i] = recvdispls[i - 1] + recvcnts[i - 1]; }
  	}


	/*** debug ***/
 	MPI_Gatherv(
  		    &o_coords[0], o_size, MPI_FLOAT,
  		    (volume != NULL ? vertex_coords->pointer() : NULL),
  		    &recvcnts[0], &recvdispls[0], MPI_FLOAT, 0, MPI_COMM_WORLD);
   	MPI_Gatherv(
 		    &o_colors[0], o_size, MPI_UNSIGNED_CHAR,
 		    (volume != NULL ? vertex_colors->pointer() : NULL),
 		    &recvcnts[0], &recvdispls[0], MPI_UNSIGNED_CHAR, 0, MPI_COMM_WORLD);
   	MPI_Gatherv(
 		    &o_normals[0], o_size, MPI_FLOAT,
 		    (volume != NULL ? vertex_normals->pointer() : NULL),
 		    &recvcnts[0], &recvdispls[0], MPI_FLOAT, 0, MPI_COMM_WORLD);
	}
	/*************/

    	if(gpuMultiStream==0){
    	  pageableMemoryFree(&o_coords,  &o_colors,  &o_normals);
    	}else{
    	  pageLockedMemoryFree(&o_coords,  &o_colors,  &o_normals);
    	}

	if(volume == NULL){
		delete [] coords;
		delete [] values;
	}
	if(density_map == NULL){
		delete [] dist_density_map;
	}
	if(color_map == NULL){
		delete [] dist_color_map;
	}
	double end = get_time();
// 	printf("\n%d:GPU:data merge: %lf ms\n", mpi_rank,end - generate_particles_end);
	/*** timer ***/
	TIMER[mpi_rank][3] += end - generate_particles_end;
	/*************/
}

void DistributedRejectionSampling::generate_particles_master_gpu(
	const kvs::UnstructuredVolumeObject *volume,
	const float *density_map,
	const unsigned char *color_map,
	size_t resolution, float min_value, float max_value,
	kvs::ValueArray<kvs::Real32> &vertex_coords,
	kvs::ValueArray<kvs::UInt8>  &vertex_colors,
	kvs::ValueArray<kvs::Real32> &vertex_normals)
{
	generate_particles_core_gpu(
		volume, density_map, color_map, resolution, min_value, max_value,
		&vertex_coords, &vertex_colors, &vertex_normals);
}

void DistributedRejectionSampling::generate_particles_slave_gpu()
{
	generate_particles_core_gpu(NULL, NULL, NULL, 0, 0.0f, 0.0f, NULL, NULL, NULL);
}

/*** debug ***/
void DistributedRejectionSampling::generate_particles_slave_ts_gpu(
        const kvs::UnstructuredVolumeObject *volume
)
{
	generate_particles_core_gpu(volume, NULL, NULL, 0, 0.0f, 0.0f, NULL, NULL, NULL);
}
/*************/

/*************/
/*** debug ***/
/*************/
// void DistributedRejectionSampling::generate_particles_test_gpu(
// 	const kvs::UnstructuredVolumeObject *volume,
// 	const float *raw_density_map,
// 	const unsigned char *raw_color_map,
// 	size_t resolution, float min_value, float max_value,
// 	kvs::ValueArray<kvs::Real32> *vertex_coords,
// 	kvs::ValueArray<kvs::UInt8>  *vertex_colors,
// 	kvs::ValueArray<kvs::Real32> *vertex_normals)
// {
// 	int mpi_rank = 0, mpi_size = 0;
// 	MPI_Comm_rank(MPI_COMM_WORLD, &mpi_rank);
// 	MPI_Comm_size(MPI_COMM_WORLD, &mpi_size);

// 	std::cout << "node " << mpi_rank << " start." << std::endl;

// 	size_t cellType = (volume != NULL ? volume->cellType() : 0);
// 	size_t veclen   = (volume != NULL ? volume->veclen() : 0);
// 	size_t nnodes   = (volume != NULL ? volume->nnodes() : 0);
// 	size_t ncells   = (volume != NULL ? volume->ncells() : 0);

//  	/*** debug ***/
// 	double start = get_time();
//  	/*************/
//  	MPI_Bcast(&time_series, 1, MPI_INTEGER, 0, MPI_COMM_WORLD);
// 	MPI_Bcast(&cellType, 1, MPI_UNSIGNED, 0, MPI_COMM_WORLD);
// 	MPI_Bcast(&veclen, 1, MPI_UNSIGNED, 0, MPI_COMM_WORLD);
// 	MPI_Bcast(&nnodes, 1, MPI_UNSIGNED, 0, MPI_COMM_WORLD);
// 	MPI_Bcast(&ncells, 1, MPI_UNSIGNED, 0, MPI_COMM_WORLD);
// 	MPI_Bcast(&resolution, 1, MPI_UNSIGNED, 0, MPI_COMM_WORLD);
// 	MPI_Bcast(&min_value, 1, MPI_FLOAT, 0, MPI_COMM_WORLD);
// 	MPI_Bcast(&max_value, 1, MPI_FLOAT, 0, MPI_COMM_WORLD);

// 	int cells_per_node = (static_cast<int>(ncells) + mpi_size - 1) / mpi_size;
// 	std::vector<int> sendcnts(mpi_size), senddispls(mpi_size);
// 	for(int i = 0; i < mpi_size; ++i){
// 		sendcnts[i] =
// 			(std::min<int>(cells_per_node * (i + 1), ncells) - cells_per_node * i) * cellType;
// 		if(i != 0){ senddispls[i] = senddispls[i - 1] + sendcnts[i - 1]; }
// 	}
// 	size_t local_ncells = sendcnts[mpi_rank] / cellType;

// 	kvs::ValueArray<kvs::Real32> coords;
// 	kvs::ValueArray<kvs::UInt32> connections(sendcnts[mpi_rank]);
// 	kvs::AnyValueArray values;
// 	float *density_map = NULL;
// 	kvs::ValueArray<kvs::UInt8> dist_color_map;
//  	/*** debug ***/
// 	if(volume != NULL){
// 		coords = volume->coords();
// 		values = volume->values();
// 	}else{
// 		coords.allocate(nnodes * 3);
// 		values.allocate<kvs::Real32>(nnodes);
// 	}
// 	if(raw_density_map != NULL){
// 		density_map = const_cast<float *>(raw_density_map);
// 	}else{
// 		density_map = new float[resolution];
// 	}
// 	if(raw_color_map != NULL){
// 		dist_color_map = kvs::ValueArray<kvs::UInt8>(raw_color_map, resolution * 3);
// 	}else{
// 		dist_color_map.allocate(resolution * 3);
// 	}
//  	/*************/

// 	MPI_Bcast(density_map, resolution, MPI_FLOAT, 0, MPI_COMM_WORLD);
// 	MPI_Bcast(dist_color_map.pointer(), resolution * 3, MPI_UNSIGNED_CHAR, 0, MPI_COMM_WORLD);
// 	MPI_Bcast(coords.pointer(), nnodes * 3, MPI_FLOAT, 0, MPI_COMM_WORLD);
// 	MPI_Bcast(values.pointer(), nnodes, MPI_FLOAT, 0, MPI_COMM_WORLD);

// 	MPI_Scatterv(
// 		(volume != NULL ? const_cast<kvs::UInt32 *>(volume->connections().pointer()) : NULL),
// 		&sendcnts[0], &senddispls[0], MPI_UNSIGNED, connections.pointer(),
// 		sendcnts[mpi_rank], MPI_UNSIGNED, 0, MPI_COMM_WORLD);

// 	kvs::UnstructuredVolumeObject *object = new kvs::UnstructuredVolumeObject(
// 		static_cast<kvs::VolumeObjectBase::CellType>(cellType),
// 		nnodes, local_ncells, veclen,
// 		coords, connections, values);

// 	kvs::CellBase<kvs::Real32> *cell = NULL;
// 	switch(cellType){
// 	case kvs::VolumeObjectBase::Tetrahedra:
// 		cell = new kvs::TetrahedralCell<Real32>(object);
// 		break;
// 	case kvs::VolumeObjectBase::QuadraticTetrahedra:
// 		cell = new kvs::QuadraticTetrahedralCell<kvs::Real32>(object);
// 		break;
// 	case kvs::VolumeObjectBase::Hexahedra:
// 		cell = new kvs::HexahedralCell<kvs::Real32>(object);
// 		break;
// 	case kvs::VolumeObjectBase::QuadraticHexahedra:
// 		cell = new kvs::QuadraticHexahedralCell<kvs::Real32>(object);
// 		break;
// 	case kvs::VolumeObjectBase::Pyramid:
// 		cell = new kvs::PyramidalCell<kvs::Real32>(object);
// 		break;
// 	}

// 	kvs::ColorMap color_map(dist_color_map, min_value, max_value);

// 	std::vector<kvs::Real32> o_coords;
// 	std::vector<kvs::UInt8>  o_colors;
// 	std::vector<kvs::Real32> o_normals;

// 	/*** gpu ***/
// 	for(size_t index = 0; index < local_ncells; ++index){
// 		// Bind the cell which is indicated by 'index'.
// 		cell->bindCell( index );

// 		// Calculate a number of particles in this cell.
// 		const float averaged_scalar = cell->averagedScalar();
// 		float density = calculate_density(
// 			averaged_scalar, min_value, max_value, resolution, density_map );
// 		if (isnan(density)) {
// 			density = 0;
// 		}
// 		const size_t nparticles = calculate_number_of_particles( density, cell->volume() );

// 		const kvs::Real32* S = cell->scalars();
// 		const size_t nnodes = object->cellType();
// 		kvs::Real32 S_min = S[0];
// 		kvs::Real32 S_max = S[0];
// 		for ( size_t i = 1; i < nnodes; i++ )
// 		{
// 			S_min = kvs::Math::Min( S_min, S[i] );
// 			S_max = kvs::Math::Max( S_max, S[i] );
// 		}
// 		const float s_min = static_cast<float>( S_min );
// 		const float s_max = static_cast<float>( S_max );
// 		const float p_max = calculate_maximum_density(
// 			s_min, s_max, min_value, max_value, resolution, density_map ) / nparticles;

// 		// Generate a set of particles in this cell represented by v0,...,v3 and s0,...,s3.
// 		size_t count = 0;
// 		while ( count < nparticles )
// 		{
// 			const kvs::Vector3f coord = cell->randomSampling();
// 			const float scalar = cell->scalar();
// 			const float density = calculate_density(
// 				scalar, min_value, max_value, resolution, density_map );

// 			const float p = density / nparticles;
// 			const float R = Generator::GetRandomNumber();
// 			if ( p > p_max * R )
// 			{
// 				// Calculate a color.
// 				const kvs::RGBColor color( color_map.at( scalar ) );

// 				// Calculate a normal.
// 				const Vector3f normal( cell->gradient() );

// 				// set coord, color, and normal to point object( this ).
// 				o_coords.push_back( coord.x() );
// 				o_coords.push_back( coord.y() );
// 				o_coords.push_back( coord.z() );

// 				o_colors.push_back( color.r() );
// 				o_colors.push_back( color.g() );
// 				o_colors.push_back( color.b() );

// 				o_normals.push_back( normal.x() );
// 				o_normals.push_back( normal.y() );
// 				o_normals.push_back( normal.z() );

// 				count++;
// 			}
// 		}
// 	}
// 	/***********/

// 	size_t particles_num = o_coords.size() / 3;
// 	std::vector<unsigned int> particles_nums(mpi_size);
// 	MPI_Allgather(
// 		&particles_num, 1, MPI_UNSIGNED, &particles_nums[0],
// 		1, MPI_UNSIGNED, MPI_COMM_WORLD);

// 	std::vector<int> recvcnts(mpi_size), recvdispls(mpi_size);
// 	for(int i = 0; i < mpi_size; ++i){
// 		recvcnts[i] = particles_nums[i] * 3;
// 		if(i != 0){ recvdispls[i] = recvdispls[i - 1] + recvcnts[i - 1]; }
// 	}

// 	if(volume != NULL){
// 		unsigned int total_particles = 0;
// 		for(int i = 0; i < mpi_size; ++i){ total_particles += particles_nums[i]; }
// 		vertex_coords->allocate(total_particles * 3);
// 		vertex_colors->allocate(total_particles * 3);
// 		vertex_normals->allocate(total_particles * 3);
// 	}

// 	MPI_Gatherv(
// 		&o_coords[0], o_coords.size(), MPI_FLOAT,
// 		(volume != NULL ? vertex_coords->pointer() : NULL),
// 		&recvcnts[0], &recvdispls[0], MPI_FLOAT, 0, MPI_COMM_WORLD);
// 	MPI_Gatherv(
// 		&o_colors[0], o_colors.size(), MPI_UNSIGNED_CHAR,
// 		(volume != NULL ? vertex_colors->pointer() : NULL),
// 		&recvcnts[0], &recvdispls[0], MPI_UNSIGNED_CHAR, 0, MPI_COMM_WORLD);
// 	MPI_Gatherv(
// 		&o_normals[0], o_normals.size(), MPI_FLOAT,
// 		(volume != NULL ? vertex_normals->pointer() : NULL),
// 		&recvcnts[0], &recvdispls[0], MPI_FLOAT, 0, MPI_COMM_WORLD);

// 	delete cell;
// 	delete object;
// 	if(volume == NULL){
// 		delete [] density_map;
// 	}
// }
/*************/
/*************/
/*************/

void DistributedRejectionSampling::generate_particles_core_cpu(
	const kvs::UnstructuredVolumeObject *volume,
	const float *raw_density_map,
	const unsigned char *raw_color_map,
	size_t resolution, float min_value, float max_value,
	kvs::ValueArray<kvs::Real32> *vertex_coords,
	kvs::ValueArray<kvs::UInt8>  *vertex_colors,
	kvs::ValueArray<kvs::Real32> *vertex_normals)
{
	int mpi_rank = 0, mpi_size = 0;
	MPI_Comm_rank(MPI_COMM_WORLD, &mpi_rank);
	MPI_Comm_size(MPI_COMM_WORLD, &mpi_size);

	std::cout << "node " << mpi_rank << " start." << std::endl;

	size_t cellType = (volume != NULL ? volume->cellType() : 0);
	size_t veclen   = (volume != NULL ? volume->veclen() : 0);
	size_t nnodes   = (volume != NULL ? volume->nnodes() : 0);
	size_t ncells   = (volume != NULL ? volume->ncells() : 0);

 	/*** debug ***/
	double start = get_time();
 	MPI_Bcast(&time_series, 1, MPI_INTEGER, 0, MPI_COMM_WORLD);
  	std::cout << mpi_rank << ":time_series = " << time_series << std::endl;
	if(time_series==0){
	MPI_Bcast(&cellType, 1, MPI_UNSIGNED, 0, MPI_COMM_WORLD);
	MPI_Bcast(&veclen, 1, MPI_UNSIGNED, 0, MPI_COMM_WORLD);
	MPI_Bcast(&nnodes, 1, MPI_UNSIGNED, 0, MPI_COMM_WORLD);
	MPI_Bcast(&ncells, 1, MPI_UNSIGNED, 0, MPI_COMM_WORLD);
	}
	MPI_Bcast(&resolution, 1, MPI_UNSIGNED, 0, MPI_COMM_WORLD);
	MPI_Bcast(&min_value, 1, MPI_FLOAT, 0, MPI_COMM_WORLD);
	MPI_Bcast(&max_value, 1, MPI_FLOAT, 0, MPI_COMM_WORLD);
 	/*************/

	int cells_per_node = (static_cast<int>(ncells) + mpi_size - 1) / mpi_size;
	std::vector<int> sendcnts(mpi_size), senddispls(mpi_size);
	for(int i = 0; i < mpi_size; ++i){
		sendcnts[i] =
			(std::min<int>(cells_per_node * (i + 1), ncells) - cells_per_node * i) * cellType;
		if(i != 0){ senddispls[i] = senddispls[i - 1] + sendcnts[i - 1]; }
	}
	size_t local_ncells = sendcnts[mpi_rank] / cellType;

	kvs::ValueArray<kvs::Real32> coords;
	kvs::ValueArray<kvs::UInt32> connections(sendcnts[mpi_rank]);
	kvs::AnyValueArray values;
	float *density_map = NULL;
	kvs::ValueArray<kvs::UInt8> dist_color_map;
	if(volume != NULL){
		coords = volume->coords();
		values = volume->values();
	}else{
		coords.allocate(nnodes * 3);
		values.allocate<kvs::Real32>(nnodes);
	}
	if(raw_density_map != NULL){
		density_map = const_cast<float *>(raw_density_map);
	}else{
		density_map = new float[resolution];
	}
	if(raw_color_map != NULL){
		dist_color_map = kvs::ValueArray<kvs::UInt8>(raw_color_map, resolution * 3);
	}else{
		dist_color_map.allocate(resolution * 3);
	}

	MPI_Bcast(density_map, resolution, MPI_FLOAT, 0, MPI_COMM_WORLD);
	MPI_Bcast(dist_color_map.pointer(), resolution * 3, MPI_UNSIGNED_CHAR, 0, MPI_COMM_WORLD);
	MPI_Bcast(coords.pointer(), nnodes * 3, MPI_FLOAT, 0, MPI_COMM_WORLD);
	MPI_Bcast(values.pointer(), nnodes, MPI_FLOAT, 0, MPI_COMM_WORLD);

	MPI_Scatterv(
		(volume != NULL ? const_cast<kvs::UInt32 *>(volume->connections().pointer()) : NULL),
		&sendcnts[0], &senddispls[0], MPI_UNSIGNED, connections.pointer(),
		sendcnts[mpi_rank], MPI_UNSIGNED, 0, MPI_COMM_WORLD);

	kvs::UnstructuredVolumeObject *object = new kvs::UnstructuredVolumeObject(
		static_cast<kvs::VolumeObjectBase::CellType>(cellType),
		nnodes, local_ncells, veclen,
		coords, connections, values);

	kvs::CellBase<kvs::Real32> *cell = NULL;
	switch(cellType){
	case kvs::VolumeObjectBase::Tetrahedra:
		cell = new kvs::TetrahedralCell<Real32>(object);
		break;
	case kvs::VolumeObjectBase::QuadraticTetrahedra:
		cell = new kvs::QuadraticTetrahedralCell<kvs::Real32>(object);
		break;
	case kvs::VolumeObjectBase::Hexahedra:
		cell = new kvs::HexahedralCell<kvs::Real32>(object);
		break;
	case kvs::VolumeObjectBase::QuadraticHexahedra:
		cell = new kvs::QuadraticHexahedralCell<kvs::Real32>(object);
		break;
	case kvs::VolumeObjectBase::Pyramid:
		cell = new kvs::PyramidalCell<kvs::Real32>(object);
		break;
	}

	kvs::ColorMap color_map(dist_color_map, min_value, max_value);

	std::vector<kvs::Real32> o_coords;
	std::vector<kvs::UInt8>  o_colors;
	std::vector<kvs::Real32> o_normals;

	for(size_t index = 0; index < local_ncells; ++index){
		// Bind the cell which is indicated by 'index'.
		cell->bindCell( index );

		// Calculate a number of particles in this cell.
		const float averaged_scalar = cell->averagedScalar();
		float density = calculate_density(
			averaged_scalar, min_value, max_value, resolution, density_map );
		if (isnan(density)) {
			density = 0;
		}
		const size_t nparticles = calculate_number_of_particles( density, cell->volume() );

		const kvs::Real32* S = cell->scalars();
		const size_t nnodes = object->cellType();
		kvs::Real32 S_min = S[0];
		kvs::Real32 S_max = S[0];
		for ( size_t i = 1; i < nnodes; i++ )
		{
			S_min = kvs::Math::Min( S_min, S[i] );
			S_max = kvs::Math::Max( S_max, S[i] );
		}
		const float s_min = static_cast<float>( S_min );
		const float s_max = static_cast<float>( S_max );
		const float p_max = calculate_maximum_density(
			s_min, s_max, min_value, max_value, resolution, density_map ) / nparticles;

		// Generate a set of particles in this cell represented by v0,...,v3 and s0,...,s3.
		size_t count = 0;
		while ( count < nparticles )
		{
			const kvs::Vector3f coord = cell->randomSampling();
			const float scalar = cell->scalar();
			const float density = calculate_density(
				scalar, min_value, max_value, resolution, density_map );

			const float p = density / nparticles;
			const float R = Generator::GetRandomNumber();
			if ( p > p_max * R )
			{
				// Calculate a color.
				const kvs::RGBColor color( color_map.at( scalar ) );

				// Calculate a normal.
				const Vector3f normal( cell->gradient() );

				// set coord, color, and normal to point object( this ).
				o_coords.push_back( coord.x() );
				o_coords.push_back( coord.y() );
				o_coords.push_back( coord.z() );

				o_colors.push_back( color.r() );
				o_colors.push_back( color.g() );
				o_colors.push_back( color.b() );

				o_normals.push_back( normal.x() );
				o_normals.push_back( normal.y() );
				o_normals.push_back( normal.z() );

				count++;
			}
		}
	}

	size_t particles_num = o_coords.size() / 3;
	std::vector<unsigned int> particles_nums(mpi_size);
	MPI_Allgather(
		&particles_num, 1, MPI_UNSIGNED, &particles_nums[0],
		1, MPI_UNSIGNED, MPI_COMM_WORLD);

	std::vector<int> recvcnts(mpi_size), recvdispls(mpi_size);
	for(int i = 0; i < mpi_size; ++i){
		recvcnts[i] = particles_nums[i] * 3;
		if(i != 0){ recvdispls[i] = recvdispls[i - 1] + recvcnts[i - 1]; }
	}

	if(volume != NULL){
		unsigned int total_particles = 0;
		for(int i = 0; i < mpi_size; ++i){ total_particles += particles_nums[i]; }
		vertex_coords->allocate(total_particles * 3);
		vertex_colors->allocate(total_particles * 3);
		vertex_normals->allocate(total_particles * 3);
	}

	MPI_Gatherv(
		&o_coords[0], o_coords.size(), MPI_FLOAT,
		(volume != NULL ? vertex_coords->pointer() : NULL),
		&recvcnts[0], &recvdispls[0], MPI_FLOAT, 0, MPI_COMM_WORLD);
	MPI_Gatherv(
		&o_colors[0], o_colors.size(), MPI_UNSIGNED_CHAR,
		(volume != NULL ? vertex_colors->pointer() : NULL),
		&recvcnts[0], &recvdispls[0], MPI_UNSIGNED_CHAR, 0, MPI_COMM_WORLD);
	MPI_Gatherv(
		&o_normals[0], o_normals.size(), MPI_FLOAT,
		(volume != NULL ? vertex_normals->pointer() : NULL),
		&recvcnts[0], &recvdispls[0], MPI_FLOAT, 0, MPI_COMM_WORLD);

	delete cell;
	delete object;
	if(volume == NULL){
		delete [] density_map;
	}
}

float DistributedRejectionSampling::calculate_density(
	const float scalar, const float min_value, const float max_value,
	const size_t resolution, const float *density_map)
{
	const float max_range = static_cast<float>( resolution - 1 );
	const float normalize_factor = max_range / ( max_value - min_value );
	const float normalized_scalar = ( scalar - min_value ) * normalize_factor;
	size_t index0 = 0;
	if (normalized_scalar < 0) {
		index0 = 0; // round to 0.
	} else {
		index0 = static_cast<size_t>(normalized_scalar);
	}
	size_t index1 = index0 + 1;
	index1 = kvs::Math::Clamp<size_t>(index1, 0, (size_t)max_range);
	const float scalar_offset = normalized_scalar - index0;

	if ( index0 == ( resolution - 1 ) )
	{
		return density_map[index0];
	}
	else
	{
		const float rho0 = density_map[ index0 ];
		const float rho1 = density_map[ index1 ];
		const float interpolated_density = ( rho1 - rho0 ) * scalar_offset + rho0;

		return interpolated_density;
	}
}

size_t DistributedRejectionSampling::calculate_number_of_particles(
	const float density, const float volume_of_cell)
{
	const float N = density * volume_of_cell;
	const float R = Generator::GetRandomNumber();

	size_t n = static_cast<size_t>(N);
	if ( N - n > R ) { ++n; }

	return( n );
}

float DistributedRejectionSampling::calculate_maximum_density(
	const float scalar0, const float scalar1,
	const float min_value, const float max_value, const size_t resolution, 
	const float *density_map)
{
	if ( scalar0 > scalar1 )
	{
		kvsMessageError("undefined use of calculate_maximum_density.");
		return 0.0f;
	}

	const float max_range = static_cast<float>( resolution - 1 );
	const float normalize_factor = max_range / ( max_value - min_value );
	const float index0_float = (scalar0 - min_value) * normalize_factor;
	size_t index0 = 0;
	if (index0_float < 0) {
		index0 = 0; // round to 0.
	} else {
		index0 = static_cast<size_t>(index0_float);
	}
	index0 += 1;
	index0 = kvs::Math::Clamp<size_t>(index0, 0, (size_t)max_range);

	const float index1_float = (scalar1 - min_value) * normalize_factor;
	size_t index1 = 0;
	if (index1_float < 0) {
		index1 = 0; // round to 0.
	} else {
		index1 = static_cast<size_t>(index1_float);
	}

	float maximum_density = density_map[ index0 ];

	for ( size_t i = index0 + 1; i <= index1; i++ )
	{
		maximum_density = density_map[ i ] > maximum_density ? density_map[ i ] : maximum_density;
	}

	const float density0 = calculate_density(
		scalar0, min_value, max_value, resolution, density_map );
	maximum_density = density0 > maximum_density ? density0 : maximum_density;

	const float density1 = calculate_density(
		scalar1, min_value, max_value, resolution, density_map );
	maximum_density = density1 > maximum_density ? density1 : maximum_density;

	return maximum_density;
}

void DistributedRejectionSampling::generate_particles_master_cpu(
	const kvs::UnstructuredVolumeObject *volume,
	const float *density_map,
	const unsigned char *color_map,
	size_t resolution, float min_value, float max_value,
	kvs::ValueArray<kvs::Real32> &vertex_coords,
	kvs::ValueArray<kvs::UInt8>  &vertex_colors,
	kvs::ValueArray<kvs::Real32> &vertex_normals)
{
	generate_particles_core_cpu(
		volume, density_map, color_map, resolution, min_value, max_value,
		&vertex_coords, &vertex_colors, &vertex_normals);
}

/*** debug ***/
// void DistributedRejectionSampling::generate_particles_slave_ts_cpu(
//         const kvs::UnstructuredVolumeObject *volume
// )
// {
// 	generate_particles_core_gpu(volume, NULL, NULL, 0, 0.0f, 0.0f, NULL, NULL, NULL);
// }
/*************/

void DistributedRejectionSampling::generate_particles_slave_cpu()
{
	generate_particles_core_cpu(NULL, NULL, NULL, 0, 0.0f, 0.0f, NULL, NULL, NULL);
}

} // end of namespace test
} // end of namespace kvs

#endif // ENABLE_MPI

