//#ifdef _OPENMP
#include <omp.h>
//#endif

#include <kvs/PointObject>

#include <vector>
#include <iostream>
#include <cassert>

#include "common.h"

#include "PointShuffle.h"
#include "Profiler.h"

extern void shuffleWithCUDA(const int howManyParticles, const int repeatLevel, const std::vector<int>& indices, std::vector<int>& shuffledIndices, const std::vector<float>& positions, std::vector<float>& shuffledPositions, const std::vector<float>& normals, std::vector<float>& shuffledNormals, const std::vector<unsigned char>& colors, std::vector<unsigned char>& shuffledColors);

using namespace kvs;

void PointShuffle::shuffle(const int howManyParticles, const int repeatLevel){
    for( int i = 0; i < howManyParticles; ++i ) {
	int index = indices[i];
	int newIndex;
	if( (index % repeatLevel) < (howManyParticles % repeatLevel) ) {
		newIndex = (index % repeatLevel) * ( howManyParticles / repeatLevel + 1 ) + index / repeatLevel;
	}
	else {
		newIndex = (index % repeatLevel) * ( howManyParticles / repeatLevel ) + index / repeatLevel + howManyParticles % repeatLevel;
	}
	shuffledIndices[index] = newIndex;
	shuffledPositions[newIndex*3] = positions[index*3];
	shuffledPositions[newIndex*3+1] = positions[index*3+1];
	shuffledPositions[newIndex*3+2] = positions[index*3+2];
        shuffledNormals[newIndex*3] = normals[index*3];
	shuffledNormals[newIndex*3+1] = normals[index*3+1];
	shuffledNormals[newIndex*3+2] = normals[index*3+2];
	shuffledColors[newIndex*3] = colors[index*3];
	shuffledColors[newIndex*3+1] = colors[index*3+1];
	shuffledColors[newIndex*3+2] = colors[index*3+2];
}
}

void PointShuffle::shuffleWithOpenMP(const int howManyParticles, const int repeatLevel){
#pragma omp parallel for
    for( int i = 0; i < howManyParticles; ++i ) {
	int index = indices[i];
	int newIndex;
	if( (index % repeatLevel) < (howManyParticles % repeatLevel) ) {
		newIndex = (index % repeatLevel) * ( howManyParticles / repeatLevel + 1 ) + index / repeatLevel;
	}
	else {
		newIndex = (index % repeatLevel) * ( howManyParticles / repeatLevel ) + index / repeatLevel + howManyParticles % repeatLevel;
	}
	shuffledIndices[index] = newIndex;
	shuffledPositions[newIndex*3] = positions[index*3];
	shuffledPositions[newIndex*3+1] = positions[index*3+1];
	shuffledPositions[newIndex*3+2] = positions[index*3+2];
        shuffledNormals[newIndex*3] = normals[index*3];
	shuffledNormals[newIndex*3+1] = normals[index*3+1];
	shuffledNormals[newIndex*3+2] = normals[index*3+2];
	shuffledColors[newIndex*3] = colors[index*3];
	shuffledColors[newIndex*3+1] = colors[index*3+1];
	shuffledColors[newIndex*3+2] = colors[index*3+2];
    }
}

void PointShuffle::setPointObject(PointObject* object)
{
    this->object = object;
    const int howManyParticles = object->coords().size() / 3;
    indices.resize(howManyParticles);
    positions.resize(howManyParticles*3);
    normals.resize(howManyParticles*3);
    colors.resize(howManyParticles*3);

    for( size_t i = 0; i < indices.size(); ++i ) {
        indices[i] = i;
        positions[i*3] = object->coords()[i*3];//10.0 * i;
        positions[i*3+1] = object->coords()[i*3+1];//20.0 * i;
        positions[i*3+2] = object->coords()[i*3+2];//30.0 * i;
        normals[i*3] =  object->normals()[i*3];//-10.0;
        normals[i*3+1] = object->normals()[i*3+1];//-20.0;
        normals[i*3+2] = object->normals()[i*3+2];//-30.0;
        colors[i*3] = object->colors()[i*3];//1.0 * i;
        colors[i*3+1] = object->colors()[i*3+1];//2.0 * i;
        colors[i*3+2] = object->colors()[i*3+2];//3.0 * i;
    }
    
    shuffledIndices.resize(howManyParticles);
    shuffledPositions.resize(howManyParticles*3);
    shuffledNormals.resize(howManyParticles*3);
    shuffledColors.resize(howManyParticles*3);    
}

void PointShuffle::shuffleWithRandom()
{
    kvs::visclient::Profiler::get()->start("shuffle(random)");
    shuffledPositions = positions;
    shuffledNormals = normals;
    shuffledColors = colors;
    const int howManyParticles = object->coords().size()/3;
    for( size_t i = howManyParticles-1; i>=1; --i) {
        const int random = rand()%i;
        std::swap(shuffledPositions[i*3], shuffledPositions[random*3]);
        std::swap(shuffledPositions[i*3+1], shuffledPositions[random*3+1]);
        std::swap(shuffledPositions[i*3+2], shuffledPositions[random*3+2]);
        std::swap(shuffledNormals[i*3], shuffledNormals[random*3]);
        std::swap(shuffledNormals[i*3+1], shuffledNormals[random*3+1]);
        std::swap(shuffledNormals[i*3+2], shuffledNormals[random*3+2]);
        std::swap(shuffledColors[i*3], shuffledColors[random*3]);
        std::swap(shuffledColors[i*3+1], shuffledColors[random*3+1]);
        std::swap(shuffledColors[i*3+2], shuffledColors[random*3+2]);
    }
    kvs::visclient::Profiler::get()->end("shuffle(random)");
}
    
void PointShuffle::shuffleWithSingle(const int repeatLevel)
{
    const int howManyParticles = object->coords().size()/3;

    kvs::visclient::Profiler::get()->start("shuffle(single)");
    shuffle( howManyParticles, repeatLevel );
    kvs::visclient::Profiler::get()->end("shuffle(single)");
}

void PointShuffle::shuffleWithOpenMP(const int repeatLevel)
{
    const int howManyParticles = object->coords().size()/3;

    kvs::visclient::Profiler::get()->start("shuffle(OpenMP)");
    shuffleWithOpenMP( howManyParticles, repeatLevel );
    kvs::visclient::Profiler::get()->end("shuffle(OpenMP)");
}

void PointShuffle::shuffleWithCUDA(const int repeatLevel)
{
    const int howManyParticles = object->coords().size()/3;

    kvs::visclient::Profiler::get()->start("shuffle(CUDA)");
    ::shuffleWithCUDA( howManyParticles, repeatLevel, indices, shuffledIndices, positions, shuffledPositions, normals, shuffledNormals, colors, shuffledColors );
    kvs::visclient::Profiler::get()->end("shuffle(CUDA)");
}

PointObject* PointShuffle::getAbstractPointObject(const int repeatLevel)
{
    const int howManyParticles = object->coords().size()/3;
    const int howManyForAbstract = (howManyParticles % repeatLevel == 0) ? howManyParticles/repeatLevel : howManyParticles/repeatLevel + 1;
    assert( howManyForAbstract <= shuffledPositions.size() );
    kvs::ValueArray<float> kvsPos(howManyForAbstract *3);
    kvs::ValueArray<unsigned char> kvsColor(howManyForAbstract *3);
    kvs::ValueArray<float> kvsNormal(howManyForAbstract *3);
        
    for( size_t i = 0; i < kvsPos.size(); ++i ) {
        kvsPos[i] = shuffledPositions[i];
        kvsColor[i] = shuffledColors[i];
        kvsNormal[i] = shuffledNormals[i];
    }
    return new PointObject( kvsPos, kvsColor, kvsNormal, 1.0 );
}

std::vector<PointObject*> PointShuffle::getDetailedPointObject(const int repeatLevel)
{
    std::vector<PointObject*> objects(repeatLevel);

    int index = 0;
    
    for( int i = 0; i < repeatLevel; ++i )
    {
        const int howManyParticles = object->coords().size()/3;
        const int howManyForDetail = howManyParticles / repeatLevel;
        kvs::ValueArray<float> kvsPos(howManyForDetail *3);
        kvs::ValueArray<unsigned char> kvsColor(howManyForDetail *3);
        kvs::ValueArray<float> kvsNormal(howManyForDetail *3);
        
        for( size_t j = 0; j < kvsPos.size(); ++j, ++index ) {
                kvsPos[j] = shuffledPositions[index];
                kvsColor[j] = shuffledColors[index];
                kvsNormal[j] = shuffledNormals[index];
        }
        objects[i] = new PointObject( kvsPos, kvsColor, kvsNormal, 1.0 );
    }
    return objects;
}
