#include <stdio.h>
#include <iostream>

#include <cuda_runtime.h>

#include <vector>
#include <cassert>

#include "common.h"

__global__ void setNewIndex(const int *originalIndices, int *newIndices, int numElements, int repeatLevel, const float* originalPositions, float* newPositions, const float *originalNormals, float* newNormals, const unsigned char* originalColors, unsigned char* newColors)
{
    int i = blockDim.x * blockIdx.x + threadIdx.x;

    int howManyParticles = numElements;
    if (i < numElements)
    {
	int index = originalIndices[i];
	int newIndex;
	if( (index % repeatLevel) < (howManyParticles % repeatLevel) ) {
		newIndex = (index % repeatLevel) * ( howManyParticles / repeatLevel + 1 ) + index / repeatLevel;
	}
	else {
		newIndex = (index % repeatLevel) * ( howManyParticles / repeatLevel ) + index / repeatLevel + howManyParticles % repeatLevel;
	}
	newIndices[index] = newIndex;
	newPositions[newIndex*3] = originalPositions[index*3];
	newPositions[newIndex*3+1] = originalPositions[index*3+1];
	newPositions[newIndex*3+2] = originalPositions[index*3+2];
	newNormals[newIndex*3] = originalNormals[index*3];
	newNormals[newIndex*3+1] = originalNormals[index*3+1];
	newNormals[newIndex*3+2] = originalNormals[index*3+2];
	newColors[newIndex*3] = originalColors[index*3];
	newColors[newIndex*3+1] = originalColors[index*3+1];
	newColors[newIndex*3+2] = originalColors[index*3+2];
    }
}

void shuffleWithCUDA(const int howManyParticles, const int repeatLevel, const std::vector<int>& indices, std::vector<int>& shuffledIndices, const std::vector<float>& positions, std::vector<float>& shuffledPositions, const std::vector<float>& normals, std::vector<float>& shuffledNormals, const std::vector<unsigned char>& colors, std::vector<unsigned char>& shuffledColors)
{
	cudaError_t err = cudaSuccess;

	const size_t size = sizeof(int) * howManyParticles;
	const size_t pointSize = sizeof(float) * howManyParticles * 3;
        const size_t colorSize = sizeof(unsigned char) * howManyParticles * 3;

        double start = get_time();
	int *cudaIndices = NULL;
	err = cudaMalloc((void **)&cudaIndices, size);

	int *cudaShuffledIndices = NULL;
	err = cudaMalloc((void **)&cudaShuffledIndices, size);

	float *cudaPositions = NULL;
	err = cudaMalloc((void **)&cudaPositions, pointSize);

	float *cudaShuffledPositions = NULL;
	err = cudaMalloc((void **)&cudaShuffledPositions, pointSize);

	float *cudaNormals = NULL;
	err = cudaMalloc((void **)&cudaNormals, pointSize);

	float *cudaShuffledNormals = NULL;
	err = cudaMalloc((void **)&cudaShuffledNormals, pointSize);

	unsigned char *cudaColors = NULL;
	err = cudaMalloc((void **)&cudaColors, colorSize);

	unsigned char *cudaShuffledColors = NULL;
	err = cudaMalloc((void **)&cudaShuffledColors, colorSize);
        double end = get_time();
        std::cout << "Cuda Malloc : " << end - start << "[ms]" << std::endl;

        start = get_time();

	cudaMemcpy( cudaIndices, &(indices.front()), size, cudaMemcpyHostToDevice);
	cudaMemcpy( cudaPositions, &(positions.front()), pointSize, cudaMemcpyHostToDevice);
	cudaMemcpy( cudaNormals, &(normals.front()), pointSize, cudaMemcpyHostToDevice);
	cudaMemcpy( cudaColors, &(colors.front()), colorSize, cudaMemcpyHostToDevice);
	assert( err == cudaSuccess);

        end = get_time();
        std::cout << "Cuda Memcpy : " << end - start << "[ms]" << std::endl;

        start = get_time();
	int numElements = howManyParticles;
	int threadsPerBlock = 256;
	int blocksPerGrid =(numElements + threadsPerBlock - 1) / threadsPerBlock;
	printf("CUDA kernel launch with %d blocks of %d threads\n", blocksPerGrid, threadsPerBlock);
	setNewIndex<<<blocksPerGrid, threadsPerBlock>>>(cudaIndices, cudaShuffledIndices, numElements, repeatLevel, cudaPositions, cudaShuffledPositions, cudaNormals, cudaShuffledNormals, cudaColors, cudaShuffledColors);
	err = cudaGetLastError();
	assert( err == cudaSuccess);
        
        end = get_time();
        std::cout << "Cuda Shuffle : " << end - start << "[ms]" << std::endl;

        start = get_time();
	cudaMemcpy( &(shuffledIndices.front()), cudaShuffledIndices, size, cudaMemcpyDeviceToHost);
	cudaMemcpy( &(shuffledPositions.front()), cudaShuffledPositions, pointSize, cudaMemcpyDeviceToHost);
	cudaMemcpy( &(shuffledNormals.front()), cudaShuffledNormals, pointSize, cudaMemcpyDeviceToHost);
	cudaMemcpy( &(shuffledColors.front()), cudaShuffledColors, colorSize, cudaMemcpyDeviceToHost);
        end = get_time();
        std::cout << "Cuda Memcpy : " << end - start << "[ms]" << std::endl;

        start = get_time();
	cudaFree(cudaPositions);
	cudaFree(cudaIndices);
	cudaFree(cudaNormals);
	cudaFree(cudaColors);
	cudaFree(cudaShuffledIndices);
	cudaFree(cudaShuffledPositions);
	cudaFree(cudaShuffledNormals);
	cudaFree(cudaShuffledColors);
        end = get_time();
        std::cout << "CudaFree : " << end - start << "[ms]" << std::endl;

        /*start = get_time();
	err = cudaDeviceReset();
        end = get_time();
        std::cout << "CudaDeviceReset : " << end - start << "[ms]" << std::endl;*/

	assert (err == cudaSuccess);
}
