
#include <stdio.h>
#include <cuda.h>
#include "common.h"


__constant__ const float v[20][3] = {
        {-0.5, +0.5, +0.5},
        {-0.5, -0.5, +0.5},
        {+0.5, -0.5, +0.5},
        {+0.5, +0.5, +0.5},
        {-0.5, +0.5, -0.5},
        {-0.5, -0.5, -0.5},
        {+0.5, -0.5, -0.5},
        {+0.5, +0.5, -0.5},
        {-0.5,  0.0, +0.5},
        { 0.0, -0.5, +0.5},
        {+0.5,  0.0, +0.5},
        { 0.0, +0.5, +0.5},
        {-0.5,  0.0, -0.5},
        { 0.0, -0.5, -0.5},
        {+0.5,  0.0, -0.5},
        { 0.0, +0.5, -0.5},
        {-0.5, +0.5,  0.0},
        {-0.5, -0.5,  0.0},
        {+0.5, -0.5,  0.0},
        {+0.5, +0.5,  0.0}
};

__constant__ const float dim = 128;
__constant__ const float kr = 32.0 / dim;
__constant__ const float kd = 6.0;


__global__ void gen_quadratic_hexahedra_data_gpu(
        float* coords,
        unsigned int* connections,
        float* values)
{
    const float center[3] = {blockIdx.x, blockIdx.y, blockIdx.z};

    float posi[3];
    posi[0] = center[0] + v[threadIdx.x][0];
    posi[1] = center[1] + v[threadIdx.x][1];
    posi[2] = center[2] + v[threadIdx.x][2];

    const unsigned int xy_threads = gridDim.x * gridDim.y * blockDim.x;
    const unsigned int index = xy_threads * blockIdx.z +
            gridDim.x * blockDim.x * blockIdx.y +
            blockIdx.x * blockDim.x + threadIdx.x;
    coords[index * 3 + 0] = posi[0];
    coords[index * 3 + 1] = posi[1];
    coords[index * 3 + 2] = posi[2];
    connections[index] = index;

    const float dx = kr * (posi[0] - dim/2.0);
    const float dy = kr * (posi[1] - dim/2.0);
    const float dz = kr * (posi[2] - dim/2.0);
    const float r = sqrt(dx*dx + dy*dy + dz*dz);
    const float cos_theta = dz / r;
    const float phi = kd * r * r * exp(-r/2) *
            (3 * cos_theta * cos_theta - 1);
    float c = phi * phi;
    if (c > 255.0) c = 255.0;
    values[index] = c;
}


extern "C" {

void gen_quadratic_hexahedra_data(
        float* coords,
        unsigned int* connections,
        float* values)
{
    CUDA_SAFE_CALL(cudaSetDevice(0));

    const size_t resolution = 128;
    const size_t lsize = resolution - 1;
    const size_t ncells = lsize * lsize * lsize;
    const size_t nnodes = ncells * 20;

    float* d_coords;
    unsigned int* d_connections;
    float* d_values;

    CUDA_SAFE_CALL(cudaMalloc((void**)&d_coords, sizeof(float) * nnodes * 3));
    CUDA_SAFE_CALL(cudaMalloc((void**)&d_connections, sizeof(unsigned int) * nnodes));
    CUDA_SAFE_CALL(cudaMalloc((void**)&d_values, sizeof(float) * nnodes));

    dim3 dimGrid(lsize, lsize, lsize);
    dim3 dimBlock(20, 1, 1);
    gen_quadratic_hexahedra_data_gpu<<<dimGrid, dimBlock>>>(
            d_coords,
            d_connections,
            d_values);

    CUDA_SAFE_CALL(cudaDeviceSynchronize());

    CUDA_SAFE_CALL(cudaMemcpy(coords, d_coords, sizeof(float) * nnodes * 3,
                cudaMemcpyDeviceToHost));
    CUDA_SAFE_CALL(cudaMemcpy(connections, d_connections, sizeof(unsigned int) * nnodes,
                cudaMemcpyDeviceToHost));
    CUDA_SAFE_CALL(cudaMemcpy(values, d_values, sizeof(float) * nnodes,
                cudaMemcpyDeviceToHost));

    CUDA_SAFE_CALL(cudaFree(d_coords));
    CUDA_SAFE_CALL(cudaFree(d_connections));
    CUDA_SAFE_CALL(cudaFree(d_values));
}

}

