#ifndef PYRAMID_CUH
#define PYRAMID_CUH


__device__ static void pyramid_interpolation_func(
        float* point,
        float* interpolation_func)
{
    const float z = point[2];
    float x, y;
    if (1 - z != 0.0) {
        x = point[0] / (1 - z) * 2;
        y = point[1] / (1 - z) * 2;
    } else {
        x = point[0] * 2;
        y = point[1] * 2;
    }

    const float xy = x * y;
    interpolation_func[0] = z;
    interpolation_func[1] = (float)(0.25 * (1 - z) * (1 - x - y + xy));
    interpolation_func[2] = (float)(0.25 * (1 - z) * (1 + x - y - xy));
    interpolation_func[3] = (float)(0.25 * (1 - z) * (1 + x + y + xy));
    interpolation_func[4] = (float)(0.25 * (1 - z) * (1 - x + y - xy));
}


__device__ static void pyramid_differential_func(
        float* point,
        float* differential_func)
{
    const float z = point[2];
    float x, y;
    if (1 - z != 0.0f) {
        x = point[0] / (1 - z) * 2;
        y = point[1] / (1 - z) * 2;
    } else {
        x = point[0] * 2;
        y = point[1] * 2;
    }

    const float xy = x * y;

    // dNdx
    differential_func[ 0] = 0.0f;
    differential_func[ 1] = (float)(0.25 * (-1 + y));
    differential_func[ 2] = (float)(0.25 * ( 1 - y));
    differential_func[ 3] = (float)(0.25 * ( 1 + y));
    differential_func[ 4] = (float)(0.25 * (-1 - y));

    // dNdy
    differential_func[ 5] = 0.0f;
    differential_func[ 6] = (float)(0.25 * (-1 + x));
    differential_func[ 7] = (float)(0.25 * (-1 - x));
    differential_func[ 8] = (float)(0.25 * ( 1 + x));
    differential_func[ 9] = (float)(0.25 * ( 1 - x));

    // dNdz
    differential_func[10] = 1.0f;
    differential_func[11] = (float)(0.25 * (-1 - xy));
    differential_func[12] = (float)(0.25 * (-1 + xy));
    differential_func[13] = (float)(0.25 * (-1 - xy));
    differential_func[14] = (float)(0.25 * (-1 + xy));
}


__device__ static void pyramid_bind_cell(float* vertices)
{
    float v12[3];
    v12[0] = vertices[2 * 3    ] - vertices[3    ];
    v12[1] = vertices[2 * 3 + 1] - vertices[3 + 1];
    v12[2] = vertices[2 * 3 + 2] - vertices[3 + 2];

    float v14[3];
    v14[0] = vertices[4 * 3    ] - vertices[3    ];
    v14[1] = vertices[4 * 3 + 1] - vertices[3 + 1];
    v14[2] = vertices[4 * 3 + 2] - vertices[3 + 2];

    float v10[3];
    v10[0] = vertices[0 * 3    ] - vertices[3    ];
    v10[1] = vertices[0 * 3 + 1] - vertices[3 + 1];
    v10[2] = vertices[0 * 3 + 2] - vertices[3 + 2];

    // [0], [1]: length of bottom plane, [2]: height of pyramid
    vertices[0] = vector_length(v12);
    vertices[1] = vector_length(v14);

    float height[3];
    // cross
    height[0] = v12[1] * v14[2] - v12[2] * v14[1];
    height[1] = v12[2] * v14[0] - v12[0] * v14[2];
    height[2] = v12[0] * v14[1] - v12[1] * v14[0];
    // normalize
    normalize(height);

    height[0] = height[0] * v10[0];
    height[1] = height[1] * v10[1];
    height[2] = height[2] * v10[2];
    vertices[2] = vector_length(height);
}

/*** debug ***/
__device__ static float pyramid_density
(
 float* density_map,
 unsigned int resolution,
 float  normalize_factor,
 float  min_value,
 float* vertices,
 float* scalars,
 int    ndiv,
 float* x,
 float* w
)
{
  float  density;
  float  v[3];
  float  scalar;
  float  interpolation_func[5];
  float  differential_func[15];
  int    nnodes = 5;
  float  A,B;
  float  J[3][3],jac;

  A = 0.5;
  B = 0.5; 
  density = 0.0;
  for(int k=0;k<ndiv;k++){
    v[2] = B+A*x[k];
    for(int j=0;j<ndiv;j++){
      v[1] = B+A*x[j];
      for(int i=0;i<ndiv;i++){
	v[0] = B+A*x[i];
	// setLocalPoint
	pyramid_interpolation_func(v,interpolation_func);
	pyramid_differential_func(v,differential_func);
// 	jacobi_matrix(nnodes, vertices, differential_func, J);
// 	jac = determinant(J);
 	jac = 1.0;
	scalar = 0.0;
	for(int m=0;m<nnodes;m++){
	  scalar += scalars[m]*interpolation_func[m];
	}
	density += 
	  w[i]*w[j]*w[k]
	  *jac*calculate_density(
				 density_map,
				 resolution,
				 normalize_factor,
				 min_value,
				 scalar);
      }
    }
  }
  return fabsf(A*A*A*density);
}
/*************/

__device__ static float pyramid_volume(float* pyramid)
{
    // [0], [1]: length of bottom plane, [2]: height of pyramid
    return (float)(pyramid[0] * pyramid[1] * pyramid[2] / 3);
}


__device__ static void pyramid_random_sampling(
        float* vertices,
        float* interpolation_func,
        float* differential_func,
	/*** debug ***/
	float s,
	float t,
	float u,
//         tinymt32_status_t* tinymt32,
	/*************/
        float* coord)
    {
    const int nnodes = 5;
    /*** debug ***/
//     const float s = tinymt32_single(tinymt32);
//     const float t = tinymt32_single(tinymt32);
//     const float u = tinymt32_single(tinymt32);
    /*************/
    float point[3];

    if (fabsf(s - 0.5) <= fabsf(t - 0.5) &&
        fabsf(t - 0.5) > fabsf(u - 0.5)) {
        point[0] = (float)(u - 0.5);
        point[1] = (float)(s - 0.5);
        point[2] = (float)(0.5 - fabsf(t - 0.5) * 2);
    } else if (fabsf(u - 0.5) <= fabsf(s - 0.5) &&
        fabsf(t - 0.5) < fabsf(s - 0.5)) {
        point[0] = (float)(t - 0.5);
        point[1] = (float)(u - 0.5);
        point[2] = (float)(0.5 - fabsf(s - 0.5) * 2);
    } else {
        point[0] = (float)(s - 0.5);
        point[1] = (float)(t - 0.5);
        point[2] = (float)(0.5 - fabsf(u - 0.5) * 2);
    }

    // setLocalPoint
    pyramid_interpolation_func(point, interpolation_func);
    pyramid_differential_func(point, differential_func);

    // transformLocalToGlobal
    coord[0] = 0;
    for (int i = 0; i < nnodes; i++) coord[0] += interpolation_func[i] * vertices[i * 3];

    coord[1] = 0;
    for (int i = 0; i < nnodes; i++) coord[1] += interpolation_func[i] * vertices[i * 3 + 1];

    coord[2] = 0;
    for (int i = 0; i < nnodes; i++) coord[2] += interpolation_func[i] * vertices[i * 3 + 2];
}

#endif

