#ifndef QUADRATIC_TETRAHEDRA_CUH
#define QUADRATIC_TETRAHEDRA_CUH


__device__ static void quadratic_tetrahedra_interpolation_func(
        float* point,
        float* interpolation_func)
{
    const float x = point[0];
    const float y = point[1];
    const float z = point[2];
    const float xy = x * y;
    const float yz = y * z;
    const float zx = z * x;

    interpolation_func[0] = 2.0f * ( 1.0f - x - y - z ) * ( 0.5f - x - y - z );
    interpolation_func[1] = 2.0f *  x * ( x - 0.5f );
    interpolation_func[2] = 2.0f *  z * ( z - 0.5f );
    interpolation_func[3] = 2.0f *  y * ( y - 0.5f );
    interpolation_func[4] = 4.0f *  x * ( 1.0f - x - y - z );
    interpolation_func[5] = 4.0f *  z * ( 1.0f - x - y - z );
    interpolation_func[6] = 4.0f *  y * ( 1.0f - x - y - z );
    interpolation_func[7] = 4.0f * zx;
    interpolation_func[8] = 4.0f * yz;
    interpolation_func[9] = 4.0f * xy;
}


__device__ static void quadratic_tetrahedra_differential_func(
        float* point,
        float* differential_func)
{
    const float x = point[0];
    const float y = point[1];
    const float z = point[2];

    // dNdx
    differential_func[0] =  4 * ( x + y + z ) - 3;
    differential_func[1] =  4 * x - 1;
    differential_func[2] =  0;
    differential_func[3] =  0;
    differential_func[4] =  4 * ( 1 - 2 * x - y - z );
    differential_func[5] = -4 * z;
    differential_func[6] = -4 * y;
    differential_func[7] =  4 * z;
    differential_func[8] =  0;
    differential_func[9] =  4 * y;

    // dNdy
    differential_func[10] =  4 * ( x + y + z ) - 3;
    differential_func[11] =  0;
    differential_func[12] =  0;
    differential_func[13] =  4 * y - 1;
    differential_func[14] = -4 * x;
    differential_func[15] = -4 * z;
    differential_func[16] =  4 * ( 1 - x - 2 * y - z );
    differential_func[17] =  0;
    differential_func[18] =  4 * z;
    differential_func[19] =  4 * x;

    // dNdz
    differential_func[20] =  4 * ( x + y + z ) - 3;
    differential_func[21] =  0;
    differential_func[22] =  4 * z - 1;
    differential_func[23] =  0;
    differential_func[24] = -4 * x;
    differential_func[25] =  4 * ( 1 - x - y - 2 * z );
    differential_func[26] = -4 * y;
    differential_func[27] =  4 * x;
    differential_func[28] =  4 * y;
    differential_func[29] =  0;
}
/*** debug ***/
__device__ static float quadratic_tetrahedra_density
(
 float* density_map,
 unsigned int resolution,
 float  normalize_factor,
 float  min_value,
 float* vertices,
 float* scalars,
 int    ndiv,
 float* x,
 float* w
 )
{
  float  density;
  float  v[3];
  float  scalar;
  float  interpolation_func[10];
  float  differential_func[30];
  int    nnodes=10;
  float  A,B;
  float  J[3][3],jac;

  A = 0.5;
  B = 0.5;
  density = 0.0;
  for(int k=0;k<ndiv;k++){
    v[2] = B+A*x[k];
    for(int j=0;j<ndiv;j++){
      v[1] = B+A*x[j];
      for(int i=0;i<ndiv;i++){
	v[0] = B+A*x[i];
	// setLocalPoint
        quadratic_tetrahedra_interpolation_func(v, interpolation_func);
        quadratic_tetrahedra_differential_func(v, differential_func);
// 	jacobi_matrix(nnodes, vertices, differential_func, J);
// 	jac = fabsf(determinant(J));
	jac = 1.0;
	scalar = 0.0;
	for(int m=0;m<nnodes;m++){
	  scalar += scalars[m]*interpolation_func[m];
	}
	density += 
	  w[i]*w[j]*w[k]
	  *jac*calculate_density(
				 density_map,
				 resolution,
				 normalize_factor,
				 min_value,
				 scalar);
      }
    }
  }
  return fabsf(A*A*A*density);
//   return fabsf(A*A*A*density/6.0f);
}
/*************/

__device__ static float quadratic_tetrahedra_volume(
        float* vertices,
        float* interpolation_func,
        float* differential_func)
{
    float v0[3] = {  0,   0,   0};
    float v1[3] = {  1,   0,   0};
    float v2[3] = {  0,   0,   1};
    float v3[3] = {  0,   1,   0};
    float v4[3] = {0.5,   0,   0};
    float v5[3] = {  0,   0, 0.5};
    float v6[3] = {  0, 0.5,   0};
    float v7[3] = {0.5,   0, 0.5};
    float v8[3] = {  0, 0.5, 0.5};
    float v9[3] = {0.5, 0.5,   0};

    float c[8][3] = {
        {
            (v0[0] + v4[0] + v5[0] + v6[0]) * 0.25,
            (v0[1] + v4[1] + v5[1] + v6[1]) * 0.25,
            (v0[2] + v4[2] + v5[2] + v6[2]) * 0.25
        },
        {
            (v4[0] + v1[0] + v7[0] + v9[0]) * 0.25,
            (v4[1] + v1[1] + v7[1] + v9[1]) * 0.25,
            (v4[2] + v1[2] + v7[2] + v9[2]) * 0.25
        },
        {
            (v5[0] + v7[0] + v2[0] + v8[0]) * 0.25,
            (v5[1] + v7[1] + v2[1] + v8[1]) * 0.25,
            (v5[2] + v7[2] + v2[2] + v8[2]) * 0.25
        },
        {
            (v6[0] + v9[0] + v8[0] + v3[0]) * 0.25,
            (v6[1] + v9[1] + v8[1] + v3[1]) * 0.25,
            (v6[2] + v9[2] + v8[2] + v3[2]) * 0.25
        },
        {
            (v4[0] + v7[0] + v5[0] + v6[0]) * 0.25,
            (v4[1] + v7[1] + v5[1] + v6[1]) * 0.25,
            (v4[2] + v7[2] + v5[2] + v6[2]) * 0.25
        },
        {
            (v4[0] + v9[0] + v7[0] + v6[0]) * 0.25,
            (v4[1] + v9[1] + v7[1] + v6[1]) * 0.25,
            (v4[2] + v9[2] + v7[2] + v6[2]) * 0.25
        },
        {
            (v8[0] + v6[0] + v5[0] + v7[0]) * 0.25,
            (v8[1] + v6[1] + v5[1] + v7[1]) * 0.25,
            (v8[2] + v6[2] + v5[2] + v7[2]) * 0.25
        },
        {
            (v8[0] + v7[0] + v9[0] + v6[0]) * 0.25,
            (v8[1] + v7[1] + v9[1] + v6[1]) * 0.25,
            (v8[2] + v7[2] + v9[2] + v6[2]) * 0.25
        }
    };

    float sum_metric = 0;
    float J[3][3];
    for (size_t i = 0; i < 8; i++) {
        // setLocalPoint
        quadratic_tetrahedra_interpolation_func(c[i], interpolation_func);
        quadratic_tetrahedra_differential_func(c[i], differential_func);

        // Jacobi Matrix
        jacobi_matrix(10, vertices, differential_func, J);

        // determinant
        float metric_element = determinant(J);

        sum_metric += fabsf(metric_element);
    }

    return sum_metric / (6.0f * 8.0f);
}


__device__ static void quadratic_tetrahedra_random_sampling(
        float* vertices,
        float* interpolation_func,
        float* differential_func,
	/*** debug ***/
	float s,
	float t,
	float u,
//         tinymt32_status_t* tinymt32,
	/*************/
        float* coord)
{
    const int nnodes = 10;
    float point[3];
    /*** debug ***/
//     const float s = tinymt32_single(tinymt32);
//     const float t = tinymt32_single(tinymt32);
//     const float u = tinymt32_single(tinymt32);
    /*************/

    if (s + t + u <= 1.0f) {
        point[0] = s;
        point[1] = t;
        point[2] = u;
    } else if (s - t + u >= 1.0f) {
        // Revise the point.
        point[0] = -u + 1.0f;
        point[1] = -s + 1.0f;
        point[2] =  t;
    } else if (s + t - u >= 1.0f) {
        // Revise the point.
        point[0] = -s + 1.0f;
        point[1] = -t + 1.0f;
        point[2] =  u;
    } else if (-s + t + u >= 1.0f) {
        // Revise the point.
        point[0] = -u + 1.0f;
        point[1] =  s;
        point[2] = -t + 1.0f;
    } else {
        // Revise the point.
        point[0] =   0.5f * s - 0.5f * t - 0.5f * u + 0.5f;
        point[1] = - 0.5f * s + 0.5f * t - 0.5f * u + 0.5f;
        point[2] = - 0.5f * s - 0.5f * t + 0.5f * u + 0.5f;
    }

    // setLocalPoint
    quadratic_tetrahedra_interpolation_func(point, interpolation_func);
    quadratic_tetrahedra_differential_func(point, differential_func);

    // transformLocalToGlobal
    coord[0] = 0;
    for (int i = 0; i < nnodes; i++) coord[0] += interpolation_func[i] * vertices[i * 3];

    coord[1] = 0;
    for (int i = 0; i < nnodes; i++) coord[1] += interpolation_func[i] * vertices[i * 3 + 1];

    coord[2] = 0;
    for (int i = 0; i < nnodes; i++) coord[2] += interpolation_func[i] * vertices[i * 3 + 2];
}

#endif

