#ifndef TETRAHEDRA_CUH
#define TETRAHEDRA_CUH


__device__ static void tetrahedra_interpolation_func(
        float* point,
        float* interpolation_func)
{
    interpolation_func[0] = point[0];
    interpolation_func[1] = point[1];
    interpolation_func[2] = point[2];
    interpolation_func[3] = 1.0f - point[0] - point[1] - point[2];
}


__device__ static void tetrahedra_differential_func(
        float* differential_func)
{
    // dNdx
    differential_func[ 0] =  1.0f;
    differential_func[ 1] =  0.0f;
    differential_func[ 2] =  0.0f;
    differential_func[ 3] = -1.0f;

    // dNdy
    differential_func[ 4] =  0.0f;
    differential_func[ 5] =  1.0f;
    differential_func[ 6] =  0.0f;
    differential_func[ 7] = -1.0f;

    // dNdz
    differential_func[ 8] =  0.0f;
    differential_func[ 9] =  0.0f;
    differential_func[10] =  1.0f;
    differential_func[11] = -1.0f;
}

/*** debug ***/
__device__ static float tetrahedra_density
(
 float* density_map,
 unsigned int resolution,
 float  normalize_factor,
 float  min_value,
 float* vertices,
 float* scalars,
 int    ndiv,
 float* x,
 float* w
 )
{
  float  density;
  float  v[3];
  float  scalar;
  float  interpolation_func[4];
  float  differential_func[12];
  int    nnodes = 4;
  float  A,B;
  float  J[3][3],jac;

  tetrahedra_differential_func(differential_func);
  jacobi_matrix(nnodes,vertices,differential_func,J);
  jac = fabsf(determinant(J));

  A = 0.5;
  B = 0.5; 
  density = 0.0;
  for(int k=0;k<ndiv;k++){
    v[2] = B+A*x[k];
    for(int j=0;j<ndiv;j++){
      v[1] = B+A*x[j];
      for(int i=0;i<ndiv;i++){
	v[0] = B+A*x[i];
	tetrahedra_interpolation_func(v,interpolation_func);
	scalar = 0.0;
	for(int m=0;m<nnodes;m++){
	  scalar += scalars[m]*interpolation_func[m];
	}
	density += 
	  jac*w[i]*w[j]*w[k]
	  *calculate_density(
			     density_map,
			     resolution,
			     normalize_factor,
			     min_value,
			     scalar);
      }
    }
  }
  return fabsf(A*A*A*density/6.0f);
}
/*************/

__device__ static float tetrahedra_volume(float* vertices)
{
    float3 v1;
    float3 v2;
    float3 v3;
    float3 cross;
    float volume;

    v1.x = vertices[3] - vertices[0];
    v1.y = vertices[4] - vertices[1];
    v1.z = vertices[5] - vertices[2];

    v2.x = vertices[6] - vertices[0];
    v2.y = vertices[7] - vertices[1];
    v2.z = vertices[8] - vertices[2];

    v3.x = vertices[9] - vertices[0];
    v3.y = vertices[10] - vertices[1];
    v3.z = vertices[11] - vertices[2];

    cross.x = v1.y * v2.z - v1.z * v2.y;
    cross.y = v1.z * v2.x - v1.x * v2.z;
    cross.z = v1.x * v2.y - v1.y * v2.x;

    volume = cross.x * v3.x +
             cross.y * v3.y +
             cross.z * v3.z;
    return fabsf(volume *= 0.166666f);
}

__device__ static void tetrahedra_random_sampling(
        float* vertices,
        float* interpolation_func,
        float* differential_func,
	float  s,
	float  t,
	float  u,
        float* coord)
{
    float point[3];
    float M[3][3];
//     const float s = tinymt32_single(tinymt32);
//     const float t = tinymt32_single(tinymt32);
//     const float u = tinymt32_single(tinymt32);

    if (s + t + u <= 1.0f) {
        point[0] = s;
        point[1] = t;
        point[2] = u;
    } else if (s - t + u >= 1.0f) {
        // Revise the point.
        point[0] = -u + 1.0f;
        point[1] = -s + 1.0f;
        point[2] =  t;
    } else if (s + t - u >= 1.0f) {
        // Revise the point.
        point[0] = -s + 1.0f;
        point[1] = -t + 1.0f;
        point[2] =  u;
    } else if (-s + t + u >= 1.0f) {
        // Revise the point.
        point[0] = -u + 1.0f;
        point[1] =  s;
        point[2] = -t + 1.0f;
    } else {
        // Revise the point.
        point[0] =   0.5f * s - 0.5f * t - 0.5f * u + 0.5f;
        point[1] = - 0.5f * s + 0.5f * t - 0.5f * u + 0.5f;
        point[2] = - 0.5f * s - 0.5f * t + 0.5f * u + 0.5f;
    }

    // setLocalPoint
    tetrahedra_interpolation_func(point, interpolation_func);
    tetrahedra_differential_func(differential_func);

    // transformLocalToGlobal
    // v03
    M[0][0] = vertices[0] - vertices[9];
    M[0][1] = vertices[1] - vertices[10];
    M[0][2] = vertices[2] - vertices[11];
    // v13
    M[1][0] = vertices[3] - vertices[9];
    M[1][1] = vertices[4] - vertices[10];
    M[1][2] = vertices[5] - vertices[11];
    // v23
    M[2][0] = vertices[6] - vertices[9];
    M[2][1] = vertices[7] - vertices[10];
    M[2][2] = vertices[8] - vertices[11];

    // M * point + v3
    coord[0] = M[0][0] * point[0] + M[1][0] * point[1] + M[2][0] * point[2] + vertices[9];
    coord[1] = M[0][1] * point[0] + M[1][1] * point[1] + M[2][1] * point[2] + vertices[10];
    coord[2] = M[0][2] * point[0] + M[1][2] * point[1] + M[2][2] * point[2] + vertices[11];
}

#endif

