#include "FrontSTRFileReader.h"

namespace kvs
{

FrontSTRFileReader::FrontSTRFileReader( const char* dirname,
                                        const int process_id,
                                        const int nprocesses,
                                        const ConnectionReadMode connection_read_mode,
                                        const CellType cell_type,
                                        const int vpower,
                                        const int max_length ):
m_process_id( process_id ),
m_nprocesses( nprocesses ),
m_vpower( vpower ),
m_connection_read_mode( connection_read_mode ),
m_cell_type( cell_type ),
m_max_length( max_length ),
m_assigned_coordsdata( NULL ),
m_assigned_nodesdata( NULL ),
m_assigned_nconnections( 0 ),
m_assigned_nvalues( 0 ),
m_assigned_ncoords( 0 )
{
    snprintf(m_connections_filename, m_max_length, "%s/connections.raw", dirname);
    snprintf(m_coords_filename, m_max_length, "%s/coords.raw", dirname);
    snprintf(m_nodes_filename, m_max_length, "%s/nodes.raw", dirname);
#if 0
    _snprintf(m_connections_filename, m_max_length, "%s/connections.raw", dirname);
    _snprintf(m_coords_filename, m_max_length, "%s/coords.raw", dirname);
    _snprintf(m_nodes_filename, m_max_length, "%s/nodes.raw", dirname);
#endif
    read();
}

FrontSTRFileReader::~FrontSTRFileReader( void )
{
}

void FrontSTRFileReader::read( void )
{
    if( m_cell_type == QuadraticTetrahedra )
    {
        m_each_nnodes = 10;
    }
    else if( m_cell_type == Tetrahedra )
    {
        m_each_nnodes = 4;
    }

    if( m_connection_read_mode == Skip )
    {
        m_assigned_connections = getConnections_SkipMode();
    }
    else if( m_connection_read_mode == Block )
    {
        m_assigned_connections = getConnections_BlockMode();
    }

    m_assigned_coords = getCoords();
    m_assigned_values = getValues();

}

ValueArray<kvs::UInt32> FrontSTRFileReader::getConnections_SkipMode( void )
{
    FILE* fp_connections = fopen( m_connections_filename, "r");
        if(!fp_connections)
        {
            printf("file:%s cannot open.", m_connections_filename);
            exit(0);
        }

    fseek( fp_connections, 0, SEEK_END );
    int nconnections = ftell( fp_connections ) / sizeof(kvs::UInt32);
    int nelements   = nconnections / m_each_nnodes;
    fseek( fp_connections, 0, SEEK_SET );

    int assigned_nelements = nelements / m_nprocesses;
    if( m_process_id < ( nelements % m_nprocesses ) )
    {
        assigned_nelements++;
    }

    m_assigned_nconnections = assigned_nelements * m_each_nnodes;
    kvs::UInt32* assigned_connectionsdata = new kvs::UInt32[ m_assigned_nconnections ];

    int stride = ( m_nprocesses - 1 ) * m_each_nnodes * sizeof(kvs::UInt32); // byte
    fseek( fp_connections, m_process_id * m_each_nnodes * sizeof(kvs::UInt32) , SEEK_SET );
    int k = 0;
    for( int i = 0; i < assigned_nelements; i++, k+=m_each_nnodes )
    {
        fread( &assigned_connectionsdata[k], sizeof(kvs::UInt32), m_each_nnodes, fp_connections );
        fseek( fp_connections, stride, SEEK_CUR );
    }

    fclose( fp_connections );

    ValueArray<kvs::UInt32> connections_array( assigned_connectionsdata, m_assigned_nconnections );
    delete assigned_connectionsdata;
    return connections_array;
}

ValueArray<kvs::UInt32> FrontSTRFileReader::getConnections_BlockMode( void )
{
    FILE* fp_connections = fopen( m_connections_filename, "r");
    if(!fp_connections)
    {
        printf("file:%s cannot open.", m_connections_filename);
        exit(0);
    }

    fseek( fp_connections, 0, SEEK_END );
    int nconnections = ftell( fp_connections ) / sizeof(kvs::UInt32);
    int nelements   = nconnections / 4;
    fseek( fp_connections, 0, SEEK_SET );

    int startid = (int)( nelements / m_nprocesses * m_process_id );
    int endid = 0;
    if( ( m_process_id + 1 ) != m_nprocesses )
    {
        endid = (int)( ( nelements / m_nprocesses * ( m_process_id + 1 ) ) - 1 );
    }
    else
    {
        endid = nelements - 1;
    }
    int assigned_nelements = endid - startid + 1 ;

    m_assigned_nconnections = assigned_nelements * 4;
    kvs::UInt32* assigned_connectionsdata = new kvs::UInt32[ m_assigned_nconnections ];

    fseek( fp_connections, startid * 4 * sizeof(kvs::UInt32), SEEK_SET );
    fread( assigned_connectionsdata, sizeof(kvs::UInt32), assigned_nelements * 4, fp_connections );
    fclose( fp_connections );

    ValueArray<kvs::UInt32> connections_array( assigned_connectionsdata, m_assigned_nconnections );
    delete assigned_connectionsdata;
    return connections_array;
}

ValueArray<float> FrontSTRFileReader::getCoords( void )
{
    FILE* fp_coords = fopen( m_coords_filename, "r");
       if(!fp_coords)
        {
            printf("file:%s cannot open.", m_coords_filename);
            exit(0);
        }
    fseek( fp_coords, 0, SEEK_END );
    int ncoords = ftell( fp_coords ) / sizeof(float);
    fseek( fp_coords, 0, SEEK_SET );

    m_assigned_ncoords = ncoords;
    ValueArray<float> coords_array( m_assigned_ncoords );
    fread( coords_array.pointer(), sizeof(float), m_assigned_ncoords, fp_coords );

    fclose( fp_coords );

    if( m_vpower > 1 )
    {
        for( int i=0; i<m_assigned_ncoords; i++ )
        {
            coords_array[i] *= m_vpower;
        }
    }

    return coords_array;
}

ValueArray<kvs::Real32> FrontSTRFileReader::getValues( void )
{
    FILE* fp_nodes = fopen( m_nodes_filename, "r");
       if(!fp_nodes)
        {
            printf("file:%s cannot open.", m_nodes_filename);
            exit(0);
        }
    fseek( fp_nodes, 0, SEEK_END );
    int nnodes = ftell( fp_nodes ) / sizeof(kvs::Real32);
    fseek( fp_nodes, 0, SEEK_SET );

    m_assigned_nvalues = nnodes;
    ValueArray<kvs::Real32> nodes_array( m_assigned_nvalues );
    fread( nodes_array.pointer(), sizeof(kvs::Real32), m_assigned_nvalues, fp_nodes );

    fclose( fp_nodes );
    return nodes_array;
}

} // end of namespace kvs
