#ifndef COMMON_KERNEL_CUH
#define COMMON_KERNEL_CUH

__device__ static size_t clamp(size_t a, size_t low, size_t high)
{
    return (a <= low ? low : a <= high ? a : high);
}

__device__ static float Clamp(float a, float low, float high)
{
    return (a <= low ? low : a <= high ? a : high);
}


// v.length()
__device__ static float vector_length(float* v)
{
    return sqrtf(v[0] * v[0] + v[1] * v[1] + v[2] * v[2]);
}


// a.normalize()
__device__ static void normalize(float* a)
{
    const float length = vector_length(a);
    const float normalize_factor = length > 0.0f ? 1.0f / length : 0.0f;
    a[0] *= normalize_factor;
    a[1] *= normalize_factor;
    a[2] *= normalize_factor;
}


// out = a.cross(b)
__device__ static void cross(float* a, float* b, float* out)
{
    out[0] = a[1] * b[2] - a[2] * b[1];
    out[1] = a[2] * b[0] - a[0] * b[2];
    out[2] = a[0] * b[1] - a[1] * b[0];
}


// a.inverse()
__device__ static void inverse(float a[3][3])
{
    const float det22[9] = {
        a[1][1] * a[2][2] - a[1][2] * a[2][1],
        a[1][0] * a[2][2] - a[1][2] * a[2][0],
        a[1][0] * a[2][1] - a[1][1] * a[2][0],
        a[0][1] * a[2][2] - a[0][2] * a[2][1],
        a[0][0] * a[2][2] - a[0][2] * a[2][0],
        a[0][0] * a[2][1] - a[0][1] * a[2][0],
        a[0][1] * a[1][2] - a[0][2] * a[1][1],
        a[0][0] * a[1][2] - a[0][2] * a[1][0],
        a[0][0] * a[1][1] - a[0][1] * a[1][0]
    };

    const float det33 = a[0][0] * det22[0] - a[0][1] * det22[1] + a[0][2] * det22[2];
    a[0][0] =  det22[0];
    a[0][1] = -det22[3];
    a[0][2] =  det22[6];
    a[1][0] = -det22[1];
    a[1][1] =  det22[4];
    a[1][2] = -det22[7];
    a[2][0] =  det22[2];
    a[2][1] = -det22[5];
    a[2][2] =  det22[8];

    const float det_inverse = 1.0f / det33;
    for (int i = 0; i < 3; i++) {
        for (int j = 0; j < 3; j++) {
            a[i][j] *= det_inverse;
        }
    }
}


__device__ static void jacobi_matrix(
        const unsigned int nnodes,
        float* vertices,
        float* differential_func,
        float out[3][3])
{
    unsigned int i;

    out[0][0] = 0.0f;
    for (i = 0; i < nnodes; i++) out[0][0] = __fadd_rn(out[0][0], __fmul_rn(differential_func[i], vertices[i * 3]));
    out[0][1] = 0.0f;
    for (i = 0; i < nnodes; i++) out[0][1] = __fadd_rn(out[0][1], __fmul_rn(differential_func[i], vertices[i * 3 + 1]));
    out[0][2] = 0.0f;
    for (i = 0; i < nnodes; i++) out[0][2] = __fadd_rn(out[0][2], __fmul_rn(differential_func[i], vertices[i * 3 + 2]));

    out[1][0] = 0.0f;
    for (i = 0; i < nnodes; i++) out[1][0] = __fadd_rn(out[1][0], __fmul_rn(differential_func[nnodes + i], vertices[i * 3]));
    out[1][1] = 0.0f;
    for (i = 0; i < nnodes; i++) out[1][1] = __fadd_rn(out[1][1], __fmul_rn(differential_func[nnodes + i], vertices[i * 3 + 1]));
    out[1][2] = 0.0f;
    for (i = 0; i < nnodes; i++) out[1][2] = __fadd_rn(out[1][2], __fmul_rn(differential_func[nnodes + i], vertices[i * 3 + 2]));

    out[2][0] = 0.0f;
    for (i = 0; i < nnodes; i++) out[2][0] = __fadd_rn(out[2][0], __fmul_rn(differential_func[nnodes * 2 + i], vertices[i * 3]));
    out[2][1] = 0.0f;
    for (i = 0; i < nnodes; i++) out[2][1] = __fadd_rn(out[2][1], __fmul_rn(differential_func[nnodes * 2 + i], vertices[i * 3 + 1]));
    out[2][2] = 0.0f;
    for (i = 0; i < nnodes; i++) out[2][2] = __fadd_rn(out[2][2], __fmul_rn(differential_func[nnodes * 2 + i], vertices[i * 3 + 2]));
}


__device__ static float determinant(float m[3][3])
{
    float det22[3];
    det22[0] = __fmul_rn(m[1][1], m[2][2]) - __fmul_rn(m[1][2], m[2][1]);
    det22[1] = __fmul_rn(m[1][0], m[2][2]) - __fmul_rn(m[1][2], m[2][0]);
    det22[2] = __fmul_rn(m[1][0], m[2][1]) - __fmul_rn(m[1][1], m[2][0]);

    return __fmul_rn(m[0][0], det22[0]) - __fmul_rn(m[0][1], det22[1]) + __fmul_rn(m[0][2], det22[2]);
}


__device__ static size_t calc_number_of_particles(
        float density,
        float volume,
        tinymt32_status_t* mt_st)
{
    const float N = density * volume;
    const float R = tinymt32_single(mt_st);

    size_t n = (size_t)N;
    if (N - n > R) ++n;

    return n;
}


__device__ static float calculate_density(
        float* density_map,
        unsigned int resolution,
        float normalize_factor,
        float min_value,
        float scalar)
{
    const float normalized_scalar = (scalar - min_value) * normalize_factor;
    const size_t index0 = (size_t)normalized_scalar;
    const size_t index1 = index0 + 1;
    const float scalar_offset = normalized_scalar - index0;

    if (index0 == resolution - 1) {
        return density_map[index0];
    } else {
        float rho0 = density_map[index0];
        float rho1 = density_map[index1];
        return (rho1 - rho0) * scalar_offset + rho0;
    }
}


__device__ static float calculate_maximum_density(
        float scalar0,
        float scalar1,
        float* density_map,
        unsigned int resolution,
        float min_value,
        float max_value)
{
    if (scalar0 > scalar1) {
        return 0.0f;
    }

    const float normalize_factor = (resolution - 1) / (max_value - min_value);
    const size_t index0 = (size_t)((scalar0 - min_value) * normalize_factor) + 1;
    const size_t index1 = (size_t)((scalar1 - min_value) * normalize_factor);

    float maximum_density = density_map[index0];
    for (size_t i = index0 + 1; i <= index1; i++) {
        maximum_density = density_map[i] > maximum_density ?
                density_map[i] : maximum_density;
    }

    const float density0 = calculate_density(
            density_map,
            resolution,
            normalize_factor,
            min_value,
            scalar0);
    maximum_density = density0 > maximum_density ? density0 : maximum_density;

    const float density1 = calculate_density(
            density_map,
            resolution,
            normalize_factor,
            min_value,
            scalar1);
    maximum_density = density1 > maximum_density ? density1 : maximum_density;

    return maximum_density;
}


__device__ static void gradient(
        const int nnodes,
        float* scalars,
        float* vertices,
        float* differential_func,
        float* out)
{
    size_t i;

    // normal
    float3 g;
    g.x = g.y = g.z = 0.0f;
    for (i = 0; i < nnodes; i++) {
        g.x += scalars[i] * differential_func[i];
        g.y += scalars[i] * differential_func[nnodes + i];
        g.z += scalars[i] * differential_func[nnodes * 2 + i];
    }

    // Jacobi Matrix
    float j[3][3];
    jacobi_matrix(nnodes, vertices, differential_func, j);

    // inverse
    float det22[9];
    det22[0] = j[1][1] * j[2][2] - j[1][2] * j[2][1];
    det22[1] = j[1][0] * j[2][2] - j[1][2] * j[2][0];
    det22[2] = j[1][0] * j[2][1] - j[1][1] * j[2][0];
    det22[3] = j[0][1] * j[2][2] - j[0][2] * j[2][1];
    det22[4] = j[0][0] * j[2][2] - j[0][2] * j[2][0];
    det22[5] = j[0][0] * j[2][1] - j[0][1] * j[2][0];
    det22[6] = j[0][1] * j[1][2] - j[0][2] * j[1][1];
    det22[7] = j[0][0] * j[1][2] - j[0][2] * j[1][0];
    det22[8] = j[0][0] * j[1][1] - j[0][1] * j[1][0];

    float det33 = j[0][0] * det22[0] - j[0][1] * det22[1] + j[0][2] * det22[2];
    if (fabsf(det33) < MATH_TINY_VALUE) {
        out[0] = 0.0f;
        out[1] = 0.0f;
        out[2] = 0.0f;
    } else {
        float inverse[3][3];
        inverse[0][0] =  det22[0] * (1.0f / det33);
        inverse[0][1] = -det22[3] * (1.0f / det33);
        inverse[0][2] =  det22[6] * (1.0f / det33);
        inverse[1][0] = -det22[1] * (1.0f / det33);
        inverse[1][1] =  det22[4] * (1.0f / det33);
        inverse[1][2] = -det22[7] * (1.0f / det33);
        inverse[2][0] =  det22[2] * (1.0f / det33);
        inverse[2][1] = -det22[5] * (1.0f / det33);
        inverse[2][2] =  det22[8] * (1.0f / det33);

        // G
        out[0] = inverse[0][0] * g.x + inverse[0][1] * g.y + inverse[0][2] * g.z;
        out[1] = inverse[1][0] * g.x + inverse[1][1] * g.y + inverse[1][2] * g.z;
        out[2] = inverse[2][0] * g.x + inverse[2][1] * g.y + inverse[2][2] * g.z;
    }
}


__device__ static void colormap_at(
        float value,
        unsigned char* color_map,
        size_t resolution,
        float min_value,
        float max_value,
        unsigned char* out)
{
    if (value <= min_value) {
        out[0] = color_map[0];
        out[1] = color_map[1];
        out[2] = color_map[2];
    } else if (value >= max_value) {
        out[0] = *(color_map + (resolution - 1) * 3);
        out[1] = *(color_map + (resolution - 1) * 3 + 1);
        out[2] = *(color_map + (resolution - 1) * 3 + 2);
    } else {
        float r = (float)(resolution - 1);
        float v = (value - min_value) / (max_value - min_value) * r;
        size_t s0 = (size_t)v;
        size_t s1 = s0 + 1;
        unsigned char c0[3], c1[3];

        c0[0] = *(color_map + 3 * s0);
        c0[1] = *(color_map + 3 * s0 + 1);
        c0[2] = *(color_map + 3 * s0 + 2);

        c1[0] = *(color_map + 3 * s1);
        c1[1] = *(color_map + 3 * s1 + 1);
        c1[2] = *(color_map + 3 * s1 + 2);

        out[0] = (unsigned char)((c1[0] - c0[0]) * v + c0[0] * s1 - c1[0] * s0);
        out[1] = (unsigned char)((c1[1] - c0[1]) * v + c0[1] * s1 - c1[1] * s0);
        out[2] = (unsigned char)((c1[2] - c0[2]) * v + c0[2] * s1 - c1[2] * s0);
    }
}

#endif

