
#include <stdio.h>
#include <cuda.h>
#include "common.h"


__constant__ const float v[9][3] = {
        {-0.5, +0.5, +0.5},
        {-0.5, -0.5, +0.5},
        {+0.5, -0.5, +0.5},
        {+0.5, +0.5, +0.5},
        {-0.5, +0.5, -0.5},
        {-0.5, -0.5, -0.5},
        {+0.5, -0.5, -0.5},
        {+0.5, +0.5, -0.5},
        { 0.0,  0.0,  0.0}
};

__constant__ int local_connection[6][5] = {
        {8, 3, 2, 1, 0},
        {8, 4, 5, 6, 7},
        {8, 0, 1, 5, 4},
        {8, 1, 2, 6, 5},
        {8, 2, 3, 7, 6},
        {8, 0, 4, 7, 3}
};

__constant__ const float dim = 128;
__constant__ const float kr = 32.0 / dim;
__constant__ const float kd = 6.0;


__global__ void gen_pyramid_data_gpu(
        float* coords,
        unsigned int* connections,
        float* values)
{
    const float center[3] = {blockIdx.x, blockIdx.y, blockIdx.z};
    const size_t lnid = local_connection[threadIdx.y][threadIdx.x];

    float posi[3];
    posi[0] = center[0] + v[lnid][0];
    posi[1] = center[1] + v[lnid][1];
    posi[2] = center[2] + v[lnid][2];

    const unsigned int threads_per_block = blockDim.x * blockDim.y;
    const unsigned int threads_per_grid_z = gridDim.x * gridDim.y * threads_per_block;
    const unsigned int index = threads_per_grid_z * blockIdx.z +
            gridDim.x * blockIdx.y * threads_per_block +
            blockIdx.x * threads_per_block +
            blockDim.x * threadIdx.y +
            threadIdx.x;
    coords[index * 3 + 0] = posi[0];
    coords[index * 3 + 1] = posi[1];
    coords[index * 3 + 2] = posi[2];
    connections[index] = index;

    const float dx = kr * (posi[0] - dim/2.0);
    const float dy = kr * (posi[1] - dim/2.0);
    const float dz = kr * (posi[2] - dim/2.0);
    const float r = sqrt(dx*dx + dy*dy + dz*dz);
    const float cos_theta = dz / r;
    const float phi = kd * r * r * exp(-r/2) *
            (3 * cos_theta * cos_theta - 1);
    float c = phi * phi;
    if (c > 255.0) c = 255.0;
    values[index] = c;
}


extern "C" {

void gen_pyramid_data(
        float* coords,
        unsigned int* connections,
        float* values)
{
    CUDA_SAFE_CALL(cudaSetDevice(0));

    const size_t resolution = 128;
    const size_t lsize = resolution - 1;
    const size_t ncells = lsize * lsize * lsize * 6;
    const size_t nnodes = ncells * 5;

    float* d_coords;
    unsigned int* d_connections;
    float* d_values;

    CUDA_SAFE_CALL(cudaMalloc((void**)&d_coords, sizeof(float) * nnodes * 3));
    CUDA_SAFE_CALL(cudaMalloc((void**)&d_connections, sizeof(unsigned int) * nnodes));
    CUDA_SAFE_CALL(cudaMalloc((void**)&d_values, sizeof(float) * nnodes));

    dim3 dimGrid(lsize, lsize, lsize);
    dim3 dimBlock(5, 6, 1);
    gen_pyramid_data_gpu<<<dimGrid, dimBlock>>>(
            d_coords,
            d_connections,
            d_values);

    CUDA_SAFE_CALL(cudaDeviceSynchronize());

    CUDA_SAFE_CALL(cudaMemcpy(coords, d_coords, sizeof(float) * nnodes * 3,
                cudaMemcpyDeviceToHost));
    CUDA_SAFE_CALL(cudaMemcpy(connections, d_connections, sizeof(unsigned int) * nnodes,
                cudaMemcpyDeviceToHost));
    CUDA_SAFE_CALL(cudaMemcpy(values, d_values, sizeof(float) * nnodes,
                cudaMemcpyDeviceToHost));

    CUDA_SAFE_CALL(cudaFree(d_coords));
    CUDA_SAFE_CALL(cudaFree(d_connections));
    CUDA_SAFE_CALL(cudaFree(d_values));
}

}

