#ifndef HEXAHEDRA_CUH
#define HEXAHEDRA_CUH


__device__ static void hexahedra_interpolation_func(
        float* point,
        float* interpolation_func)
{
    const float x = point[0];
    const float y = point[1];
    const float z = point[2];
    const float xy = x * y;
    const float yz = y * z;
    const float zx = z * x;
    const float xyz = xy * z;

    interpolation_func[0] = z - zx - yz + xyz;
    interpolation_func[1] = zx - xyz;
    interpolation_func[2] = xyz;
    interpolation_func[3] = yz - xyz;
    interpolation_func[4] = 1.0f - x - y - z + xy + yz + zx - xyz;
    interpolation_func[5] = x - xy - zx + xyz;
    interpolation_func[6] = xy - xyz;
    interpolation_func[7] = y - xy - yz + xyz;
}


__device__ static void hexahedra_differential_func(
        float* point,
        float* differential_func)
{
    const float x = point[0];
    const float y = point[1];
    const float z = point[2];
    const float xy = x * y;
    const float yz = y * z;
    const float zx = z * x;

    // dNdx
    differential_func[ 0] =  - z + yz;
    differential_func[ 1] =  z - yz;
    differential_func[ 2] =  yz;
    differential_func[ 3] =  - yz;
    differential_func[ 4] =  - 1.0f + y +z - yz;
    differential_func[ 5] =  1.0f - y - z + yz;
    differential_func[ 6] =  y - yz;
    differential_func[ 7] =  - y + yz;

    // dNdy
    differential_func[ 8] =  - z + zx;
    differential_func[ 9] =  - zx;
    differential_func[10] =  zx;
    differential_func[11] =  z - zx;
    differential_func[12] =  - 1.0f + x + z - zx;
    differential_func[13] =  - x + zx;
    differential_func[14] =  x - zx;
    differential_func[15] =  1.0f - x - z + zx;

    // dNdz
    differential_func[16] =  1.0f - y - x + xy;
    differential_func[17] =  x - xy;
    differential_func[18] =  xy;
    differential_func[19] =  y - xy;
    differential_func[20] =  - 1.0f + y + x - xy;
    differential_func[21] =  - x + xy;
    differential_func[22] =  - xy;
    differential_func[23] =  - y + xy;
}

/*** debug ***/
__device__ static float hexahedra_density
(
 float* density_map,
 unsigned int resolution,
 float  normalize_factor,
 float  min_value,
 float* vertices,
 float* scalars,
 int    ndiv,
 float* x,
 float* w
 )
{
  float density;
  float  v[3];
  float  scalar;
  float  interpolation_func[8];
  float  differential_func[24];
  int    nnodes=8;
  float  A,B;
  float  J[3][3],jac;

  A = 0.5;
  B = 0.5;
  density = 0.0;
  for(int k=0;k<ndiv;k++){
    v[2] = B+A*x[k];
    for(int j=0;j<ndiv;j++){
      v[1] = B+A*x[j];
      for(int i=0;i<ndiv;i++){
	v[0] = B+A*x[i];
	// setLocalPoint
	hexahedra_interpolation_func(v, interpolation_func);
	hexahedra_differential_func(v, differential_func);
// 	jacobi_matrix(nnodes, vertices, differential_func, J);
//  	jac = fabsf(determinant(J));
 	jac = 1.0;
	scalar = 0.0;
	for(int m=0;m<nnodes;m++){
	  scalar += scalars[m]*interpolation_func[m];
	}
	density += 
	  w[i]*w[j]*w[k]
	  *jac*calculate_density(
				 density_map,
				 resolution,
				 normalize_factor,
				 min_value,
				 scalar);
      }
    }
  }
  return fabsf(A*A*A*density);
}
/*************/

__device__ static float hexahedra_volume(
        float* vertices,
        float* interpolation_func,
        float* differential_func)
{
    const size_t resolution = 3;
    const float sampling_length = 1.0f / (float)resolution;
    const float adjustment = sampling_length * 0.5f;

    float sampling_position[3] = {
        -adjustment,
        -adjustment,
        -adjustment
    };
    float sum_metric = 0;
    float J[3][3];

    for (size_t k = 0; k < resolution; k++) {
        sampling_position[2] += sampling_length;
        for (size_t j = 0; j < resolution; j++) {
            sampling_position[1] += sampling_length;
            for (size_t i = 0; i < resolution; i++) {
                sampling_position[0] += sampling_length;

                // setLocalPoint
                hexahedra_interpolation_func(sampling_position, interpolation_func);
                hexahedra_differential_func(sampling_position, differential_func);

                // Jacobi Matrix
                jacobi_matrix(8, vertices, differential_func, J);

                // determiant
                float metric_element = determinant(J);
                sum_metric += fabsf(metric_element);
            }
            sampling_position[0] = -adjustment;
        }
        sampling_position[1] = -adjustment;
    }

    return sum_metric / (resolution * resolution * resolution);
}


__device__ static void hexahedra_random_sampling(
        float* vertices,
        float* interpolation_func,
        float* differential_func,
	/*** debug ***/
	float s,
	float t,
	float u,
//         tinymt32_status_t* tinymt32,
	/*************/
        float* coord)
    {
    const int nnodes = 8;
    float point[3];
    /*** debug ***/
    point[0] = s;
    point[1] = t;
    point[2] = u;
//     point[0] = tinymt32_single(tinymt32);
//     point[1] = tinymt32_single(tinymt32);
//     point[2] = tinymt32_single(tinymt32);
    /*************/

    // setLocalPoint
    hexahedra_interpolation_func(point, interpolation_func);
    hexahedra_differential_func(point, differential_func);

    // transformLocalToGlobal
    coord[0] = 0;
    for (int i = 0; i < nnodes; i++) coord[0] += interpolation_func[i] * vertices[i * 3];

    coord[1] = 0;
    for (int i = 0; i < nnodes; i++) coord[1] += interpolation_func[i] * vertices[i * 3 + 1];

    coord[2] = 0;
    for (int i = 0; i < nnodes; i++) coord[2] += interpolation_func[i] * vertices[i * 3 + 2];
}

#endif

