/*****************************************************************************/
/**
 *  @file   main.cpp
 *  @brief  Example program for kvs::PointRenderer class.
 *  @author Naohisa Sakamoto
 */
/*----------------------------------------------------------------------------
 *
 *  Copyright 2007 Visualization Laboratory, Kyoto University.
 *  All rights reserved.
 *  See http://www.viz.media.kyoto-u.ac.jp/kvs/copyright/ for details.
 *
 *  $Id: main.cpp 381 2009-11-30 06:10:28Z naohisa.sakamoto $
 */
/*****************************************************************************/

#define REMOTE

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <kvs/Message>
#include <kvs/PointImporter>
#include <kvs/PointObject>
#include <kvs/PointRenderer>
#include <kvs/glut/Application>
#include <kvs/glut/Screen>
#ifdef REMOTE
#include <kvs/ParticleVolumeRenderer>
#else
#include <kvs/glew/ParticleVolumeRenderer>
#endif
#include <kvs/Date>
#include <kvs/Time>
#include <kvs/CommandLine>
#include <kvs/TransferFunction>
#include <kvs/UnstructuredVolumeObject>
#include <kvs/UnstructuredVolumeImporter>
#include <kvs/CellByCellUniformSampling>
#include <kvs/CellByCellRejectionSampling>
#include <kvs/CellByCellMetropolisSampling>
#include <kvs/CellByCellLayeredSampling>
/*** debug ***/
#include <kvs/PointExporter>
#include <kvs/KVSMLObjectPoint>
/*************/
#include "CellByCellUniformSampling.h"
#include "CellByCellRejectionSampling.h"
#include "CellByCellMetropolisSampling.h"
#include "common.h"
#include "KVSMLObjectPointWriter.h"
#include "point_object_summation.h"
#include "TestVolume.h"
#include "FrontSTRFileReader.h"
#include "DistributedUnstructuredVolumeImporter.h"
#include "AVSUcd.h"
/*** debug ***/
#include "AVSField.h"
#include <kvs/Vector>
/*************/

#ifdef ENABLE_MPI
#include <mpi.h>
#include "mpi_controller.h"
#include "DistributedUniformSampling.h"
#include "DistributedRejectionSampling.h"
#include "DistributedMetropolisSampling.h"
#endif

extern "C" {
    void gen_quadratic_hexahedra_data(
        float* coords,
        unsigned int* connections,
        float* values);
  
    void gen_pyramid_data(
        float* coords,
        unsigned int* connections,
        float* values);

/*** debug ***/
    int init_random(void);
    int term_random(void);
/*************/
    int init_tinymt(void);
    int term_tinymt(void);
}

// int HTTR_NUM = 10;
int HTTR_NUM = 200;
// int GT5D_NUM = 2;
int GT5D_NUM = 3200;

/*** debug ***/
extern int mpi_rank;
extern int time_series;
extern int gl_integration_ndiv;
extern int random_number_method;
extern int gpuPrefixScan;
extern int gpuMultiStream;
extern int thread_size;
extern int jacobCalc;
extern float TIMER[10][100];

float TOTAL[10];
float MAXVAL[10];
/*************/

/*===========================================================================*/
/**
 *  @brief  User-defined argument class.
 */
/*===========================================================================*/
class Argument : public kvs::CommandLine
{
public:

    std::string input_data;
    kvs::TransferFunction transfer_function;
    size_t subpixel_level;
    size_t repeat_level;
    float sampling_step;
    bool dump;
    char sampling_method;
    std::string hexahedra_dir;
    std::string quadratic_tetrahedra_dir;
    bool pyramid;
    bool test_volume;
    /*** debug ***/
    bool gt5d;
    int gl_integration;
    size_t random_number;
    std::string scheduling;
    bool doublebuff;
    bool prefix;
    bool jacob;
    /*********************/

    Argument( int argc, char** argv ):
        kvs::CommandLine( argc, argv )
        {
            // Add help option (generate help message automatically).
            add_help_option();

            // Add input value.
            add_value( "input file file name.",false );
            add_option( "t","transfer_function",1,false );
            add_option( "s","subpixel_level",1,false );
            add_option( "r","repeat_level",1,false );
            add_option( "step","sampling_step",1,false );
            add_option( "S","sampling method",1,false );
            add_option( "d","DEBUG:dump result",0,false );
            add_option( "H","DEBUG:read Hexahedra data dir",1,false );
            add_option( "qt","DEBUG:read Quadratic Tetrahedra data dir",1,false );
            add_option( "p","DEBUG:generate pyramid data",0,false );
            add_option( "V","DEBUG:test volume calculation",0,false );
            add_option( "G","GT5D data",0, false );
            /*** Added by MHIR ***/
            add_option( "I","Gauss Legendre Integration",1,false );
            add_option( "R","Generate quasi Random number",1,false );
            add_option( "sc","Scheduling method",1,false );
            add_option( "pf","Prefix Scan",0,false );
            add_option( "db","Double Buffering",0,false );
            add_option( "thread","Max Thread size",1,false );
            add_option( "J","High-precision density calculation",0,false );
            /*********************/

            if ( !this->parse() ) exit( EXIT_FAILURE );

            subpixel_level = 1;
            repeat_level = 4;
            sampling_step = 0.5f;
            dump = false;
            sampling_method = 'u';
            hexahedra_dir = "";
            quadratic_tetrahedra_dir = "";
            pyramid = false;
            test_volume = false;
            /*** debug ***/
            gt5d = false;
            gl_integration = 0;
            random_number = 0;
            scheduling = "";
            gpuPrefixScan = 0;
            gpuMultiStream = 0;
            prefix=false;
            doublebuff=false;
            thread_size=256;
            jacob=false;
            /*************/

            if( this->hasOption("t") ) transfer_function.read( this->optionValue<std::string>("t") );
            else transfer_function.create( 256 );
            if( this->hasOption("s") ) subpixel_level = this->optionValue<size_t>("s");
            if( this->hasOption("r") ) repeat_level = this->optionValue<size_t>("r");
            if( this->hasOption("step") ) sampling_step = this->optionValue<float>("step");
            if( this->hasOption("d") ) dump = true;
            if( this->hasOption("S") ) sampling_method = this->optionValue<char>("S");
            if( this->hasOption("H") ) hexahedra_dir = this->optionValue<std::string>("H");
            if( this->hasOption("qt") ) quadratic_tetrahedra_dir = this->optionValue<std::string>("qt");
            if( this->hasOption("p") ) pyramid = true;
            if( this->hasOption("V") ) test_volume = true;
            if( this->hasOption("G") ) gt5d = true;
            /*** Added by MHIR ***/
            if( this->hasOption("I") ) gl_integration = this->optionValue<size_t>("I");
            if( this->hasOption("R") ) random_number = this->optionValue<size_t>("R");
            if( this->hasOption("sc") ) scheduling = this->optionValue<std::string>("sc");
            if( this->hasOption("pf") ) prefix=true;
            if( this->hasOption("db") ) doublebuff=true;
            if( this->hasOption("thread") ) thread_size = this->optionValue<int>("thread");
            if( this->hasOption("J") ) jacob=true;
            gl_integration_ndiv = std::max(0,gl_integration);
            random_number_method = random_number;
            if(scheduling=="time") time_series=1;
            if(prefix) gpuPrefixScan=1;
            if(doublebuff) gpuMultiStream=1;
            if(jacob) jacobCalc=1;
            /*********************/
            /*** timer ***/
            for(int i=0;i<10;i++){
                TOTAL[i]=0.0;
                MAXVAL[i]=0.0;
                for(int j=0;j<100;j++){
                    TIMER[i][j]=0.0;
                }
            }
            /*************/

            input_data = this->value<std::string>();
        }
};

inline const size_t GetRevisedSubpixelLevel(
    const size_t subpixel_level,
    const size_t repetition_level )
{
    return( static_cast<size_t>( subpixel_level * std::sqrt( (double)repetition_level ) + 0.5f ) );
}

/*===========================================================================*/
/**
 *  Key press event.
 */
/*===========================================================================*/

class KeyPressEvent : public kvs::KeyPressEventListener
{
private:

public:
    KeyPressEvent( void ){}

    void update( kvs::KeyEvent* event )
        {
            switch( event->key() )
            {
            case kvs::Key::l:
                screen()->controlTarget() = kvs::ScreenBase::TargetLight;
                break;
            case kvs::Key::c:
                screen()->controlTarget() = kvs::ScreenBase::TargetCamera;
                break;
            case kvs::Key::o:
                screen()->controlTarget() = kvs::ScreenBase::TargetObject;
                break;

            case kvs::Key::s:
            {
                kvs::Date today;
                kvs::Time now;
                std::string base1 = "test";
                std::string base2 = base1 + "_" + today.toString();
                std::string base3 = base2 + "_" + now.toString("-");
                std::string filename = base3 + ".bmp";
                kvs::ColorImage image( screen()->camera()->snapshot() );
                image.write( filename.c_str() );
                std::cout<<"Snapshot >>" << filename << std::endl;
            }

            default: break;
            }
        }
};

//***********************************//
//     generate Volume Data          //
//***********************************//
class GenerateVolumeData
{
public:

    enum CellType
    {
        QuadraticHexahedra,
        Pyramid
    };

    kvs::UnstructuredVolumeObject* m_volume;

public:

    GenerateVolumeData( const size_t resolution )
        {
            m_volume = new kvs::UnstructuredVolumeObject( );

            //this->quadraticHexahedra( resolution );
        }

    ~GenerateVolumeData( void )
        {
            delete m_volume;
        }

    void quadraticHexahedra( const size_t resolution )
        {
            kvs::VolumeObjectBase::CellType cell_type =
                kvs::VolumeObjectBase::QuadraticHexahedra;
            const size_t lsize = resolution - 1;
            const size_t ncells = lsize * lsize * lsize;
            const size_t nnodes = ncells * 20;
            kvs::ValueArray<float> coords( nnodes * 3 );
            //kvs::ValueArray<size_t> connections( nnodes );
            kvs::ValueArray<unsigned int> connections( nnodes );
            kvs::ValueArray<float> values( nnodes );

            double start = get_time();
#if 0
            kvs::Vector3f v[20];
            v[ 0].set( -0.5, +0.5, +0.5 );
            v[ 1].set( -0.5, -0.5, +0.5 );
            v[ 2].set( +0.5, -0.5, +0.5 );
            v[ 3].set( +0.5, +0.5, +0.5 );
            v[ 4].set( -0.5, +0.5, -0.5 );
            v[ 5].set( -0.5, -0.5, -0.5 );
            v[ 6].set( +0.5, -0.5, -0.5 );
            v[ 7].set( +0.5, +0.5, -0.5 );
            v[ 8].set( -0.5,  0.0, +0.5 );
            v[ 9].set(  0.0, -0.5, +0.5 );
            v[10].set( +0.5,  0.0, +0.5 );
            v[11].set(  0.0, +0.5, +0.5 );
            v[12].set( -0.5,  0.0, -0.5 );
            v[13].set(  0.0, -0.5, -0.5 );
            v[14].set( +0.5,  0.0, -0.5 );
            v[15].set(  0.0, +0.5, -0.5 );
            v[16].set( -0.5, +0.5,  0.0 );
            v[17].set( -0.5, -0.5,  0.0 );
            v[18].set( +0.5, -0.5,  0.0 );
            v[19].set( +0.5, +0.5,  0.0 );

            const float dim = resolution;
            const float kr = 32.0 / dim;
            const float kd = 6.0;
            float* coord = coords.pointer();
            size_t index = 0;

            for( size_t k = 0; k < lsize; k++ )
            {
                for( size_t j = 0; j < lsize; j++ )
                {
                    for( size_t i = 0; i < lsize; i++ )
                    {
                        kvs::Vector3f center(i,j,k);

                        for( size_t lid = 0; lid < 20; lid++, index++ )
                        {
                            kvs::Vector3f posi = center + v[lid];
                            *coord = posi.x(); coord++;
                            *coord = posi.y(); coord++;
                            *coord = posi.z(); coord++;
                            connections[ index ] = index;

                            const float dx = kr * ( posi.x() - dim/2.0 );
                            const float dy = kr * ( posi.y() - dim/2.0 );
                            const float dz = kr * ( posi.z() - dim/2.0 );
                            const float r = std::sqrt( dx*dx + dy*dy + dz*dz );
                            const float cos_theta = dz / r;
                            const float phi = kd * r * r * std::exp( -r/2 ) *
                                ( 3 * cos_theta * cos_theta - 1 );
                            float c = phi * phi;
                            if( c > 255.0 ) c = 255.0;

                            values[index] = c;
                        }
                    }
                }
            }
#else
            gen_quadratic_hexahedra_data(coords.pointer(), connections.pointer(), values.pointer());
#endif

#if 0 // debug
            float* coords_gpu = (float*)malloc(sizeof(float) * nnodes * 3);
            unsigned int* connections_gpu = (unsigned int*)malloc(sizeof(unsigned int) * nnodes);
            float* values_gpu = (float*)malloc(sizeof(float) * nnodes);

            gen_quadratic_hexahedra_data(coords_gpu, connections_gpu, values_gpu);

            float* cod = coords.pointer();
            for (size_t i = 0; i < nnodes * 3; i++) {
                if (fabsf(cod[i] - coords_gpu[i]) > 1.e-5) {
                    printf("%zu %f %f\n", i, cod[i], coords_gpu[i]);
                }
            }
            printf("coord cmp end\n");
            unsigned int* con = connections.pointer();
            for (size_t i = 0; i < nnodes; i++) {
                if (con[i] != connections_gpu[i]) {
                    printf("%zu %u %u\n", i, con[i], connections_gpu[i]);
                }
            }
            printf("conn cmp end\n");
            float* val = values.pointer();
            for (size_t i = 0; i < nnodes; i++) {
                if (fabsf(val[i] - values_gpu[i]) > 1.e-3) {
                    printf("%zu %f %f\n", i, val[i], values_gpu[i]);
                }
            }
            printf("values cmp end\n");

            free(coords_gpu);
            free(connections_gpu);
            free(values_gpu);
#endif
            double end = get_time();
            printf("gpu gen : %f\n", end - start);
            m_volume->setCellType( cell_type );
            m_volume->setNNodes( nnodes );
            m_volume->setNCells( ncells );
            m_volume->setVeclen( 1 );
            m_volume->setCoords( coords );
            m_volume->setConnections( connections );
            m_volume->setValues( kvs::AnyValueArray( values ) );
            m_volume->updateMinMaxCoords();
            m_volume->updateMinMaxValues();
        }

    void pyramid( const size_t resolution )
        {
            kvs::VolumeObjectBase::CellType cell_type =
                kvs::VolumeObjectBase::Pyramid;
            const size_t lsize = resolution - 1;
            const size_t ncells = lsize * lsize * lsize * 6;
            const size_t nnodes = ncells * 5;
            kvs::ValueArray<float> coords( nnodes * 3 );
            kvs::ValueArray<unsigned int> connections( nnodes );
            kvs::ValueArray<float> values( nnodes );

            double start = get_time();
#if 0
            kvs::Vector3f v[9];
            v[ 0].set( -0.5, +0.5, +0.5 );
            v[ 1].set( -0.5, -0.5, +0.5 );
            v[ 2].set( +0.5, -0.5, +0.5 );
            v[ 3].set( +0.5, +0.5, +0.5 );
            v[ 4].set( -0.5, +0.5, -0.5 );
            v[ 5].set( -0.5, -0.5, -0.5 );
            v[ 6].set( +0.5, -0.5, -0.5 );
            v[ 7].set( +0.5, +0.5, -0.5 );
            v[ 8].set(  0.0,  0.0,  0.0 );

            int local_connection[6][5] =
                {
                    {8,3,2,1,0},
                    {8,4,5,6,7},
                    {8,0,1,5,4},
                    {8,1,2,6,5},
                    {8,2,3,7,6},
                    {8,0,4,7,3}
                };

            const float dim = resolution;
            const float kr = 32.0 / dim;
            const float kd = 6.0;
            float* coord = coords.pointer();
            size_t index = 0;

            for( size_t k = 0; k < lsize; k++ )
            {
                for( size_t j = 0; j < lsize; j++ )
                {
                    for( size_t i = 0; i < lsize; i++ )
                    {
                        kvs::Vector3f center(i,j,k);

                        for( size_t lcell_id = 0; lcell_id < 6; lcell_id++ )
                        {
                            for( size_t lnode_id = 0; lnode_id < 5; lnode_id++, index++ )
                            {
                                const size_t lnid = local_connection[ lcell_id ][ lnode_id ];
                                kvs::Vector3f posi = center + v[ lnid ];
                                *coord = posi.x(); coord++;
                                *coord = posi.y(); coord++;
                                *coord = posi.z(); coord++;
                                connections[ index ] = index;

                                const float dx = kr * ( posi.x() - dim/2.0 );
                                const float dy = kr * ( posi.y() - dim/2.0 );
                                const float dz = kr * ( posi.z() - dim/2.0 );
                                const float r = std::sqrt( dx*dx + dy*dy + dz*dz );
                                const float cos_theta = dz / r;
                                const float phi = kd * r * r * std::exp( -r/2 ) *
                                    ( 3 * cos_theta * cos_theta - 1 );
                                float c = phi * phi;
                                if( c > 255.0 ) c = 255.0;

                                values[index] = c;
                            }
                        }
                    }
                }
            }
#else
            gen_pyramid_data(coords.pointer(), connections.pointer(), values.pointer());
#endif

#if 0 // debug
            float* coords_gpu = (float*)malloc(sizeof(float) * nnodes * 3);
            unsigned int* connections_gpu = (unsigned int*)malloc(sizeof(unsigned int) * nnodes);
            float* values_gpu = (float*)malloc(sizeof(float) * nnodes);

            gen_pyramid_data(coords_gpu, connections_gpu, values_gpu);

            float* cod = coords.pointer();
            for (size_t i = 0; i < nnodes * 3; i++) {
                if (fabsf(cod[i] - coords_gpu[i]) > 1.e-6) {
                    printf("%zu %f %f\n", i, cod[i], coords_gpu[i]);
                }
            }
            printf("coord cmp end\n");
            unsigned int* con = connections.pointer();
            for (size_t i = 0; i < nnodes; i++) {
                if (con[i] != connections_gpu[i]) {
                    printf("%zu %u %u\n", i, con[i], connections_gpu[i]);
                }
            }
            printf("conn cmp end\n");
            float* val = values.pointer();
            for (size_t i = 0; i < nnodes; i++) {
                if (fabsf(val[i] - values_gpu[i]) > 1.e-3) {
                    printf("%zu %f %f\n", i, val[i], values_gpu[i]);
                }
            }
            printf("values cmp end\n");

            free(coords_gpu);
            free(connections_gpu);
            free(values_gpu);
#endif
            double end = get_time();
            printf("gpu gen : %f\n", end - start);
            m_volume->setCellType( cell_type );
            m_volume->setNNodes( nnodes );
            m_volume->setNCells( ncells );
            m_volume->setVeclen( 1 );
            m_volume->setCoords( coords );
            m_volume->setConnections( connections );
            m_volume->setValues( kvs::AnyValueArray( values ) );
            m_volume->updateMinMaxCoords();
            m_volume->updateMinMaxValues();
        }
};

/*** debug ***/
//***********************************//
//     Analyze volume and generate TF//
//***********************************//
class SliceAnalysis
{
protected:

    float m_min_extremum;
    float m_max_extremum;

    const kvs::ColorMap& m_color_map;
    kvs::ValueArray<float> m_opacity;
    kvs::UnstructuredVolumeObject* m_volume;

public:

    SliceAnalysis( kvs::UnstructuredVolumeObject* volume,
                   kvs::Vector3ui resol,
                   const kvs::ColorMap& color_map ):
        m_color_map( color_map ),m_volume( volume )
        {
            m_opacity.allocate( color_map.resolution() );

            this->extract_extremum( resol );
        }

    SliceAnalysis( const kvs::ColorMap& color_map ):
        m_color_map( color_map )
        {
            m_opacity.allocate( color_map.resolution() );
        }

    void analyze( kvs::Vector3ui resol, kvs::UnstructuredVolumeObject* volume )
        {
            m_volume = volume;
            this->extract_extremum( resol );
        }

    float minExtremum( void ){ return m_min_extremum; }
    float maxExtremum( void ){ return m_max_extremum; }
    kvs::ValueArray<float>& opacity( void ){ return m_opacity; }

protected:

    void extract_extremum( kvs::Vector3ui resol )
        {
            m_opacity.fill( 0x00 );

            const kvs::UInt32 line_size = resol.x();
            const kvs::UInt32 slice_size = line_size * resol.y();

            const size_t k = resol.z() / 2;

            for( size_t j = 2; j < resol.y()-2; j++ )
            {
                for( size_t i = 2; i < resol.x()-2; i++ )
                {
                    const size_t index_c = i   +  j    * line_size + k * slice_size;
                    const size_t index_l = i-1 +  j    * line_size + k * slice_size;
                    const size_t index_r = i+1 +  j    * line_size + k * slice_size;
                    const size_t index_u = i   + (j-1) * line_size + k * slice_size;
                    const size_t index_d = i   + (j+1) * line_size + k * slice_size;
                    const size_t index_lu= i+1 + (j+1) * line_size + k * slice_size;
                    const size_t index_ld= i+1 + (j-1) * line_size + k * slice_size;
                    const size_t index_ru= i-1 + (j+1) * line_size + k * slice_size;
                    const size_t index_rd= i-1 + (j-1) * line_size + k * slice_size;
                    const size_t index_ll= i-2 +  j     * line_size + k * slice_size;
                    const size_t index_rr= i+2 +  j     * line_size + k * slice_size;
                    const size_t index_uu= i   + (j+2)  * line_size + k * slice_size;
                    const size_t index_dd= i   + (j-2)  * line_size + k * slice_size;

                    const float value_c = m_volume->values().at<float>( index_c );
                    const float value_l = m_volume->values().at<float>( index_l );
                    const float value_r = m_volume->values().at<float>( index_r );
                    const float value_u = m_volume->values().at<float>( index_u );
                    const float value_d = m_volume->values().at<float>( index_d );
                    const float value_lu= m_volume->values().at<float>( index_lu );
                    const float value_ld= m_volume->values().at<float>( index_ld );
                    const float value_ru= m_volume->values().at<float>( index_ru );
                    const float value_rd= m_volume->values().at<float>( index_rd );
                    const float value_ll= m_volume->values().at<float>( index_ll );
                    const float value_rr= m_volume->values().at<float>( index_rr );
                    const float value_uu= m_volume->values().at<float>( index_uu );
                    const float value_dd= m_volume->values().at<float>( index_dd );


                    // Local Max
                    if( (value_c > value_l) &&
                        (value_c > value_r) &&
                        (value_c > value_u) &&
                        (value_c > value_d) &&
                        (value_c > value_lu) &&
                        (value_c > value_ld) &&
                        (value_c > value_ru) &&
                        (value_c > value_rd) &&
                        (value_c > value_ll) &&
                        (value_c > value_rr) &&
                        (value_c > value_uu) &&
                        (value_c > value_dd) )

                    {
                        float v[12] = {
                            value_l, value_r, value_u, value_d,
                            value_lu, value_ld, value_ru, value_rd,
                            value_ll, value_rr, value_uu, value_dd };
                        for( size_t id=0;id<12;id++)v[id]-=value_c;
                        kvs::Vector<float> vec(12,v);
                        //this->update_min_max( value_c );
                        this->update_opacity( value_c, vec.length() );
                    }

                    // Local Min
                    if( (value_c < value_l) &&
                        (value_c < value_r) &&
                        (value_c < value_u) &&
                        (value_c < value_d) &&
                        (value_c < value_lu) &&
                        (value_c < value_ld) &&
                        (value_c < value_ru) &&
                        (value_c < value_rd) &&
                        (value_c < value_ll) &&
                        (value_c < value_rr) &&
                        (value_c < value_uu) &&
                        (value_c < value_dd) )

                    {
                        float v[12] = {
                            value_l, value_r, value_u, value_d,
                            value_lu, value_ld, value_ru, value_rd,
                            value_ll, value_rr, value_uu, value_dd };
                        for( size_t id=0;id<12;id++)v[id]-=value_c;
                        kvs::Vector<float> vec(12,v);
                        //this->update_min_max( value_c );
                        this->update_opacity( value_c, vec.length() );
                    }
                }
            }
            this->normalize_opacity();
        }

    void update_min_max( float extremum )
        {
            static int cnt = 0;
            if( cnt == 0 )
            {
                m_min_extremum = extremum;
                m_max_extremum = extremum;
                cnt++;
            }
            else
            {
                m_min_extremum = m_min_extremum < extremum ? m_min_extremum : extremum;
                m_max_extremum = m_max_extremum > extremum ? m_max_extremum : extremum;
            }
        }

    void update_opacity( float extremum, float grad )
        {
            const size_t resol = m_color_map.resolution();
            kvs::ValueArray<float> triangle( resol );
            const float local_min = m_volume->minValue();
            const float local_max = m_volume->maxValue();

            for( size_t i = 0; i < resol; i++ )
            {
                float x = (local_max-local_min) / (resol-1) * i + local_min;
                triangle[i] = grad * triangle_func( x, extremum, local_min, local_max );
            }

            for( size_t i = 0; i < resol; i++ )
            {
                m_opacity[i] += triangle[i];
            }
        }

    void normalize_opacity( void )
        {
            const size_t resol = m_color_map.resolution();
            float max = 0.0f;
            for( size_t i = 0; i < resol; i++ )
            {
                max = max < m_opacity[i] ? m_opacity[i] : max;
//                 std::cout<<m_opacity[i]<<std::endl;
            }

            for( size_t i = 0; i < resol; i++ )
            {
                m_opacity[i] = 1.0f / max * m_opacity[i];
                if( m_opacity[i] > 0.5f ) m_opacity[i] = 0.5f;
            }
        }

    float triangle_func(float x, float center, float left, float right )
        {
            float l = right - left;
            float width = l / 25.0f;
            float y = 0.0f;

            // zero range
            float zero_range = l / 25.0f;
            if( ( -zero_range < x ) && ( x < zero_range ) ) return 0.0f;

            if( x < center - width ) y = 0.0f;
            else if( (center - width < x) && (x < center) )
            {
                y = 1.0f/width * ( x - center + width );
            }
            else if( (center < x ) && (x < center + width) )
            {
                y = -1.0f/width * ( x - center - width );
            }
            else y = 0.0f;

            return y;
        }
};


void Import( kvs::UnstructuredVolumeObject* volume,
             kvs::test::AVSField* field,
             size_t time_step,
	     int first_step )
{
    if( (int)time_step == first_step )
    {
        const kvs::Vector3ui resol = field->dim();
        const kvs::UInt32 nnodes = resol.x() * resol.y() * resol.z();
        const kvs::Vector3ui elem = resol - kvs::Vector3ui(1);
        const kvs::UInt32 nelem = elem.x() * elem.y() * elem.z();

        //volume->setVeclen( ucd->veclens()[ ucd->componentID() ] );
        volume->setVeclen( 1 );
        volume->setNNodes( nnodes );
        volume->setNCells( nelem );
        volume->setCellType( kvs::UnstructuredVolumeObject::Hexahedra );
        volume->setCoords( field->coords() );
        volume->setValues( kvs::AnyValueArray( field->values( 0 ) ) );
//         volume->setValues( kvs::AnyValueArray( field->values( time_step ) ) );


        // Generate connections
        const kvs::UInt32 line_size = resol.x();
        const kvs::UInt32 slice_size = line_size * resol.y();
        kvs::ValueArray<kvs::UInt32> connections( nelem * 8 );

        kvs::UInt32 m_index[8];
        kvs::UInt32* connec = connections.pointer();

        for( size_t k = 0; k < elem.z(); k++ )
        {
            for( size_t j = 0; j < elem.y(); j++ )
            {
                for( size_t i = 0; i < elem.x(); i++ )
                {
                    m_index[0] = i + j * line_size + k * slice_size;
                    m_index[1] = m_index[0] + 1;
                    m_index[2] = m_index[1] + line_size;
                    m_index[3] = m_index[0] + line_size;
                    m_index[4] = m_index[0] + slice_size;
                    m_index[5] = m_index[1] + slice_size;
                    m_index[6] = m_index[2] + slice_size;
                    m_index[7] = m_index[3] + slice_size;

                    for( size_t p = 0; p < 8; p++)
                    {
                        *connec = m_index[p];
                        connec++;
                    }
                }
            }
        }
        volume->setConnections( connections );
        volume->updateMinMaxCoords();
        volume->updateMinMaxValues();
    }
    else
    {
        volume->setValues( kvs::AnyValueArray( field->values( 0 ) ) );
        volume->updateMinMaxValues();
    }
}

std::string createString( const int step, const std::string filename )
{
    char* buf = new char[256];
    sprintf( buf, "_%04d.kvsml", step );
    std::string str = filename;
    str += buf;
    delete [] buf;
    return( str );
}
/*************/

kvs::PointObject* sampling(
    Argument& param,
    kvs::glut::Screen& screen,
    kvs::VolumeObjectBase* volume,
    size_t subpixel_level,
    float sampling_step)
{
    kvs::PointObject* object;
    kvs::TransferFunction& tf = param.transfer_function;

    /*** debug ***/
    tf.setRange(volume);
    /*************/

    // volume calculate test.
    if (param.test_volume) {
        object = new kvs::test::TestVolume(
            screen.camera(), volume, subpixel_level, sampling_step, tf);
        return object;
    }
#ifdef CPU_SAMPLING_TIME
    std::cout << std::endl << "CPU - ";
#else
    std::cout << std::endl << "GPU - ";
#endif
    switch (param.sampling_method) {
    case 'u':
        std::cout << "Uniform sampling" << std::endl;
        object = new kvs::test::CellByCellUniformSampling(
            screen.camera(), volume, subpixel_level, sampling_step, tf);
        break;
    case 'r':
        std::cout << "Rejection sampling" << std::endl;
        object = new kvs::test::CellByCellRejectionSampling(
            screen.camera(), volume, subpixel_level, sampling_step, tf);
        break;
    case 'm':
        std::cout << "Metropolis sampling" << std::endl;
        object =  new kvs::test::CellByCellMetropolisSampling(
            screen.camera(), volume, subpixel_level, sampling_step, tf);
        break;
    default:
        std::cout << "Unknown sampling method:" << param.sampling_method << std::endl;
        exit(1);
    }

    return object;
}


kvs::UnstructuredVolumeObject* read_avsucd(char* filename)
{
    kvs::test::AVSUcd* ucd = new kvs::test::AVSUcd(filename);

    kvs::UnstructuredVolumeObject* volume = new kvs::UnstructuredVolumeObject();
    volume->setVeclen( ucd->veclens()[ ucd->componentID() ] );
    volume->setNNodes( ucd->nnodes() );
    volume->setNCells( ucd->nelements() );
    volume->setCellType( kvs::UnstructuredVolumeObject::Tetrahedra );
    volume->setCoords( ucd->coords() );
    volume->setConnections( ucd->connections() );
    volume->setValues( kvs::AnyValueArray( ucd->values() ) );
    volume->updateMinMaxCoords();
    volume->updateMinMaxValues();

    delete ucd;

    return volume;
}


/*===========================================================================*/
/**
 *  @brief  Main function.
 *  @param  argc [i] argument counter
 *  @param  argv [i] argument values
 */
/*===========================================================================*/
int main( int argc, char** argv )
{
    int retval = 0;

    Argument param( argc, argv );

    mpi_rank = 0;
    time_series = 0;
    gl_integration_ndiv = 0;
    random_number_method = 0;
    gpuPrefixScan=0;
    gpuMultiStream=0;
    thread_size=256;
    jacobCalc=0;
    //TIMER[10][100];

    int mpi_size = 1;
    int ts_file_no = -1;
    int ts_file_max_no = 0;

#ifdef ENABLE_MPI
    MPI_Init(&argc, &argv);
    MPI_Comm_rank(MPI_COMM_WORLD, &mpi_rank);
    MPI_Comm_size(MPI_COMM_WORLD, &mpi_size);
    if(mpi_rank == 0){
#endif
        kvs::glut::Application app( argc, argv );
        kvs::glut::Screen screen( &app );
	//       Argument param( argc, argv );
        KeyPressEvent key_press_event;
	const float sampling_step = param.sampling_step;
 	const size_t subpixel_level =
            ::GetRevisedSubpixelLevel( param.subpixel_level, param.repeat_level );

        std::cout << "sampling step = " << sampling_step <<
            ", subpixel_lv. = " << subpixel_level << std::endl;

        // setup TinyMT
        if (init_tinymt()) {
            std::cout << "Error. init tinymt failed." << std::endl;
            exit(1);
        }

	kvs::PointObject* object;

	//**** createHexahedraPointObject
 	if (param.hexahedra_dir != "") {
            // hexahedra x16 data.
            const int DATA_NUM = 16;

            size_t total_particles = 0;
            kvs::UnstructuredVolumeObject* volume;
            kvs::PointObject* p_object[DATA_NUM];

            for (int i = 0; i < DATA_NUM; i++) {
                char filename[128];
                sprintf(filename, "P00%02d/7k_P00%02d.kvsml", i+1, i+1);
                std::string filepath = param.hexahedra_dir + '/' + filename;
                std::cout << filepath << std::endl;
		double start = get_time();
                volume = new kvs::UnstructuredVolumeImporter(filepath);
		double end = get_time();
		std::cout << "0:Read file: " << end-start << " ms" << std::endl;
		/*** timer ***/
		TIMER[mpi_rank][0] += end - start;
		/*************/

                std::cout << *volume << std::endl;

 		p_object[i] = sampling(param, screen, volume,
 				       subpixel_level, sampling_step);

                delete volume;
                total_particles += p_object[i]->nvertices();
            }
 	    if (!param.test_volume) {
                printf("total %zu\n", total_particles);
            }

            object = new kvs::test::PointObjectSummation(p_object, DATA_NUM, total_particles);

            for (int i = 0; i < DATA_NUM; i++) {
                delete p_object[i];
            }
        } else if (param.pyramid) {
            //*****  createPyramidPointObject
            // test pyramid
            kvs::UnstructuredVolumeObject* volume;
            GenerateVolumeData gen_vol(128);

            std::cout << "gen pyramid data." << std::endl;
	    double start = get_time();
            gen_vol.pyramid(128);
            volume = gen_vol.m_volume;
	    double end = get_time();
	    std::cout << "0:Read file: " << end-start << " ms" << std::endl;
	    /*** timer ***/
	    TIMER[mpi_rank][0] += end - start;
	    /*************/

            std::cout << *volume << std::endl;

 	    object = sampling(param, screen, volume, subpixel_level, sampling_step);

        } else if (param.quadratic_tetrahedra_dir != "") {
            //****** createQuadraticTetrahedra
            // quadratic tetrahedra
	    double start = get_time();
            kvs::FrontSTRFileReader* reader = new kvs::FrontSTRFileReader(
                param.quadratic_tetrahedra_dir.c_str(),
                0,
                1,
                kvs::FrontSTRFileReader::Skip,
                kvs::UnstructuredVolumeObject::QuadraticTetrahedra);

            kvs::UnstructuredVolumeObject* volume
                = new kvs::DistributedUnstructuredVolumeImporter(reader);

            delete reader;
	    double end = get_time();
	    std::cout << "0:Read file: " << end-start << " ms" << std::endl;
	    /*** timer ***/
	    TIMER[mpi_rank][0] += end - start;
	    /*************/

            std::cout << *volume << std::endl;
 	    object = sampling(param, screen, volume, subpixel_level, sampling_step);

        } else if (param.input_data == "" && param.gt5d == false) {
            //*******  createQuadraticHexahedra
            // test quadratic hexahedra
            kvs::UnstructuredVolumeObject* volume;
            GenerateVolumeData gen_vol(128);

            std::cout << "gen quadratic hexahedra data." << std::endl;
            double start = get_time();
            gen_vol.quadraticHexahedra(128);
            volume = gen_vol.m_volume;
            double end = get_time();
            std::cout << "0:Read file: " << end-start << " ms" << std::endl;
            /*** timer ***/
            TIMER[mpi_rank][0] += end - start;
            /*************/

            std::cout << *volume << std::endl;

            object = sampling(param, screen, volume, subpixel_level, sampling_step);

        }
        else if( param.gt5d == true )
        {
            double gt5_start = get_time();
            std::cout<<"start GT5D sampling\n";
            kvs::TransferFunction& tf = param.transfer_function;

            std::string header   = param.input_data+"/pd3d.fld";
            std::string coord    = param.input_data+"/co3d.dat";
            std::string variable = param.input_data+"/pd3d.dat";
//             std::string header = "/home/uehara/JAEA/H24_PBVR/Data/GT5D/pd3d.fld";
//             std::string coord = "/home/uehara/JAEA/H24_PBVR/Data/GT5D/co3d.dat";
//             std::string variable = "/home/uehara/JAEA/H24_PBVR/Data/GT5D/pd3d.dat";
//             std::string header = "/home/mizuho/PBVR/Data/GT5D/pd3d.fld";
//             std::string coord = "/home/mizuho/PBVR/Data/GT5D/co3d.dat";
//             std::string variable = "/home/mizuho/PBVR/Data/GT5D/pd3d.dat";

            kvs::test::AVSField* field =
                new kvs::test::AVSField( header, coord, variable );

            const kvs::Vector3ui resol = field->dim();

            kvs::UnstructuredVolumeObject* volume = new kvs::UnstructuredVolumeObject;
            SliceAnalysis analysis( tf.colorMap() );
            for( size_t i = 0; i < GT5D_NUM; i++ )
            {
#ifdef ENABLE_MPI
                if((time_series==0)||(mpi_rank==i%mpi_size)){
                    if(time_series==1) MPI_Barrier(MPI_COMM_WORLD);
#endif
                    double start = get_time();
                    while(field->currentTimeStep() != i){
                        field->progress();
                    }
                    field->progress();
                    Import( volume, field, i ,0);
                    double end = get_time();
                    std::cout << mpi_rank << ":";
                    std::cout << "Read file: " << end-start << " ms" << std::endl;
                    /*** timer ***/
                    TIMER[mpi_rank][0] += end - start;
                    /*************/
                    std::cout << mpi_rank << ":";
                    std::cout<<"time step = "<<i<<std::endl;
                    kvs::Timer timer;
                    timer.start();

                    const float min_val = volume->minValue();
                    const float max_val = volume->maxValue();

                    tf.setRange( min_val, max_val );
                    analysis.analyze( resol, volume );
                    kvs::OpacityMap opacity_map( analysis.opacity(), min_val, max_val );

                    kvs::TransferFunction tfunc( tf.colorMap(), opacity_map );

                    kvs::PointObject* pobject;
                    switch (param.sampling_method) {
                    case 'u':
                        std::cout << "Uniform sampling" << std::endl;
                        pobject = new kvs::test::CellByCellUniformSampling(
                            screen.camera(), volume, subpixel_level, sampling_step, tfunc);
                        break;
                    case 'r':
                        std::cout << "Rejection sampling" << std::endl;
                        pobject = new kvs::test::CellByCellRejectionSampling(
                            screen.camera(), volume, subpixel_level, sampling_step, tfunc);
                        break;
                    case 'm':
                        std::cout << "Metropolis sampling" << std::endl;
                        pobject = new kvs::test::CellByCellMetropolisSampling(
                            screen.camera(), volume, subpixel_level, sampling_step, tfunc );
                        break;
                    }
//         std::cout<<"nparticles="<<pobject->nvertices()<<std::endl;
                    timer.stop();
                    std::cout<<mpi_rank << ":" ;
                    std::cout<<"sampling time:"<<timer.msec()<<std::endl<<std::endl;
                    /* output rusult */
//         std::string output = createString(i,"fine");
//         kvs::KVSMLObjectPoint* kvsmlpoint =
//           new kvs::PointExporter<kvs::KVSMLObjectPoint>( pobject );
//         kvsmlpoint->setWritingDataType( kvs::KVSMLObjectPoint::ExternalBinary );
//         kvsmlpoint->write( output );
//         delete kvsmlpoint;
                    delete pobject;
#ifdef ENABLE_MPI
                } // end of if() MPI_Barrier
#endif
            } // end of for(i<GT5FD_NUM)
#ifdef ENABLE_MPI
//          if(time_series==1) MPI_Barrier(MPI_COMM_WORLD);
            int exit_command = MPIController::EXIT_PROCESS;
            MPI_Bcast(&exit_command, 1, MPI_INT, 0, MPI_COMM_WORLD);
            for(int i=1;i<mpi_size;i++){
                MPI_Recv(&TIMER[i][0], 100, MPI_REAL, i, 0, MPI_COMM_WORLD, 
                         MPI_STATUS_IGNORE);
            }
#endif
            double gt5_end = get_time();
            if(time_series==1){
                printf("GPU:time series data: %lf ms\n", gt5_end - gt5_start);
            }
            delete volume;
            field->close();
            delete field;
            /*** timer ***/
            // ALL
            for(int i=0;i<mpi_size;i++){
                TOTAL[i]  = 0.0;
                MAXVAL[i] = 0.0;
            }
            float t00[30],t01[30];
            for(int j=0;j<21;j++){
                t00[j]=0.0;
                t01[j]=0.0;
                for(int i=0;i<mpi_size;i++){
                    t00[j] += TIMER[i][j];
                    t01[j] = std::max(t01[j],TIMER[i][j]);
                    TOTAL[i] += TIMER[i][j];
                    MAXVAL[i] = std::max(MAXVAL[i],TIMER[i][j]);
                }
            }
            std::cout << "ALL," ;
            for(int i=0;i<21;i++){
                std::cout << t00[i]/mpi_size << ",";
            }
            t00[22] = 0.0;
            for(int i=0;i<mpi_size;i++) t00[22] += TOTAL[i];
            std::cout << t00[22]/mpi_size << std::endl;
            std::cout << "ALL," ;
            for(int i=0;i<21;i++){
                std::cout << t01[i] << ",";
            }
            t01[22] = 0.0;
            for(int i=0;i<mpi_size;i++) t01[22] = std::max(TOTAL[i],t01[22]);
            std::cout << t01[22] << std::endl;

            // "PREFIXSCAN"
            for(int i=0;i<mpi_size;i++){
                TOTAL[i]  = 0.0;
                MAXVAL[i] = 0.0;
            }
            for(int j=13;j<17;j++){
                t00[j-13]=0.0;
                t01[j-13]=0.0;
                for(int i=0;i<mpi_size;i++){
                    t00[j-13] += TIMER[i][j];
                    t01[j-13] = std::max(t01[j-13],TIMER[i][j]);
                    TOTAL[i] += TIMER[i][j];
                    MAXVAL[i] = std::max(MAXVAL[i],TIMER[i][j]);
                }
            }
            std::cout << "PREFIXSCAN," ;
            for(int i=13;i<17;i++){
                std::cout << t00[i-13]/mpi_size << ",";
            }
            t00[4] = 0.0;
            for(int i=0;i<mpi_size;i++) t00[4] += TOTAL[i];
            std::cout << t00[4]/mpi_size << std::endl;

            std::cout << "PREFIXSCAN," ;
            for(int i=13;i<17;i++){
                std::cout << t01[i-13] << ",";
            }
            t01[4] = 0.0;
            for(int i=0;i<mpi_size;i++) t01[4] = std::max(TOTAL[i],t01[4]);
            std::cout << t01[4] << std::endl;

            // "DOUBLEBUFF";
            for(int i=0;i<mpi_size;i++){
                TOTAL[i]  = 0.0;
                MAXVAL[i] = 0.0;
            }
            for(int j=17;j<21;j++){
                t00[j-17]=0.0;
                t01[j-17]=0.0;
                for(int i=0;i<mpi_size;i++){
//         std::cout << TIMER[i][j] << ",";
                    t00[j-17] += TIMER[i][j];
                    t01[j-17] = std::max(t01[j-17],TIMER[i][j]);
                    TOTAL[i] += TIMER[i][j];
                    MAXVAL[i] = std::max(MAXVAL[i],TIMER[i][j]);
                }
//           std::cout << std::endl;
            }
            std::cout << "DOUBLEBUFF,";
            for(int i=17;i<21;i++){
                std::cout << t00[i-17]/mpi_size << ",";
            }
            t00[4] = 0.0;
            for(int i=0;i<mpi_size;i++) t00[4] += TOTAL[i];
            std::cout << t00[4]/mpi_size << std::endl;

            std::cout << "DOUBLEBUFF,";
            for(int i=17;i<21;i++){
                std::cout << t01[i-17] << ",";
            }
            t01[4] = 0.0;
            for(int i=0;i<mpi_size;i++) t01[4] = std::max(TOTAL[i],t01[4]);
            std::cout << t01[4] << std::endl;


//         std::cout << "TOTAL," ;
            for(int i=0;i<mpi_size;i++){
                TOTAL[i]  = 0.0;
                MAXVAL[i] = 0.0;
            }
            for(int j=2;j<3;j++){
                t00[j-2]=0.0;
                t01[j-2]=0.0;
                for(int i=0;i<mpi_size;i++){
//         std::cout << TIMER[i][j] << ",";
                    t00[j-2] += TIMER[i][j];
                    t01[j-2] = std::max(t01[j-2],TIMER[i][j]);
                    TOTAL[i] += TIMER[i][j];
                    MAXVAL[i] = std::max(MAXVAL[i],TIMER[i][j]);
                }
//           std::cout << std::endl;
            }
            std::cout << "TOTAL," ;
            for(int i=2;i<3;i++){
                std::cout << t00[i-2]/mpi_size << ",";
            }
            t00[1] = 0.0;
            for(int i=0;i<mpi_size;i++) t00[1] += TOTAL[i];
            std::cout << t00[1]/mpi_size << std::endl;

            std::cout << "TOTAL," ;
            for(int i=2;i<3;i++){
                std::cout << t01[i-2] << ",";
            }
            t01[1] = 0.0;
            for(int i=0;i<mpi_size;i++) t01[1] += std::max(TOTAL[i],t01[1]);
            std::cout << t01[1] << std::endl;

            // particle
//         t00[1] = 0.0;
//         std::cout << "particles," ;
//         for(int i=0;i<mpi_size;i++){
//           t00[1] += TIMER[i][50];
//           std::cout << TIMER[i][50] << ",";
//         }
//         std::cout << t00[1] << std::endl;
            /*************/
            return 0;
        } else {
            // normal input data.
            struct stat st;
            if (stat(param.input_data.c_str(), &st)) {
                std::cout << "Error. read failed:" << param.input_data << std::endl;
                exit(1);
            }

            if (S_ISDIR(st.st_mode)) {
                // For HTTR data.
                char filename[256];
                double ts_start,ts_end;
                if(time_series==0){
                    ts_file_max_no = HTTR_NUM;
                }else{
                    ts_file_max_no = ((HTTR_NUM-1)/mpi_size+1)*mpi_size;
                    ts_start = get_time();
                }
                kvs::test::AVSUcd* ucd = new kvs::test::AVSUcd();
                kvs::UnstructuredVolumeObject* volume = new kvs::UnstructuredVolumeObject();
                kvs::PointObject* obj_list[HTTR_NUM];
                size_t total_particles = 0;
                for (int i = 0; i < ts_file_max_no; i++) {
                    ts_file_no = i;
#ifdef ENABLE_MPI
                    if((time_series==0) || (mpi_rank == ts_file_no%mpi_size)){
#endif
                        /* set file name */
                        int file_no = std::min(HTTR_NUM,i+1); 
                        sprintf(filename, "%s/binary0%03d.dat", param.input_data.c_str(), file_no);
                        double start = get_time();

                        /* read file */
//             volume = read_avsucd(filename);
                        if(i==0){
                            ucd->read(filename,0);
                            volume->setVeclen( ucd->veclens()[ ucd->componentID() ] );
                            volume->setNNodes( ucd->nnodes() );
                            volume->setNCells( ucd->nelements() );
                            volume->setCellType( kvs::UnstructuredVolumeObject::Tetrahedra );
                            volume->setCoords( ucd->coords() );
                            volume->setConnections( ucd->connections() );
                            volume->setValues( kvs::AnyValueArray( ucd->values() ) );
                            volume->updateMinMaxCoords();
                            volume->updateMinMaxValues();
                        }else{
                            ucd->read(filename,-1);
                            volume->setValues( kvs::AnyValueArray( ucd->values() ) );
                            volume->updateMinMaxCoords();
                            volume->updateMinMaxValues();
                        }
                        double read_end = get_time();
#ifdef ENABLE_MPI
                        if(time_series==1) MPI_Barrier(MPI_COMM_WORLD);
#endif
                        std::cout << mpi_rank << ":";
                        std::cout << filename ;
                        std::cout << ":Read file: " << read_end-start << " ms" << std::endl;
                        /*** timer ***/
                        TIMER[mpi_rank][0] += read_end - start;
                        /*************/
                        std::cout << *volume << std::endl;
                        kvs::PointObject* object = sampling(param, screen, volume, subpixel_level, sampling_step);
//              obj_list[i] = sampling(param, screen, volume, subpixel_level, sampling_step);
//                  total_particles += obj_list[i]->nvertices();
                        delete object;
                        double end = get_time();
                        std::cout << "time step " << i << " : " << end-start << " ms" << std::endl;
#ifdef ENABLE_MPI
                    }else{
                    }
#endif
                }
                delete volume;
                delete ucd;

                if(time_series==1){
                    ts_end = get_time();
                    printf("GPU:time series data: %lf ms\n", ts_end - ts_start);
                }else{
//           object = new kvs::test::PointObjectSummation(obj_list, HTTR_NUM, total_particles);
//           for (int i = 0; i < HTTR_NUM; i++) {
//             delete obj_list[i];
//           }
                }
            } else {
                // one file
                double start = get_time();
                kvs::UnstructuredVolumeObject* volume;
                volume = new kvs::UnstructuredVolumeImporter(param.input_data);
                double end = get_time();
                std::cout << ":Read file: " << end-start << " ms" << std::endl;
                /*** timer ***/
                TIMER[mpi_rank][0] += end - start;
                /*************/

                std::cout << *volume << std::endl;

                object = sampling(param, screen, volume, subpixel_level, sampling_step);
                delete volume;
            }
        }
#ifdef ENABLE_MPI
        int exit_command = MPIController::EXIT_PROCESS;
        MPI_Bcast(&exit_command, 1, MPI_INT, 0, MPI_COMM_WORLD);
        /*** timer ***/
        for(int i=1;i<mpi_size;i++){
            MPI_Recv(&TIMER[i][0], 100, MPI_REAL, i, 0, MPI_COMM_WORLD, 
                     MPI_STATUS_IGNORE);
        }
        /*** timer ***/
        // ALL
        for(int i=0;i<mpi_size;i++){
            TOTAL[i]  = 0.0;
            MAXVAL[i] = 0.0;
        }
        float t00[30],t01[30];
        for(int j=0;j<21;j++){
            t00[j]=0.0;
            t01[j]=0.0;
            for(int i=0;i<mpi_size;i++){
                t00[j] += TIMER[i][j];
                t01[j] = std::max(t01[j],TIMER[i][j]);
                TOTAL[i] += TIMER[i][j];
                MAXVAL[i] = std::max(MAXVAL[i],TIMER[i][j]);
            }
        }
        std::cout << "ALL," ;
        for(int i=0;i<21;i++){
            std::cout << t00[i]/mpi_size << ",";
        }
        t00[22] = 0.0;
        for(int i=0;i<mpi_size;i++) t00[22] += TOTAL[i];
        std::cout << t00[22]/mpi_size << std::endl;

        std::cout << "ALL," ;
        for(int i=0;i<21;i++){
            std::cout << t01[i] << ",";
        }
        t01[22] = 0.0;
        for(int i=0;i<mpi_size;i++) t01[22] = std::max(TOTAL[i],t01[22]);
        std::cout << t01[22] << std::endl;

        // "PREFIXSCAN"
        for(int i=0;i<mpi_size;i++){
            TOTAL[i]  = 0.0;
            MAXVAL[i] = 0.0;
        }
        for(int j=13;j<17;j++){
            t00[j-13]=0.0;
            t01[j-13]=0.0;
            for(int i=0;i<mpi_size;i++){
                t00[j-13] += TIMER[i][j];
                t01[j-13] = std::max(t01[j-13],TIMER[i][j]);
                TOTAL[i] += TIMER[i][j];
                MAXVAL[i] = std::max(MAXVAL[i],TIMER[i][j]);
            }
        }
        std::cout << "PREFIXSCAN," ;
        for(int i=13;i<17;i++){
            std::cout << t00[i-13]/mpi_size << ",";
        }
        t00[4] = 0.0;
        for(int i=0;i<mpi_size;i++) t00[4] += TOTAL[i];
        std::cout << t00[4]/mpi_size << std::endl;

        std::cout << "PREFIXSCAN," ;
        for(int i=13;i<17;i++){
            std::cout << t01[i-13] << ",";
        }
        t01[4] = 0.0;
        for(int i=0;i<mpi_size;i++) t01[4] = std::max(TOTAL[i],t01[4]);
        std::cout << t01[4] << std::endl;

        // "DOUBLEBUFF"
        for(int i=0;i<mpi_size;i++){
            TOTAL[i]  = 0.0;
            MAXVAL[i] = 0.0;
        }
        for(int j=17;j<21;j++){
            t00[j-17]=0.0;
            t01[j-17]=0.0;
            for(int i=0;i<mpi_size;i++){
                t00[j-17] += TIMER[i][j];
                t01[j-17] = std::max(t01[j-17],TIMER[i][j]);
                TOTAL[i] += TIMER[i][j];
                MAXVAL[i] = std::max(MAXVAL[i],TIMER[i][j]);
            }
        }
        std::cout << "DOUBLEBUFF,";
        for(int i=17;i<21;i++){
            std::cout << t00[i-17]/mpi_size << ",";
        }
        t00[4] = 0.0;
        for(int i=0;i<mpi_size;i++) t00[4] += TOTAL[i];
        std::cout << t00[4]/mpi_size << std::endl;

        std::cout << "DOUBLEBUFF,";
        for(int i=17;i<21;i++){
            std::cout << t01[i-17] << ",";
        }
        t01[4] = 0.0;
        for(int i=0;i<mpi_size;i++) t01[4] = std::max(TOTAL[i],t01[4]);
        std::cout << t01[4] << std::endl;

        // "TOTAL"
        for(int i=0;i<mpi_size;i++){
            TOTAL[i]  = 0.0;
            MAXVAL[i] = 0.0;
        }
        for(int j=2;j<3;j++){
            t00[j-2]=0.0;
            t01[j-2]=0.0;
            for(int i=0;i<mpi_size;i++){
                //         std::cout << TIMER[i][j] << ",";
                t00[j-2] += TIMER[i][j];
                t01[j-2] = std::max(t01[j-2],TIMER[i][j]);
                TOTAL[i] += TIMER[i][j];
                MAXVAL[i] = std::max(MAXVAL[i],TIMER[i][j]);
            }
            //           std::cout << std::endl;
        }
        std::cout << "TOTAL," ;
        for(int i=2;i<3;i++){
            std::cout << t00[i-2]/mpi_size << ",";
        }
        t00[1] = 0.0;
        for(int i=0;i<mpi_size;i++) t00[1] += TOTAL[i];
        std::cout << t00[1]/mpi_size << std::endl;

        std::cout << "TOTAL," ;
        for(int i=2;i<3;i++){
            std::cout << t01[i-2] << ",";
        }
        t01[1] = 0.0;
        for(int i=0;i<mpi_size;i++) t01[1] += std::max(TOTAL[i],t01[1]);
        std::cout << t01[1] << std::endl;

        // particle
//     t00[1] = 0.0;
//     std::cout << "particles," ;
//     for(int i=0;i<mpi_size;i++){
//       t00[1] += TIMER[i][50];
//       std::cout << TIMER[i][50] << ",";
//     }
//     std::cout << t00[1] << std::endl;
        /*************/
#endif
        // free TinyMT params
        term_tinymt();

        if (param.test_volume) {
            // test end
            return 0;
        }

        printf("total %zu\n", object->nvertices());

        if (param.dump) {
            KVSMLObjectPointWriter p_writer(object, "kvs-gpu");
#ifdef ENABLE_MPI
            MPI_Finalize();
#endif
            return 0;
        }

        size_t sp_lv = param.subpixel_level;

#ifdef REMOTE
        kvs::ParticleVolumeRenderer* renderer =
            new kvs::ParticleVolumeRenderer( object, sp_lv );
#else
        size_t rp_lv = param.repeat_level;
        kvs::glew::ParticleVolumeRenderer* renderer =
            new kvs::glew::ParticleVolumeRenderer( object, sp_lv, rp_lv );
#endif
        if ( !renderer )
        {
            kvsMessageError( "Cannot creat a point renderer.");
        }

        // For shading.
        renderer->enableShading();
        renderer->disableShading();
        renderer->setShader( kvs::Shader::Lambert( 0.5, 0.5 ) );
        renderer->setShader( kvs::Shader::Phong( 0.6, 0.6, 0.8, 15.0 ) );
        renderer->setShader( kvs::Shader::BlinnPhong( 0.1, 0.7, 0.2, 100.0 ) );

        // LOD
#ifndef REMOTE
        renderer->enableLODControl();
#endif
        screen.addKeyPressEvent( &key_press_event );
        screen.registerObject( object, renderer );
        screen.setGeometry( 0, 0, 768, 768 );
        screen.setTitle( "ParticleVolumeRenderer" );
        screen.show();

        retval = app.run();
#ifdef ENABLE_MPI
    } // end of if(mpi_rank == 0 )
    else{ // ELSE (mpi_rank !=0)
#ifdef CPU_SAMPLING_TIME
        if(time_series==1){
            if(param.gt5d){
                kvs::TransferFunction& tf = param.transfer_function;

                std::string header   = param.input_data+"/pd3d.fld";
                std::string coord    = param.input_data+"/co3d.dat";
                std::string variable = param.input_data+"/pd3d.dat";
//             std::string header = "/home/mizuho/PBVR/Data/GT5D/pd3d.fld";
//             std::string coord = "/home/mizuho/PBVR/Data/GT5D/co3d.dat";
//             std::string variable = "/home/mizuho/PBVR/Data/GT5D/pd3d.dat";

                kvs::test::AVSField* field =
                    new kvs::test::AVSField( header, coord, variable );

                const kvs::Vector3ui resol = field->dim();

                kvs::UnstructuredVolumeObject* volume = new kvs::UnstructuredVolumeObject;
                SliceAnalysis analysis( tf.colorMap() );

                ts_file_max_no = ((GT5D_NUM-1)/mpi_size+1) * mpi_size;
                size_t total_particles = 0;
                for (int i = 0; i < ts_file_max_no; i++) {
                    if(mpi_rank == i%mpi_size){
                        MPI_Barrier(MPI_COMM_WORLD);
                        ts_file_no = std::min(GT5D_NUM-1,i); 
                        double start = get_time();
                        while(field->currentTimeStep() != ts_file_no){
                            field->progress();
                        }
                        field->progress();
                        Import( volume, field, ts_file_no ,mpi_rank);
                        double end = get_time();
                        std::cout << mpi_rank << ":";
                        std::cout << "Read file: " << end-start << " ms" << std::endl;
                        /*** timer ***/
                        TIMER[mpi_rank][0] += end - start;
                        /*************/
                        std::cout << mpi_rank << ":";
                        std::cout<<"time step = "<<ts_file_no<<std::endl;
                        kvs::Timer timer;
                        timer.start();

                        const float min_val = volume->minValue();
                        const float max_val = volume->maxValue();

                        tf.setRange( min_val, max_val );
                        analysis.analyze( resol, volume );
                        kvs::OpacityMap opacity_map( analysis.opacity(), min_val, max_val );

                        kvs::TransferFunction tfunc( tf.colorMap(), opacity_map );

                        int sampling_method = 0;
                        MPI_Bcast(&sampling_method, 1, MPI_INT, 0, MPI_COMM_WORLD);
                        if(sampling_method == MPIController::EXIT_PROCESS){
                            break;
                        }
                        switch(sampling_method){
                        case MPIController::UNIFORM_SAMPLING:
                            kvs::test::DistributedUniformSampling::generate_particles_slave_ts_cpu(volume);
                            break;
                        case MPIController::REJECTION_SAMPLING:
                            kvs::test::DistributedRejectionSampling::generate_particles_slave_ts_cpu(volume);
                            break;
                        case MPIController::METROPOLIS_SAMPLING:
                            kvs::test::DistributedMetropolisSampling::generate_particles_slave_ts_cpu(volume);
                            break;
                        }
                        timer.stop();
                        std::cout<<mpi_rank<<":sampling time:"<<timer.msec()<<std::endl<<std::endl;
                    }
                }
            }else{
                // For HTTR data.
                char filename[256];
                ts_file_max_no = ((HTTR_NUM-1)/mpi_size+1) * mpi_size;
                kvs::test::AVSUcd* ucd = new kvs::test::AVSUcd();
                kvs::UnstructuredVolumeObject* volume = new kvs::UnstructuredVolumeObject();
                size_t total_particles = 0;
                for (int i = 0; i < ts_file_max_no; i++) {
                    ts_file_no = i;
                    if(mpi_rank == ts_file_no%mpi_size){
                        int file_no = std::min(HTTR_NUM,i+1); 
                        sprintf(filename, "%s/binary0%03d.dat", param.input_data.c_str(), file_no);
                        double start = get_time();
//            volume = read_avsucd(filename);
                        if(ts_file_no==mpi_rank){
                            ucd->read(filename,0);
                            volume->setVeclen( ucd->veclens()[ ucd->componentID() ] );
                            volume->setNNodes( ucd->nnodes() );
                            volume->setNCells( ucd->nelements() );
                            volume->setCellType( kvs::UnstructuredVolumeObject::Tetrahedra );
                            volume->setCoords( ucd->coords() );
                            volume->setConnections( ucd->connections() );
                            volume->setValues( kvs::AnyValueArray( ucd->values() ) );
                            volume->updateMinMaxCoords();
                            volume->updateMinMaxValues();
                        }else{
                            ucd->read(filename,-1);
                            volume->setValues( kvs::AnyValueArray( ucd->values() ) );
                            volume->updateMinMaxCoords();
                            volume->updateMinMaxValues();
                        }
                        double end = get_time();

                        MPI_Barrier(MPI_COMM_WORLD);
                        std::cout << mpi_rank << ":";
                        std::cout << filename ;
                        std::cout << ":Read file: " << end-start << " ms" << std::endl;
                        /*** timer ***/
                        TIMER[mpi_rank][0] += end - start;
                        /*************/
                        std::cout << *volume << std::endl;

                        int sampling_method = 0;
                        MPI_Bcast(&sampling_method, 1, MPI_INT, 0, MPI_COMM_WORLD);
                        if(sampling_method == MPIController::EXIT_PROCESS){
                            break;
                        }
                        switch(sampling_method){
                        case MPIController::UNIFORM_SAMPLING:
                            kvs::test::DistributedUniformSampling::generate_particles_slave_ts_cpu(volume);
                            break;
                        case MPIController::REJECTION_SAMPLING:
                            kvs::test::DistributedRejectionSampling::generate_particles_slave_ts_cpu(volume);
                            break;
                        case MPIController::METROPOLIS_SAMPLING:
                            kvs::test::DistributedMetropolisSampling::generate_particles_slave_ts_cpu(volume);
                            break;
                        }
                    }
                }
                delete volume;
                delete ucd;
            }
        }else{
            for(;;){
                int sampling_method = 0;
                MPI_Bcast(&sampling_method, 1, MPI_INT, 0, MPI_COMM_WORLD);
                if(sampling_method == MPIController::EXIT_PROCESS){
                    break;
                }
                switch(sampling_method){
                case MPIController::UNIFORM_SAMPLING:
                    kvs::test::DistributedUniformSampling::generate_particles_slave_cpu();
                    break;
                case MPIController::REJECTION_SAMPLING:
                    kvs::test::DistributedRejectionSampling::generate_particles_slave_cpu();
                    break;
                case MPIController::METROPOLIS_SAMPLING:
                    kvs::test::DistributedMetropolisSampling::generate_particles_slave_cpu();
                    break;
                }
            }
        }
#else // end of ifdef CPU_SAMPLING
        // setup TinyMT
        init_tinymt();
        if(time_series==1){
            if(param.gt5d){
                kvs::TransferFunction& tf = param.transfer_function;

                std::string header   = param.input_data+"/pd3d.fld";
                std::string coord    = param.input_data+"/co3d.dat";
                std::string variable = param.input_data+"/pd3d.dat";
//             std::string header = "/home/mizuho/PBVR/Data/GT5D/pd3d.fld";
//             std::string coord = "/home/mizuho/PBVR/Data/GT5D/co3d.dat";
//             std::string variable = "/home/mizuho/PBVR/Data/GT5D/pd3d.dat";

                kvs::test::AVSField* field =
                    new kvs::test::AVSField( header, coord, variable );

                const kvs::Vector3ui resol = field->dim();

                kvs::UnstructuredVolumeObject* volume = new kvs::UnstructuredVolumeObject;
                SliceAnalysis analysis( tf.colorMap() );

                ts_file_max_no = ((GT5D_NUM-1)/mpi_size+1) * mpi_size;
                size_t total_particles = 0;
                for (int i = 0; i < ts_file_max_no; i++) {
                    if(mpi_rank == i%mpi_size){
                        MPI_Barrier(MPI_COMM_WORLD);
                        ts_file_no = std::min(GT5D_NUM-1,i); 
                        double start = get_time();
                        while(field->currentTimeStep() != ts_file_no){
                            field->progress();
                        }
                        field->progress();
                        Import( volume, field, ts_file_no, mpi_rank);
                        double end = get_time();
                        std::cout << mpi_rank << ":";
                        std::cout << "Read file: " << end-start << " ms" << std::endl;
                        /*** timer ***/
                        TIMER[mpi_rank][0] += end - start;
                        /*************/
                        std::cout << mpi_rank << ":";
                        std::cout<<"time step = "<<ts_file_no<<std::endl;
                        kvs::Timer timer;
                        timer.start();

                        const float min_val = volume->minValue();
                        const float max_val = volume->maxValue();

                        tf.setRange( min_val, max_val );
                        analysis.analyze( resol, volume );
                        kvs::OpacityMap opacity_map( analysis.opacity(), min_val, max_val );

                        kvs::TransferFunction tfunc( tf.colorMap(), opacity_map );

                        int sampling_method = 0;
                        MPI_Bcast(&sampling_method, 1, MPI_INT, 0, MPI_COMM_WORLD);
                        if(sampling_method == MPIController::EXIT_PROCESS){
                            break;
                        }
                        switch(sampling_method){
                        case MPIController::UNIFORM_SAMPLING:
                            kvs::test::DistributedUniformSampling::generate_particles_slave_ts_gpu(volume);
                            break;
                        case MPIController::REJECTION_SAMPLING:
                            kvs::test::DistributedRejectionSampling::generate_particles_slave_ts_gpu(volume);
                            break;
                        case MPIController::METROPOLIS_SAMPLING:
                            kvs::test::DistributedMetropolisSampling::generate_particles_slave_ts_gpu(volume);
                            break;
                        }
                        timer.stop();
                        std::cout<<mpi_rank<<":sampling time:"<<timer.msec()<<std::endl<<std::endl;
                    }
                }
            }else{
                // For HTTR data.
                char filename[256];
                ts_file_max_no = ((HTTR_NUM-1)/mpi_size+1) * mpi_size;
                kvs::test::AVSUcd* ucd = new kvs::test::AVSUcd();
                kvs::UnstructuredVolumeObject* volume = new kvs::UnstructuredVolumeObject();
                size_t total_particles = 0;
                for (int i = 0; i < ts_file_max_no; i++) {
                    ts_file_no = i;
                    if(mpi_rank == ts_file_no%mpi_size){
                        int file_no = std::min(HTTR_NUM,i+1); 
                        sprintf(filename, "%s/binary0%03d.dat", param.input_data.c_str(), file_no);
                        double start = get_time();
//            volume = read_avsucd(filename);
                        if(ts_file_no==mpi_rank){
                            ucd->read(filename,0);
                            volume->setVeclen( ucd->veclens()[ ucd->componentID() ] );
                            volume->setNNodes( ucd->nnodes() );
                            volume->setNCells( ucd->nelements() );
                            volume->setCellType( kvs::UnstructuredVolumeObject::Tetrahedra );
                            volume->setCoords( ucd->coords() );
                            volume->setConnections( ucd->connections() );
                            volume->setValues( kvs::AnyValueArray( ucd->values() ) );
                            volume->updateMinMaxCoords();
                            volume->updateMinMaxValues();
                        }else{
                            ucd->read(filename,-1);
                            volume->setValues( kvs::AnyValueArray( ucd->values() ) );
                            volume->updateMinMaxCoords();
                            volume->updateMinMaxValues();
                        }
                        double end = get_time();

                        MPI_Barrier(MPI_COMM_WORLD);
                        std::cout << mpi_rank << ":";
                        std::cout << filename ;
                        std::cout << ":Read file: " << end-start << " ms" << std::endl;
                        /*** timer ***/
                        TIMER[mpi_rank][0] += end - start;
                        /*************/
                        std::cout << *volume << std::endl;

                        int sampling_method = 0;
                        MPI_Bcast(&sampling_method, 1, MPI_INT, 0, MPI_COMM_WORLD);
                        if(sampling_method == MPIController::EXIT_PROCESS){
                            break;
                        }
                        switch(sampling_method){
                        case MPIController::UNIFORM_SAMPLING:
                            kvs::test::DistributedUniformSampling::generate_particles_slave_ts_gpu(volume);
                            break;
                        case MPIController::REJECTION_SAMPLING:
                            kvs::test::DistributedRejectionSampling::generate_particles_slave_ts_gpu(volume);
                            break;
                        case MPIController::METROPOLIS_SAMPLING:
                            kvs::test::DistributedMetropolisSampling::generate_particles_slave_ts_gpu(volume);
                            break;
                        }
                    }
                }
                delete volume;
                delete ucd;
            }
        }else{
            for(;;){
                int sampling_method = 0;
                MPI_Bcast(&sampling_method, 1, MPI_INT, 0, MPI_COMM_WORLD);
                if(sampling_method == MPIController::EXIT_PROCESS){
                    break;
                }
                switch(sampling_method){
                case MPIController::UNIFORM_SAMPLING:
                    kvs::test::DistributedUniformSampling::generate_particles_slave_gpu();
                    break;
                case MPIController::REJECTION_SAMPLING:
                    kvs::test::DistributedRejectionSampling::generate_particles_slave_gpu();
                    break;
                case MPIController::METROPOLIS_SAMPLING:
                    kvs::test::DistributedMetropolisSampling::generate_particles_slave_gpu();
                    break;
                }
            }
            MPI_Send(&TIMER[mpi_rank][0], 100, MPI_REAL, 0, 0, MPI_COMM_WORLD); 
        }
        // free TinyMT params
        term_tinymt();
#endif
    }
    MPI_Finalize();
#endif
    return retval;
}
