/****************************************************************************/
/**
 *  @file AVSUcd.cpp
 */
/*----------------------------------------------------------------------------
 *
 *  Copyright (c) Visualization Laboratory, Kyoto University.
 *  All rights reserved.
 *  See http://www.viz.media.kyoto-u.ac.jp/kvs/copyright/ for details.
 *
 *  $Id: AVSUcd.cpp 656 2010-12-13 16:04:04Z naohisa.sakamoto $
 */
/****************************************************************************/
#include "AVSUcd.h"
#include "File.h"
#include "Message.h"
#include "ValueArray.h"
#include "IgnoreUnusedVariable.h"
#include <cstdlib>
#include <cstring>
#include <limits>

namespace
{

const size_t MaxLineLength = 512;

const char* const Delimiter = " \n\r";

const std::string CycleTypeToString[ kvs::AVSUcd::CycleTypeSize ] =
{
    "unknown",
    "data",
    "geometry",
    "data_geometry"
};

const std::string ElementTypeToString[ kvs::AVSUcd::ElementTypeSize ] =
{
    "unknown",
    "point",
    "line",
    "triangle",
    "quadrangle",
    "tetrahedra",
    "tetrahedra2",
    "pyramid",
    "prism",
    "hexahedra",
    "hexahedra2"
};

}

namespace
{

const kvs::AVSUcd::FormatType CheckFormatType( FILE* const ifs )
{
    char buffer[ ::MaxLineLength ];

    kvs::AVSUcd::FormatType format_type = kvs::AVSUcd::SingleStep;

    while ( fgets( buffer, ::MaxLineLength, ifs ) != 0 )
    {
        if ( buffer[0] == '#' )
        {
            // Skip comment line.
            continue;
        }
        else
        {
            // Skip first token.
            strtok( buffer, ::Delimiter );

            const char* const second_token = strtok( 0, ::Delimiter );

            if ( second_token != 0 )
            {
                format_type = kvs::AVSUcd::SingleStep;
            }
            else
            {
                format_type = kvs::AVSUcd::MultiStep;
            }

            break;
        }
    }

    // Go back file-pointer to head.
    //fseek( ifs, 0, SEEK_SET );

    return( format_type );
}

}


namespace kvs
{

/*==========================================================================*/
/**
 *  Constructs a new empty AVSUcd.
 */
/*==========================================================================*/
AVSUcd::AVSUcd( void )
    : FileFormatBase()
    , m_nsteps( 0 )
    , m_cycle_type( kvs::AVSUcd::CycleTypeUnknown )
    , m_element_type( kvs::AVSUcd::ElementTypeUnknown )
    , m_step_id( 0 )
    , m_step_comment( "" )
    , m_nnodes( 0 )
    , m_nelements( 0 )
    , m_nvalues_per_node( 0 )
    , m_ncomponents_per_node( 0 )
    , m_component_id( 0 )
    , nheader( 0 )
{
}

/*==========================================================================*/
/**
 *  Constructs a new AVSUcd and reads the component of the step in the file.
 *
 *  @param filename     [in] Filename.
 *  @param step_id      [in] StepID.
 *  @param component_id [in] ComponentID.
 */
/*==========================================================================*/
AVSUcd::AVSUcd(
    const std::string& filename,
    const size_t       step_id,
    const size_t       component_id )
    : FileFormatBase()
    , m_nsteps( 0 )
    , m_cycle_type( kvs::AVSUcd::CycleTypeUnknown )
    , m_element_type( kvs::AVSUcd::ElementTypeUnknown )
    , m_step_id( step_id )
    , m_step_comment( "" )
    , m_nnodes( 0 )
    , m_nelements( 0 )
    , m_nvalues_per_node( 0 )
    , m_ncomponents_per_node( 0 )
    , m_component_id( component_id )
    , nheader( 0 )
{
    if( this->read( filename ) )
    {
        m_is_success = true;
    }
    else
    {
        m_is_success = false;
    }
}

/*==========================================================================*/
/**
 *  Destroys the AVSUcd.
 */
/*==========================================================================*/
AVSUcd::~AVSUcd( void )
{
}

const size_t AVSUcd::nsteps( void ) const
{
    return( m_nsteps );
}

const AVSUcd::CycleType AVSUcd::cycleType( void ) const
{
    return( m_cycle_type );
}

const AVSUcd::ElementType AVSUcd::elementType( void ) const
{
    return( m_element_type );
}

const size_t AVSUcd::stepID( void ) const
{
    return( m_step_id );
}

const std::string& AVSUcd::stepComment( void ) const
{
    return( m_step_comment );
}

const size_t AVSUcd::nnodes( void ) const
{
    return( m_nnodes );
}

const size_t AVSUcd::nelements( void ) const
{
    return( m_nelements );
}

const size_t AVSUcd::nvaluesPerNode( void ) const
{
    return( m_nvalues_per_node );
}

const size_t AVSUcd::ncomponentsPerNode( void ) const
{
    return( m_ncomponents_per_node );
}

const std::vector<size_t>& AVSUcd::veclens( void ) const
{
    return( m_veclens );
}

const std::vector<std::string>& AVSUcd::componentNames( void ) const
{
    return( m_component_names );
}

const std::vector<std::string>& AVSUcd::componentUnits( void ) const
{
    return( m_component_units );
}

const size_t AVSUcd::componentID( void ) const
{
    return( m_component_id );
}

const std::string& AVSUcd::componentName( void ) const
{
    return( m_component_names[ m_component_id ] );
}

const std::string& AVSUcd::componentUnit( void ) const
{
    return( m_component_units[ m_component_id ] );
}

const AVSUcd::Coords& AVSUcd::coords( void ) const
{
    return( m_coords );
}

const AVSUcd::Connections& AVSUcd::connections( void ) const
{
    return( m_connections );
}

const AVSUcd::Values& AVSUcd::values( void ) const
{
    return( m_values );
}

void AVSUcd::setNSteps( const size_t nsteps )
{
    m_nsteps = nsteps;
}

void AVSUcd::setCycleType( const CycleType cycle_type )
{
    m_cycle_type = cycle_type;
}

void AVSUcd::setElementType( const ElementType element_type )
{
    m_element_type = element_type;
}

void AVSUcd::setStepID( const size_t step_id )
{
    m_step_id = step_id;
}

void AVSUcd::setStepComment( const std::string& step_comment )
{
    m_step_comment = step_comment;
}

void AVSUcd::setNNodes( const size_t nnodes )
{
    m_nnodes = nnodes;
}

void AVSUcd::setNElements( const size_t nelements )
{
    m_nelements = nelements;
}

void AVSUcd::setNValuesPerNode( const size_t nvalues_per_node )
{
    m_nvalues_per_node = nvalues_per_node;
}

void AVSUcd::setNComponentsPerNode( const size_t ncomponents_per_node )
{
    m_ncomponents_per_node = ncomponents_per_node;
}

void AVSUcd::setVeclens( const std::vector<size_t>& veclens )
{
    m_veclens = veclens;
}

void AVSUcd::setComponentNames( const std::vector<std::string>& component_names )
{
    m_component_names.clear();

    const size_t size = component_names.size();
    for ( size_t i = 0; i < size; ++i )
    {
        m_component_names.push_back( component_names[ i ] );
    }
}

void AVSUcd::setComponentUnits( const std::vector<std::string>& component_units )
{
    m_component_units.clear();

    const size_t size = component_units.size();
    for ( size_t i = 0; i < size; ++i )
    {
        m_component_units.push_back( component_units[ i ] );
    }
}

void AVSUcd::setComponentID( const size_t component_id )
{
    m_component_id = component_id;
}

void AVSUcd::setCoords( const Coords& coords )
{
    m_coords = coords;
}

void AVSUcd::setConnections( const Connections& connections )
{
    m_connections = connections;
}

void AVSUcd::setValues( const Values& values )
{
    m_values = values;
}

/*==========================================================================*/
/**
 *  Reads AVS UCD data from the file.
 *
 *  @param filename [in] Filename.
 */
/*==========================================================================*/
const bool AVSUcd::read( const std::string& filename )
{
    m_filename = filename;

    FILE* const ifs = fopen( filename.c_str(), "rb" );

    if ( !ifs )
    {
        kvsMessageError( "Cannot open %s.", filename.c_str() );
        m_is_success = false;
        return( m_is_success );
    }

    try
    {
        //this->read_control_file( ifs );
        this->read_binary( ifs );

        m_is_success = true;
    }
    catch ( const char* const error )
    {
        kvsMessageError( "%s: %s", filename.c_str(), error );

        m_is_success = false;
    }

    fclose( ifs );

    return( m_is_success );
}

const bool AVSUcd::read( const std::string& filename, int  read_header )
{
    m_filename = filename;

    FILE* const ifs = fopen( filename.c_str(), "rb" );

    if ( !ifs )
    {
        kvsMessageError( "Cannot open %s.", filename.c_str() );
        m_is_success = false;
        return( m_is_success );
    }

    try
    {
        //this->read_control_file( ifs );
        this->read_binary( ifs, read_header );

        m_is_success = true;
    }
    catch ( const char* const error )
    {
        kvsMessageError( "%s: %s", filename.c_str(), error );

        m_is_success = false;
    }

    fclose( ifs );

    return( m_is_success );
}


/*==========================================================================*/
/**
 *  Writes AVS UCD data to the file.
 *
 *  @param filename [in] Filename.
 */
/*==========================================================================*/
const bool AVSUcd::write( const std::string& filename )
{
    m_filename = filename;

    FILE* const ofs = fopen( filename.c_str(), "wb" );

    if ( !ofs )
    {
        kvsMessageError( "Cannot open %s.", filename.c_str() );
        m_is_success = false;
        return( m_is_success );
    }

    try
    {
        this->write_single_step_format( ofs );

        m_is_success = true;
    }
    catch ( const char* const error )
    {
        kvsMessageError( "%s: %s", filename.c_str(), error );

        m_is_success = false;
    }

    fclose( ofs );

    return( m_is_success );
}

std::ostream& operator <<( std::ostream& os, const AVSUcd& rhs )
{
    os << "Filename                      : " << rhs.m_filename                            << std::endl;
    os << "Number of steps               : " << rhs.m_nsteps                              << std::endl;
    os << "Cycle type                    : " << ::CycleTypeToString[rhs.m_cycle_type]     << std::endl;
    os << "Element type                  : " << ::ElementTypeToString[rhs.m_element_type] << std::endl;

    os << "Step ID                       : " << rhs.m_step_id      << std::endl;
    os << "Step comment                  : " << rhs.m_step_comment << std::endl;

    os << "Number of nodes               : " << rhs.m_nnodes               << std::endl;
    os << "Number of elements            : " << rhs.m_nelements            << std::endl;
    os << "Number of values per node     : " << rhs.m_nvalues_per_node     << std::endl;
    os << "Number of components per node : " << rhs.m_ncomponents_per_node << std::endl;

    os << "Veclens of each component     : ";
    for ( size_t i = 0; i < rhs.m_ncomponents_per_node; ++i )
    {
        os << rhs.m_veclens[ i ] << " ";
    }
    os << std::endl;

    os << "Names of each component       : ";
    for ( size_t i = 0; i < rhs.m_ncomponents_per_node; ++i )
    {
        os << rhs.m_component_names[ i ] << " ";
    }
    os << std::endl;

    os << "Units of each component       : ";
    for ( size_t i = 0; i < rhs.m_ncomponents_per_node; ++i )
    {
        os << rhs.m_component_units[ i ] << " ";
    }
    os << std::endl;

    os << "Component ID                  : " << rhs.m_component_id << std::endl;

    return( os );
}

#if 0
void AVSUcd::read_control_file( FILE* const ifs )
{
    char buffer[ ::MaxLineLength ];

    while ( fgets( buffer, ::MaxLineLength, ifs ) != 0 )
    {
        if ( buffer[0] == '#' )
        {
            // Skip comment line.
            continue;
        }
        else
        {
            const char* cycle_type = strtok( buffer, ::Delimiter );
            m_cycle_type =
                strcmp( cycle_type, "data" ) ?
                Data : strcmp( cycle_type, "geom" ) ?
                Geom : strcmp( cycle_type, "data_geom" ) ?
                DataGeom : CycleTypeUnknown;
        }
    }

    const char* file_pass = this->read_file_pass( ifs );

    FILE* ifs_data = fopen( file_pass, "rb" );

    if ( !ifs_data )
    {
        kvsMessageError( "Cannot open %s.", file_pass );
    }

    this->read_binary( ifs_data );

    fclose( ifs_data );
}
#endif

#if 0
const char* AVSUcd::read_file_pass( FILE* const ifs )
{
    char buffer[ ::MaxLineLength ];

    while ( fgets( buffer, ::MaxLineLength, ifs ) != 0 )
    {
        if ( buffer[0] == '#' )
        {
            // Skip comment line.
            continue;
        }
        else
        {
            const char* file_pass = strtok( buffer, ::Delimiter );

            return file_pass;
        }
    }

    std::cout<<"No file pass\n";
    return buffer;
}
#endif

void AVSUcd::read_binary( FILE* const ifs , int read_header )
{
    int header;
    int footer;

    //Head information


    // check first byte size.
    fread( &header, 4, 1, ifs );
    int filetype;
    if ( header == 7 )
    {
        filetype=1;
    }
    else
    {
        filetype=0;
    }
    rewind( ifs );


    if( read_header==0 )
    {
        nheader = 0;

        // keyword
        char keyword[8];
        if( filetype==1 )fread( &header, 4, 1, ifs );
        fread( keyword, 7, 1, ifs );
        if( filetype==1 )fread( &footer, 4, 1, ifs );
        nheader += 7 + filetype*4*2;

        // version
        float version = 0.0f;
        if( filetype==1 )fread( &header, 4, 1, ifs );
        fread( &version, 4, 1, ifs );
        if( filetype==1 )fread( &footer, 4, 1, ifs );
        nheader += 4*1 + filetype*4*2;

        // title
        char title[70];
        if( filetype==1 )fread( &header, 4, 1, ifs );
        fread( title, 70, 1, ifs );
        if( filetype==1 )fread( &footer, 4, 1, ifs );
        nheader += 70*1 + filetype*4*2;

        // step num.
        int stepn = 0;
        if( filetype==1 )fread( &header, 4, 1, ifs );
        fread( &stepn, 4, 1, ifs );
        if( filetype==1 )fread( &footer, 4, 1, ifs );
        nheader += 4*1 + filetype*4*2;

        // step time
        float stept = 0.0f;
        if( filetype==1 )fread( &header, 4, 1, ifs );
        fread( &stept, 4, 1, ifs );
        if( filetype==1 )fread( &footer, 4, 1, ifs );
        nheader += 4*1 + filetype*4*2;

        //Node information

        // num. nodes
        long int nnodes = 0;
        if( filetype==1 )fread( &header, 4, 1, ifs );
        fread( &nnodes, 4, 1, ifs );
        if( filetype==1 )fread( &footer, 4, 1, ifs );
        nheader += 4*1 + filetype*4*2;

        m_nnodes = ( size_t )nnodes;

        // description type should be 1.
        int desc_type = 0;
        if( filetype==1 )fread( &header, 4, 1, ifs );
        fread( &desc_type, 4, 1, ifs );
        if( filetype==1 )fread( &footer, 4, 1, ifs );
        nheader += 4*1 + filetype*4*2;

        // coordinates
        int id = 0;
        float coord[3];
        m_coords.allocate( m_nnodes * 3 );
        size_t i3 = 0;
        for( size_t i = 0; i < m_nnodes; i++, i3 += 3 )
        {
            if( filetype==1 )fread( &header, 4, 1, ifs );
            fread( &id, 4, 1, ifs );
            fread( coord, 4, 3, ifs );
            if( filetype==1 )fread( &footer, 4, 1, ifs );
            m_coords[ i3     ] = coord[0];
            m_coords[ i3 + 1 ] = coord[1];
            m_coords[ i3 + 2 ] = coord[2];
        }
        nheader += m_nnodes*( 4*4 + filetype*4*2 );

        //Cell information

        // num. element
        unsigned int nelements = 0;
        if( filetype==1 )fread( &header, 4, 1, ifs );
        fread( &nelements, 4, 1, ifs );
        if( filetype==1 )fread( &footer, 4, 1, ifs );
        nheader += 4*1 + filetype*4*2;

        m_nelements = ( size_t )nelements;

        // element num.
        if( filetype==1 )fread( &header, 4, 1, ifs );
        fseek( ifs, m_nelements * 4, SEEK_CUR );
        if( filetype==1 )fread( &footer, 4, 1, ifs );
        nheader += 4*m_nelements + filetype*4*2;

        // material num.
        if( filetype==1 )fread( &header, 4, 1, ifs );
        fseek( ifs, m_nelements * 4, SEEK_CUR );
        if( filetype==1 )fread( &footer, 4, 1, ifs );
        nheader += 4*m_nelements + filetype*4*2;

        // element type
        char element_type;
        if( filetype==1 )fread( &header, 4, 1, ifs );
        fread( &element_type, 1, 1, ifs );
        fseek( ifs, m_nelements - 1, SEEK_CUR );
        if( filetype==1 )fread( &footer, 4, 1, ifs );
        nheader += m_nelements + filetype*4*2;

        m_element_type =
            element_type == 4  ? Tetrahedra:
            element_type == 11 ? Tetrahedra2:
            element_type == 7  ? Hexahedra:
            element_type == 14 ? Hexahedra2: ElementTypeUnknown;


        // connection
        int nelem_com = 0;
        if ( m_element_type == Tetrahedra )
        {
            nelem_com = 4;
        }
        else if ( m_element_type == Tetrahedra2 )
        {
            nelem_com = 10;
        }
        else if ( m_element_type == Hexahedra )
        {
            nelem_com = 8;
        }
        else if ( m_element_type == Hexahedra2 )
        {
            nelem_com = 20;
        }
        else
        {
            std::cout<<"Unknown element type\n";
        }

        m_connections.allocate( nelem_com * m_nelements );

        if( filetype==1 )fread( &header, 4, 1, ifs );
        fread( m_connections.pointer(), 4, nelem_com * m_nelements , ifs );
        if( filetype==1 )fread( &footer, 4, 1, ifs );
        nheader += 4*nelem_com*m_nelements + filetype*4*2;

        /// adjust connection by minus 1 ///
        for( size_t i = 0; i < nelem_com * m_nelements; i++ ) m_connections[i]--;

        //Data information

        // num. component
        unsigned int ncomponents;
        if( filetype==1 )fread( &header, 4, 1, ifs );
        fread( &ncomponents, 4, 1, ifs );
        if( filetype==1 )fread( &footer, 4, 1, ifs );
        nheader += 4 + filetype*4*2;
        m_ncomponents = ncomponents;

        // data type
        int data_type = 0;
        if( filetype==1 )fread( &header, 4, 1, ifs );
        fread( &data_type, 4, 1, ifs );
        if( filetype==1 )fread( &footer, 4, 1, ifs );
        nheader += 4 + filetype*4*2;

        // data header
        for( size_t i = 0; i < m_ncomponents; i++ )
        {
            char component_name[16];
            char unit[16];
            int veclen = 0;
            int null_data_flag = 0;
            float null_data = 0.0f;

            if( filetype==1 )fread( &header, 4, 1, ifs );
            fread( component_name, 16, 1, ifs );
            fread( unit, 16, 1, ifs );
            fread( &veclen, 4, 1, ifs );
            fread( &null_data_flag, 4, 1, ifs );
            fread( &null_data, 4, 1, ifs );
            if( filetype==1 )fread( &footer, 4, 1, ifs );
            nheader += ( 16+16+4+4+4 ) + filetype*4*2;

            //m_component_names[i] = component_name;
            m_component_names.push_back( component_name );
            //m_component_units[i] = unit;
            m_component_units.push_back( unit );
            //m_veclens[i] = static_cast<size_t>( veclen );
            m_veclens.push_back( static_cast<size_t>( veclen ) );

            //std::cout<<"component_name:"<<std::string(component_name)<<std::endl;
            /*
                std::cout<<"component_name:";
                for (int c = 0; c < 16; c++) {
                    std::cout << component_name[c];
                }
                std::cout << std::endl;
                //std::cout<<"unit:"<<std::string(unit)<<std::endl;
                std::cout<<"unit:";
                for (int c = 0; c < 16; c++) {
                    std::cout << unit[c];
                }
                std::cout << std::endl;
                std::cout<<"veclen:"<<veclen<<std::endl;
            */
        }

    }
    else
    {
        fseek( ifs, nheader, SEEK_CUR );
    }

    // value
    const size_t veclen = m_veclens[ m_component_id ];
    if( read_header==0 )
    {
        m_values.allocate( veclen * m_nnodes );
    }

    size_t nskip_front = 0;
    size_t nskip_rear = 0;
    for( size_t i = 0; i < m_ncomponents; i++ )
    {
        if( i < m_component_id )
        {
            nskip_front += m_veclens[i];
        }
        else if( i > m_component_id )
        {
            nskip_rear += m_veclens[i];
        }
    }

    float* value = new float[ veclen ];
//     kvs::Real32* p_val = m_values.pointer();

    for( size_t i = 0; i < m_nnodes; i++ )
    {
        if( filetype==1 )fread( &header, 4, 1, ifs );
        fseek( ifs, 4 * nskip_front, SEEK_CUR );
        fread( value, 4, veclen, ifs );
        fseek( ifs, 4 * nskip_rear, SEEK_CUR );
        if( filetype==1 )fread( &footer, 4, 1, ifs );

        for( size_t j = 0; j < veclen; j++ )
        {
            m_values.at( i*veclen+j ) = value[j];
//             *p_val = value[j];
//             p_val++;
        }
    }

    delete[] value;
    /*************/
}

void AVSUcd::read_binary( FILE* const ifs )
{
    read_binary( ifs , 0 );
}


void AVSUcd::write_single_step_format( FILE* const ofs ) const
{
    fprintf( ofs, "%u %u %u 0 0\n",
             static_cast<unsigned int>( m_nnodes ),
             static_cast<unsigned int>( m_nelements ),
             static_cast<unsigned int>( m_nvalues_per_node ) );

    this->write_coords( ofs );
    this->write_connections( ofs );
    this->write_components( ofs );
    this->write_values( ofs );
}

void AVSUcd::write_coords( FILE* const ofs ) const
{
    size_t node_id = 1;

    const kvs::Real32*       coord = m_coords.pointer();
    const kvs::Real32* const end   = coord + m_coords.size();

    while ( coord < end )
    {
        const kvs::Real32 x = *( coord++ );
        const kvs::Real32 y = *( coord++ );
        const kvs::Real32 z = *( coord++ );

        fprintf( ofs, "%u %f %f %f\n",
                 static_cast<unsigned int>( node_id++ ),
                 x,
                 y,
                 z );
    }
}

void AVSUcd::write_connections( FILE* const ofs ) const
{
    size_t element_id = 1;

    const kvs::UInt32*       connection = m_connections.pointer();
    const kvs::UInt32* const end        = connection + m_connections.size();

    if ( m_element_type == Tetrahedra )
    {
        while ( connection < end )
        {
            const kvs::UInt32 node_id0 = *( connection++ );
            const kvs::UInt32 node_id1 = *( connection++ );
            const kvs::UInt32 node_id2 = *( connection++ );
            const kvs::UInt32 node_id3 = *( connection++ );

            fprintf( ofs, "%u 0 tet %u %u %u %u\n",
                     static_cast<unsigned int>( element_id++ ),
                     node_id0,
                     node_id1,
                     node_id2,
                     node_id3 );
        }
    }
    else if ( m_element_type == Tetrahedra2 )
    {
        while ( connection < end )
        {
            const kvs::UInt32 node_id0  = *( connection++ );
            const kvs::UInt32 node_id1  = *( connection++ );
            const kvs::UInt32 node_id2  = *( connection++ );
            const kvs::UInt32 node_id3  = *( connection++ );
            const kvs::UInt32 node_id4  = *( connection++ );
            const kvs::UInt32 node_id5  = *( connection++ );
            const kvs::UInt32 node_id6  = *( connection++ );
            const kvs::UInt32 node_id7  = *( connection++ );
            const kvs::UInt32 node_id8  = *( connection++ );
            const kvs::UInt32 node_id9  = *( connection++ );

            fprintf( ofs, "%u 0 tet %u %u %u %u %u %u %u %u %u %u\n",
                     static_cast<unsigned int>( element_id++ ),
                     node_id0,
                     node_id1,
                     node_id2,
                     node_id3,
                     node_id4,
                     node_id5,
                     node_id6,
                     node_id7,
                     node_id8,
                     node_id9 );
        }
    }
    else if ( m_element_type == Hexahedra )
    {
        while ( connection < end )
        {
            const kvs::UInt32 node_id0 = *( connection++ );
            const kvs::UInt32 node_id1 = *( connection++ );
            const kvs::UInt32 node_id2 = *( connection++ );
            const kvs::UInt32 node_id3 = *( connection++ );
            const kvs::UInt32 node_id4 = *( connection++ );
            const kvs::UInt32 node_id5 = *( connection++ );
            const kvs::UInt32 node_id6 = *( connection++ );
            const kvs::UInt32 node_id7 = *( connection++ );

            fprintf( ofs, "%u 0 hex %u %u %u %u %u %u %u %u\n",
                     static_cast<unsigned int>( element_id++ ),
                     node_id0,
                     node_id1,
                     node_id2,
                     node_id3,
                     node_id4,
                     node_id5,
                     node_id6,
                     node_id7 );
        }
    }
    else if ( m_element_type == Hexahedra2 )
    {
        while ( connection < end )
        {
            fprintf( ofs, "%u 0 hex2", static_cast<unsigned int>( element_id++ ) );

            const size_t nnodes = 20;
            for ( size_t i = 0; i < nnodes; i++ )
            {
                const kvs::UInt32 node_id = *( connection++ );
                fprintf( ofs, " %u", node_id );
            }

            fprintf( ofs, "\n" );
        }
    }
    else
    {
        throw "Unknown element type.";
    }
}

void AVSUcd::write_components( FILE* ofs ) const
{
    fprintf( ofs, "%u", static_cast<unsigned int>( m_ncomponents_per_node ) );
    for ( size_t i = 0; i < m_ncomponents_per_node; ++i )
    {
        fprintf( ofs, " %u", static_cast<unsigned int>( m_veclens[ i ] ) );
    }
    fprintf( ofs, "\n" );

    for ( size_t i = 0; i < m_ncomponents_per_node; ++i )
    {
        fprintf( ofs, "%s,%s\n", m_component_names[ i ].c_str(), m_component_units[ i ].c_str() );
    }
}

void AVSUcd::write_values( FILE* ofs ) const
{
    size_t node_id = 1;

    const kvs::Real32*       value = m_values.pointer();
    const kvs::Real32* const end   = value + m_values.size();

    while ( value < end )
    {
        fprintf( ofs, "%u %f\n", static_cast<unsigned int>( node_id++ ), *( value++ ) );
    }
}

const bool AVSUcd::CheckFileExtension( const std::string& filename )
{
    const kvs::File file( filename );
    if ( file.extension() == "ucd" || file.extension() == "UCD" ||
            file.extension() == "inp" || file.extension() == "INP" )
    {
        return( true );
    }

    return( false );
}

const bool AVSUcd::CheckFileFormat( const std::string& filename )
{
    FILE* ifs = fopen( filename.c_str(), "rb" );
    if( !ifs )
    {
        kvsMessageError( "Cannot open %s.", filename.c_str() );
        return( false );
    }

    const AVSUcd::FormatType format_type = ::CheckFormatType( ifs );
    if ( format_type == AVSUcd::SingleStep )
    {
        char buffer[ ::MaxLineLength ];
        while ( fgets( buffer, ::MaxLineLength, ifs ) != 0 )
        {
            // Skip comment line.
            if ( buffer[0] == '#' )
            {
                continue;
            }
            else
            {
                const long nnodes = strtol( strtok( buffer, ::Delimiter ), NULL, 10 );
                if ( nnodes == 0 )
                {
                    fclose( ifs );
                    return( false );
                }

                const long nelements = strtol( strtok( 0, ::Delimiter ), NULL, 10 );
                if ( nelements == 0 )
                {
                    fclose( ifs );
                    return( false );
                }

                const long nvalues_per_node = strtol( strtok( 0, ::Delimiter ), NULL, 10 );
                if ( nvalues_per_node == 0 )
                {
                    fclose( ifs );
                    return( false );
                }

                fclose( ifs );
                return( true );
            }
        }
    }
    else if ( format_type == AVSUcd::MultiStep )
    {
        char buffer[ ::MaxLineLength ];
        while ( fgets( buffer, ::MaxLineLength, ifs ) != 0 )
        {
            // Skip comment line.
            if ( buffer[0] == '#' )
            {
                continue;
            }
            else
            {
                const long nsteps = strtol( strtok( buffer, ::Delimiter ), NULL, 10 );
                if ( nsteps == 0 )
                {
                    fclose( ifs );
                    return( false );
                }

                if ( fgets( buffer, ::MaxLineLength, ifs ) != 0 )
                {
                    const char* const cycle_type = strtok( buffer, ::Delimiter );
                    if ( !strcmp( cycle_type, "data" ) ||
                            !strcmp( cycle_type, "geom" ) ||
                            !strcmp( cycle_type, "data_geom" ) )
                    {
                        fclose( ifs );
                        return( true );
                    }
                }

                fclose( ifs );
                return( false );
            }
        }
    }

    fclose( ifs );
    return( false );
}

} // end of namespace kvs
