#include "KVSMLObjectPointMPIWriter.h"
#include "JobDispatcher.h"
#include <string>
#include <sstream>

#include <mpi.h>

KVSMLObjectPointMPIWriter::KVSMLObjectPointMPIWriter( JobDispatcher* pjd )
{
    jd = pjd;
}

void KVSMLObjectPointMPIWriter::Write( kvs::PointObject* object, std::string basename )
{
    int size = MPI::COMM_WORLD.Get_size();
    int rank = MPI::COMM_WORLD.Get_rank();

    m_object = object;
    m_basename = basename;
    m_lastRank = rank;

    if ( rank == 0 )
    {
        size_t nvertices;
        MPI::Status stat;
        MPI::COMM_WORLD.Recv( &nvertices,sizeof( size_t ),MPI_BYTE,MPI_ANY_SOURCE,1,stat );

        int src = stat.Get_source();
        m_lastRank = src;
        JobDispatcher::job_info ji = jd->GetWorkerJobInfo( src-1 );

        std::stringstream suffix;
        suffix << '_' << std::setw( 2 ) << std::setfill( '0' ) << ji.step
               << '-' << std::setw( 2 ) << std::setfill( '0' ) << ji.volume;
        m_basename += suffix.str();

        write_main_tag( src,nvertices );
        write_coords( src,nvertices );
        write_colors( src,nvertices );
        write_normals( src,nvertices );
    }
    else
    {
        size_t nvertices = m_object->nvertices();
        size_t nmemb = nvertices * 3;
        MPI::COMM_WORLD.Send( &nvertices,sizeof( size_t ),MPI_BYTE,0,1 );
        send_coords( nvertices );
        send_colors( nvertices );
        send_normals( nvertices );
    }
}

void KVSMLObjectPointMPIWriter::write_main_tag( int src, size_t nvertices )
{
    std::string kvsml_filename    = m_basename + ".kvsml";
    std::string coords_filename   = m_basename + "_coords.dat";
    std::string colors_filename   = m_basename + "_colors.dat";
    std::string normals_filename  = m_basename + "_normals.dat";

    FILE* kvsml_file = fopen( kvsml_filename.c_str(), "w" );

    fprintf( kvsml_file, "<KVSML>\n" );
    fprintf( kvsml_file, "<Object type=\"PointObject\">\n" );
    fprintf( kvsml_file, "<PointObject>\n" );
    fprintf( kvsml_file, "<Vertex nvertices=\"%zu\">\n",nvertices );
    fprintf( kvsml_file, "<Coord>\n" );
    fprintf( kvsml_file, "<DataArray type=\"float\" format=\"binary\" file=\"%s\"/>\n", coords_filename.c_str() );
    fprintf( kvsml_file, "</Coord>\n" );
    fprintf( kvsml_file, "<Color>\n" );
    fprintf( kvsml_file, "<DataArray type=\"uchar\" format=\"binary\" file=\"%s\"/>\n", colors_filename.c_str() );
    fprintf( kvsml_file, "</Color>\n" );
    fprintf( kvsml_file, "<Normal>\n" );
    fprintf( kvsml_file, "<DataArray type=\"float\" format=\"binary\" file=\"%s\"/>\n", normals_filename.c_str() );
    fprintf( kvsml_file, "</Normal>\n" );
    fprintf( kvsml_file, "<Size>\n" );
    fprintf( kvsml_file, "<DataValue>\n" );
    fprintf( kvsml_file, "1.0f\n" );
    fprintf( kvsml_file, "</DataValue>\n" );
    fprintf( kvsml_file, "</Size>\n" );
    fprintf( kvsml_file, "</Vertex>\n" );
    fprintf( kvsml_file, "</PointObject>\n" );
    fprintf( kvsml_file, "</Object>\n" );
    fprintf( kvsml_file, "</KVSML>\n" );

    fclose( kvsml_file );
}

void KVSMLObjectPointMPIWriter::write_coords( int src, size_t nvertices )
{
    std::string coords_filename = m_basename + "_coords.dat";
    FILE* stream = fopen( coords_filename.c_str(), "wb" );

    size_t nmemb = nvertices * 3;
    float* ptr = new float[nmemb];
    MPI::COMM_WORLD.Recv( ptr,nmemb,MPI_FLOAT,src,1 );

    size_t size = sizeof( float );

    fwrite( ptr, size, nmemb, stream );

    fclose( stream );
}

void KVSMLObjectPointMPIWriter::write_colors( int src, size_t nvertices )
{
    std::string colors_filename = m_basename +"_colors.dat";
    FILE* stream = fopen( colors_filename.c_str(), "wb" );

    size_t nmemb = nvertices * 3;
    unsigned char* ptr = new unsigned char[nmemb];
    MPI::COMM_WORLD.Recv( ptr,nmemb,MPI_UNSIGNED_CHAR,src,1 );

    size_t size = sizeof( unsigned char );

    fwrite( ptr, size, nmemb, stream );

    delete[] ptr;
    fclose( stream );
}

void KVSMLObjectPointMPIWriter::write_normals( int src, size_t nvertices )
{
    std::string normals_filename = m_basename + "_normals.dat";
    FILE* stream = fopen( normals_filename.c_str(), "wb" );

    size_t nmemb = nvertices * 3;
    float* ptr = new float[nmemb];
    MPI::COMM_WORLD.Recv( ptr,nmemb,MPI_FLOAT,src,1 );

    size_t size = sizeof( float );

    fwrite( ptr, size, nmemb, stream );

    delete[] ptr;
    fclose( stream );
}

void KVSMLObjectPointMPIWriter::send_coords( size_t nvertices )
{
    const float* ptr = m_object->coords().pointer();
    size_t nmemb = nvertices * 3;
    MPI::COMM_WORLD.Send( ptr,nmemb,MPI_FLOAT,0,1 );
}

void KVSMLObjectPointMPIWriter::send_colors( size_t nvertices )
{
    const unsigned char* ptr = m_object->colors().pointer();
    size_t nmemb = nvertices * 3;
    MPI::COMM_WORLD.Send( ptr,nmemb,MPI_UNSIGNED_CHAR,0,1 );
}

void KVSMLObjectPointMPIWriter::send_normals( size_t nvertices )
{
    const float* ptr = m_object->normals().pointer();
    size_t nmemb = nvertices * 3;
    MPI::COMM_WORLD.Send( ptr,nmemb,MPI_FLOAT,0,1 );
}

