#include "../FunctionParser/function.h"
#include "../FunctionParser/function_op.h"
#include "../FunctionParser/function_parser.h"

#include "../TransferFunction.h"

#include "../CellBase.h"
#include "../CellByCellParticleGenerator.h"

#ifdef KMATH
#include <kmath_random.h>
KMATH_Random km_random;
#endif

#include <map>

// Dictionaries
typedef std::map<std::string, kvs::TransferFunction*> TransferFunctionMap;
typedef std::map<std::string, FuncParser::Variable>   VariableMap;
typedef std::map<std::string, std::string>            FunctionStringMap;
typedef std::map<std::string, FuncParser::Function>   FunctionMap;
typedef std::map<std::string, size_t>                 QuantityMap;
typedef std::map<std::string, std::string>            TransfuncQuantityBindMap;

using FuncParser::Variable;
using FuncParser::Variables;
using FuncParser::Function;
using FuncParser::FunctionParser;
using kvs::CellByCellParticleGenerator::CalculateDensityConstaint;
using kvs::CellByCellParticleGenerator::CalculateDensity;

void InitPrimitiveTransferFunction( TransferFunctionMap* map,
                                    const size_t resolution,
                                    const float  min_value,
                                    const float  max_value )
{
    ( *map )["t1"] = new kvs::TransferFunction();
    ( *map )["t2"] = new kvs::TransferFunction();
    ( *map )["t3"] = new kvs::TransferFunction();
    ( *map )["t4"] = new kvs::TransferFunction();
    ( *map )["t5"] = new kvs::TransferFunction();

    ( *map )["t1"]->setRange( min_value,max_value );
    ( *map )["t2"]->setRange( min_value,max_value );
    ( *map )["t3"]->setRange( min_value,max_value );
    ( *map )["t4"]->setRange( min_value,max_value );
    ( *map )["t5"]->setRange( min_value,max_value );

    ( *map )["t1"]->create( resolution );
    ( *map )["t2"]->create( resolution );
    ( *map )["t3"]->create( resolution );
    ( *map )["t4"]->create( resolution );
    ( *map )["t5"]->create( resolution );
}

void AddVariant( const std::string& nm, VariableMap* var_map, Variables* vars )
{
    ( *var_map )[nm] = Variable();
    Variable& var = ( *var_map )[nm];
    char *str = new char[nm.size()+1];
    memcpy( str, (char*)nm.c_str(), nm.size()+1 );
    var.tag( str );
    vars->push_back( var );
}

void InitVolumeFunction( FunctionMap* map, const FunctionStringMap& strs, Variables* vars )
{
    for ( FunctionStringMap::const_iterator i = strs.begin(); i != strs.end(); i++ )
    {
        const std::string& nm  = i->first;
        const std::string& str = i->second;
        ( *map )[nm] = Function();
        FunctionParser parse( const_cast<std::string&>( str ), str.size()+1 );
        FunctionParser::Error err = parse.express( ( *map )[nm], *vars );
        if ( err != FunctionParser::ERR_NONE )
        {
            std::cerr << FunctionParser::error_type_to_string[err] << std::endl;
        }
    }
}

void RegistVolumeFunction( const FunctionMap& map, VariableMap* var_map, Variables* vars )
{
    for ( FunctionMap::const_iterator i = map.begin(); i != map.end(); i++ )
    {
        const std::string& nm = i->first;
        AddVariant( nm,var_map,vars );
    }
}

const float CalculateCellOpacity(const TransferFunctionMap& ptf_map,
                                 const TransfuncQuantityBindMap& tfq_map,
                                 const QuantityMap& qmap,
                                 const FunctionMap& volfunc_map,
                                 const FuncParser::Function& synth_func,
                                 kvs::CellBase<kvs::Real32>* cell,
                                 const size_t n,
                                 VariableMap* var_map)
{
//  const kvs::Vector3f coord = cell->vertices()[0]; // coord of the cell
    const kvs::Vector3f coord(n,n,n);

    (*var_map)["X"] = coord.x(); // X coord of the cell
    (*var_map)["Y"] = coord.y(); // Y coord of the cell
    (*var_map)["Z"] = coord.z(); // Z coord of the cell

    for ( FunctionMap::const_iterator i = volfunc_map.begin(); i != volfunc_map.end(); i++ )
    {
        const std::string& nm = i->first;
        const Function& f = i->second;

        // set evaluation value to variabe
        (*var_map)[nm] = f.eval();
    }

    // pull t1,t2,...,t5 with X,Y,Z,q1,q2,...,qn,f1,f2,...,fn
    for ( TransferFunctionMap::const_iterator t = ptf_map.begin(); t != ptf_map.end(); t++ )
    {
        const std::string& nm = t->first;
        const kvs::TransferFunction* f = t->second;

        // physical quantity of the cell
        const std::string& qnm = tfq_map.find(nm)->second;
        const char c1 = qnm.data()[0];
        float scalar;

        // bind primitive transfer function's variable.
        //   To use regex is better than this method,
        //   but regex is unsapported in before C++11...
        if( c1 == 'q' )
        {
//          const size_t qi = qmap.find(qnm)->second;
//          cell->bindCell(n, qi);
//          scalar = cell->averagedScalar();
            scalar = 100;
        }
        else if( c1 == 'f' )
        {
            scalar = (*var_map)[qnm].eval();
        }
        else if( c1 == 'X' )
        {
            scalar = coord.x(); // X-coord of the cell
        }
        else if( c1 == 'Y' )
        {
            scalar = coord.y(); // Y-coord of the cell
        }
        else if( c1 == 'Z' )
        {
            scalar = coord.z(); // Z-coord of the cell
        }
        else
        {
            // error
        }

        // set value to stab variable
        (*var_map)[nm] = f->opacityMap().at(scalar);
    }

    // eval synthesized transfer function
    float opacity = synth_func.eval();

    return opacity;
}

const kvs::RGBColor CalculatePointColor(const TransferFunctionMap& ptf_map,
                                        const TransfuncQuantityBindMap& tfq_map,
                                        const QuantityMap& qmap,
                                        const FunctionMap& volfunc_map,
                                        const FuncParser::Function& synth_func,
                                        kvs::CellBase<kvs::Real32>* cell,
                                        const size_t n,
                                        const kvs::Vector3f& coord,
                                        VariableMap* var_map)
{
    (*var_map)["X"] = coord.x(); // X coord of the point
    (*var_map)["Y"] = coord.y(); // Y coord of the point
    (*var_map)["Z"] = coord.z(); // Z coord of the point

    for ( FunctionMap::const_iterator i = volfunc_map.begin(); i != volfunc_map.end(); i++ )
    {
        const std::string& nm = i->first;
        const Function& f = i->second;

        // set evaluation value to variabe
        (*var_map)[nm] = f.eval();
    }

    kvs::UInt8 col[3];
    for ( unsigned c = 0; c < 3; c++ ) // for each RGB element
    {

        // pull t1,t2,...,t5 with X,Y,Z,q1,q2,...,qn,f1,f2,...,fn
        for ( TransferFunctionMap::const_iterator t = ptf_map.begin(); t != ptf_map.end(); t++ )
        {
            const std::string& nm = t->first;
            const kvs::TransferFunction* f = t->second;

            // physical quantity of the point
            const std::string& qnm = tfq_map.find(nm)->second;
            const char c1 = qnm.data()[0];
            float scalar;

            // bind primitive transfer function's variable.
            //   To use regex is better than this method,
            //   but regex is unsapported in before C++11...
            if( c1 == 'q' )
            {
//              const size_t qi = qmap.find(qnm)->second;
//              cell->bindCell(n, qi);
//              scalar = cell->averagedScalar();
                scalar = 100;
            }
            else if( c1 == 'f' )
            {
                scalar = (*var_map)[qnm].eval();
            }
            else if( c1 == 'X' )
            {
                scalar = coord.x(); // X-coord of the point
            }
            else if( c1 == 'Y' )
            {
                scalar = coord.y(); // Y-coord of the point
            }
            else if( c1 == 'Z' )
            {
                scalar = coord.z(); // Z-coord of the point
            }
            else
            {
                // error
            }

            // reference transfer function
            const kvs::RGBColor color( f->colorMap().at( scalar ) );
            const kvs::UInt8 rgb[] = {color.r(),color.g(),color.b()};

            // set value to stab variable
            (*var_map)[nm] = rgb[c];
        }

        // eval synthesized transfer function
        col[c] = synth_func.eval();
    }

    return kvs::RGBColor( col ); // color of the point
}

int main( int argc, char* argv[] )
{

    const float  min_value  =   0.0;
    const float  max_value  = 100.0;
    const size_t resolution = 100;

    // inputed ----------------------------------------------------------

    TransferFunctionMap ptf_map; // primitive transfer function map
    InitPrimitiveTransferFunction( &ptf_map,resolution,min_value,max_value );

    TransfuncQuantityBindMap tfq_map;
    tfq_map["t1"] = "q0";
    tfq_map["t2"] = "q1";
    tfq_map["t3"] = "q2";
    tfq_map["t4"] = "q3";
    tfq_map["t5"] = "q4";

    FunctionStringMap volfuncstr_map;
    volfuncstr_map["f1"] = "X^2+Y^2+Z^2";
    volfuncstr_map["f2"] = "X+Y+Z";

    std::string synth_func_str = "t1+t2";

    QuantityMap qmap;
    qmap["q0"] = 0;
    qmap["q1"] = 1;
    qmap["q2"] = 2;

    // ------------------------------------------------------------------

    // Construct dictionaries and Parse equations -----------------------

    // Variable containers
    Variables vars;
    VariableMap var_map;

    // Variables for X,Y,Z coord
    AddVariant( "X",&var_map,&vars );
    AddVariant( "Y",&var_map,&vars );
    AddVariant( "Z",&var_map,&vars );

    // Vairables for quantities
    for ( QuantityMap::iterator q = qmap.begin(); q != qmap.end(); q++ )
    {
        const std::string& nm = q->first;
        AddVariant( nm,&var_map,&vars );
    }

    // variables which have evaluation value of primitive transfer function
    for ( TransferFunctionMap::iterator i = ptf_map.begin(); i != ptf_map.end(); i++ )
    {
        const std::string& nm = i->first;
        AddVariant( nm,&var_map,&vars );
    }

    // initialize and parse volume shape functions
    FunctionMap volfunc_map;
    InitVolumeFunction( &volfunc_map, volfuncstr_map, &vars );
    RegistVolumeFunction( volfunc_map, &var_map, &vars );

    // initialize synthesized transfer function
    Function synth_func;
    FunctionParser synth_func_parse( synth_func_str, synth_func_str.size()+1 );
    FunctionParser::Error err = synth_func_parse.express( synth_func, vars );

    // ------------------------------------------------------------------

    // pre-sampling -----------------------------------------------------

    float sampling_volume_inverse;
    float max_opacity = 0;
    float max_density = 0;

    for ( TransferFunctionMap::iterator i = ptf_map.begin(); i != ptf_map.end(); i++ )
    {
        float t_max_opacity = 0;
        float t_max_density = 0;
        const kvs::TransferFunction* f = i->second;

//      CalculateDensityConstaint(
//          camera,
//          BaseClass::volume(),
//          static_cast<float>( m_subpixel_level ),
//          m_sampling_step,
//          &sampling_volume_inverse,
//          &t_max_opacity,
//          &t_max_density);

        if (t_max_opacity > max_opacity) max_opacity = t_max_opacity;
        if (t_max_density > max_density) max_density = t_max_density;
    }

    // ------------------------------------------------------------------

    // at sampling routine ----------------------------------------------

    kvs::CellBase<kvs::Real32>* cell;

    // apply the synthesized transfer function for each cells or each points
    for ( size_t n = 0; n < 1; n++ )
    {
//      cell->bindCell(n);
//      const kvs::Vector3f& coord = cell->vertices()[0];
        const kvs::Vector3f coord(10,10,10);

        // set variables q1,q2,...,qn from scalar value of cell
        for ( QuantityMap::iterator q = qmap.begin(); q != qmap.end(); q++ )
        {
            const std::string& nm = q->first;
//          const size_t qi = q->second;
//          cell->bindCell(n,qi);
//          var_map[nm] = cell->averagedScalar();
            var_map[nm] = 100;
        }

        // Calculate cell Opacity
        {
            float opacity;
            float density;

            // calculate opacity
            opacity = CalculateCellOpacity(ptf_map, tfq_map, qmap,
                                           volfunc_map, synth_func,
                                           cell, n, &var_map);

            // calculate density
            density = CalculateDensity(opacity,
                                       sampling_volume_inverse,
                                       max_opacity, max_density);
        }


        for ( unsigned p = 0; p < 1; p++ ) // for each points
        {

            kvs::Vector3f coord(p,p,p); // coord of the point

            // Calculate point color
            const kvs::RGBColor color = CalculatePointColor(ptf_map, tfq_map, qmap,
                                                            volfunc_map, synth_func,
                                                            cell, n, coord, &var_map);

            std::cout << color << std::endl;
        }

    }

    return 0;
}
